/******************************************************************************/
/*                                                                            */
/*    Copyright (c)The 3Delight Developers.                                   */
/*    All Rights Reserved.                                                    */
/*                                                                            */
/******************************************************************************/

/*
	The RenderMan (R) Interface Procedures and RIB Protocol are:
	Copyright 1988, 1989, Pixar.  All rights reserved.
	RenderMan (R) is a registered trademark of Pixar.
*/
#ifndef __ri_h
#define __ri_h

#undef DL_IMPORT

#if defined(WIN32) || defined(_WIN32)
#define DL_IMPORT __declspec(dllimport)
#else
#define DL_IMPORT
#endif


#define __RI_ANSI 1


#ifdef  __cplusplus
extern "C" {
#endif

/* ===========================================================================
    Type definitions
   =========================================================================*/

typedef short RtBoolean;
typedef float RtFloat;

/* Define RtInt as a 32 bit signed integer */
typedef int RtInt;

typedef char *RtToken;

typedef RtFloat RtColor[3];
typedef RtFloat RtPoint[3];
typedef RtFloat RtVector[3];
typedef RtFloat RtNormal[3];
typedef RtFloat RtHpoint[4];
typedef RtFloat RtMatrix[4][4];
typedef RtFloat RtBasis[4][4];
typedef RtFloat RtBound[6];
typedef char *RtString;

typedef void *RtPointer;

#define RtVoid void

typedef RtFloat (*RtFilterFunc)(RtFloat, RtFloat, RtFloat, RtFloat);
typedef RtVoid  (*RtErrorHandler)(RtInt, RtInt, char *);

typedef RtPointer RtObjectHandle;
typedef RtPointer RtLightHandle;


/* ===========================================================================
    Constant definitions
   =========================================================================*/

#define RI_FALSE        0
#define RI_TRUE         (!RI_FALSE)
#define RI_INFINITY     1.0e38f
#define RI_EPSILON      1.0e-10f
#define RI_NULL         ((RtToken)0)


/* ===========================================================================
    Standard token definitions
   =========================================================================*/

DL_IMPORT extern RtToken RI_FRAMEBUFFER;
DL_IMPORT extern RtToken RI_FILE;
DL_IMPORT extern RtToken RI_ZFILE;

DL_IMPORT extern RtToken RI_RGB;
DL_IMPORT extern RtToken RI_RGBA;
DL_IMPORT extern RtToken RI_RGBZ;
DL_IMPORT extern RtToken RI_RGBAZ;
DL_IMPORT extern RtToken RI_A;
DL_IMPORT extern RtToken RI_Z;
DL_IMPORT extern RtToken RI_AZ;

DL_IMPORT extern RtToken RI_PERSPECTIVE;
DL_IMPORT extern RtToken RI_ORTHOGRAPHIC;

DL_IMPORT extern RtToken RI_HIDDEN;
DL_IMPORT extern RtToken RI_PAINT;

DL_IMPORT extern RtToken RI_CONSTANT;
DL_IMPORT extern RtToken RI_SMOOTH;

DL_IMPORT extern RtToken RI_FLATNESS;
DL_IMPORT extern RtToken RI_FOV;

DL_IMPORT extern RtToken RI_AMBIENTLIGHT;
DL_IMPORT extern RtToken RI_POINTLIGHT;
DL_IMPORT extern RtToken RI_DISTANTLIGHT;
DL_IMPORT extern RtToken RI_SPOTLIGHT;

DL_IMPORT extern RtToken RI_INTENSITY;
DL_IMPORT extern RtToken RI_LIGHTCOLOR;
DL_IMPORT extern RtToken RI_FROM;
DL_IMPORT extern RtToken RI_TO;
DL_IMPORT extern RtToken RI_CONEANGLE;
DL_IMPORT extern RtToken RI_CONEDELTAANGLE;
DL_IMPORT extern RtToken RI_BEAMDISTRIBUTION;

DL_IMPORT extern RtToken RI_MATTE;
DL_IMPORT extern RtToken RI_METAL;
DL_IMPORT extern RtToken RI_SHINYMETAL;
DL_IMPORT extern RtToken RI_PLASTIC;
DL_IMPORT extern RtToken RI_PAINTEDPLASTIC;

DL_IMPORT extern RtToken RI_KA;
DL_IMPORT extern RtToken RI_KD;
DL_IMPORT extern RtToken RI_KS;
DL_IMPORT extern RtToken RI_ROUGHNESS;
DL_IMPORT extern RtToken RI_KR;
DL_IMPORT extern RtToken RI_TEXTURENAME;
DL_IMPORT extern RtToken RI_SPECULARCOLOR;

DL_IMPORT extern RtToken RI_DEPTHCUE;
DL_IMPORT extern RtToken RI_FOG;
DL_IMPORT extern RtToken RI_BUMPY;

DL_IMPORT extern RtToken RI_MINDISTANCE;
DL_IMPORT extern RtToken RI_MAXDISTANCE;
DL_IMPORT extern RtToken RI_BACKGROUND;
DL_IMPORT extern RtToken RI_DISTANCE;
DL_IMPORT extern RtToken RI_AMPLITUDE;

DL_IMPORT extern RtToken RI_RASTER;
DL_IMPORT extern RtToken RI_SCREEN;
DL_IMPORT extern RtToken RI_CAMERA;
DL_IMPORT extern RtToken RI_WORLD;
DL_IMPORT extern RtToken RI_OBJECT;

DL_IMPORT extern RtToken RI_INSIDE;
DL_IMPORT extern RtToken RI_OUTSIDE;
DL_IMPORT extern RtToken RI_LH;
DL_IMPORT extern RtToken RI_RH;

DL_IMPORT extern RtToken RI_P;
DL_IMPORT extern RtToken RI_PZ;
DL_IMPORT extern RtToken RI_PW;
DL_IMPORT extern RtToken RI_N;
DL_IMPORT extern RtToken RI_NP;
DL_IMPORT extern RtToken RI_CS;
DL_IMPORT extern RtToken RI_OS;
DL_IMPORT extern RtToken RI_S;
DL_IMPORT extern RtToken RI_T;
DL_IMPORT extern RtToken RI_ST;

DL_IMPORT extern RtToken RI_BILINEAR;
DL_IMPORT extern RtToken RI_BICUBIC;

DL_IMPORT extern RtToken RI_PRIMITIVE;
DL_IMPORT extern RtToken RI_INTERSECTION;
DL_IMPORT extern RtToken RI_UNION;
DL_IMPORT extern RtToken RI_DIFFERENCE;

DL_IMPORT extern RtToken RI_PERIODIC;
DL_IMPORT extern RtToken RI_NONPERIODIC;
DL_IMPORT extern RtToken RI_CLAMP;
DL_IMPORT extern RtToken RI_BLACK;

DL_IMPORT extern RtToken RI_IGNORE;
DL_IMPORT extern RtToken RI_PRINT;
DL_IMPORT extern RtToken RI_ABORT;
DL_IMPORT extern RtToken RI_HANDLER;

DL_IMPORT extern RtToken RI_BOUNDS;

DL_IMPORT extern RtToken RI_LIMITS;
DL_IMPORT extern RtToken RI_SHADOW;
DL_IMPORT extern RtToken RI_BIAS0;
DL_IMPORT extern RtToken RI_BIAS1;
DL_IMPORT extern RtToken RI_SAMPLE;

DL_IMPORT extern RtToken RI_SEARCHPATH;
DL_IMPORT extern RtToken RI_SHADER;
DL_IMPORT extern RtToken RI_TEXTURE;
DL_IMPORT extern RtToken RI_DISPLAY;

DL_IMPORT extern RtToken RI_WIDTH, RI_CONSTANTWIDTH;

DL_IMPORT extern RtToken RI_COMMENT;
DL_IMPORT extern RtToken RI_STRUCTURE;
DL_IMPORT extern RtToken RI_VERBATIM;

/* ===========================================================================
    Curves
   =========================================================================*/

DL_IMPORT extern RtToken RI_LINEAR;
DL_IMPORT extern RtToken RI_CUBIC;

/* ===========================================================================
    3delight tokens declaration
   =========================================================================*/

DL_IMPORT extern RtToken RI_BUCKETSENDER;
DL_IMPORT extern RtToken RI_FUNCTION;




/* ===========================================================================
    Patch basis stuff
   =========================================================================*/

DL_IMPORT extern RtBasis RiBezierBasis;
DL_IMPORT extern RtBasis RiBSplineBasis;
DL_IMPORT extern RtBasis RiCatmullRomBasis;
DL_IMPORT extern RtBasis RiHermiteBasis;

DL_IMPORT extern RtInt RI_BEZIERSTEP;
DL_IMPORT extern RtInt RI_BSPLINESTEP;
DL_IMPORT extern RtInt RI_CATMULLROMSTEP;
DL_IMPORT extern RtInt RI_HERMITESTEP;


/* ===========================================================================
    Renderman Interface calls
   =========================================================================*/

DL_IMPORT RtVoid RiBegin(RtToken name);
DL_IMPORT RtVoid RiEnd(void);
DL_IMPORT RtVoid RiFrameBegin(RtInt frame);
DL_IMPORT RtVoid RiFrameEnd(void);
DL_IMPORT RtVoid RiMotionBegin(RtInt n, ...);
DL_IMPORT RtVoid RiMotionBeginV(RtInt n, RtFloat times[]);
DL_IMPORT RtVoid RiMotionEnd(void);
DL_IMPORT RtVoid RiSolidBegin(RtToken operation);
DL_IMPORT RtVoid RiSolidEnd(void);
DL_IMPORT RtVoid RiWorldBegin(void);
DL_IMPORT RtVoid RiWorldEnd(void);
DL_IMPORT RtObjectHandle RiObjectBegin( void );
DL_IMPORT RtVoid RiObjectEnd(void);
DL_IMPORT RtVoid RiObjectInstance(RtObjectHandle handle);

DL_IMPORT RtVoid RiFormat(RtInt xres, RtInt yres, RtFloat aspect);
DL_IMPORT RtVoid RiFrameAspectRatio(RtFloat aspect);
DL_IMPORT RtVoid RiScreenWindow(RtFloat left, RtFloat right,
								RtFloat bottom, RtFloat top);
DL_IMPORT RtVoid RiClipping(RtFloat hither, RtFloat yon);
DL_IMPORT RtVoid RiCropWindow(RtFloat xmin, RtFloat xmax,
							  RtFloat ymin, RtFloat ymax);
DL_IMPORT RtVoid RiDepthOfField(RtFloat fstop, RtFloat focallength,
								RtFloat focaldistance);
DL_IMPORT RtVoid RiProjection(RtToken name, ...);
DL_IMPORT RtVoid RiProjectionV(RtToken name, RtInt n,
							   RtToken tokens[], RtPointer parms[]);
DL_IMPORT RtVoid RiShutter(RtFloat min, RtFloat max);

DL_IMPORT RtVoid RiDisplay(char *name, RtToken type, RtToken mode, ...);
DL_IMPORT RtVoid RiDisplayV(char *name, RtToken type, RtToken mode,
							RtInt n, RtToken tokens[],
							RtPointer parms[]);
DL_IMPORT RtVoid RiExposure(RtFloat gain, RtFloat gamma);
DL_IMPORT RtVoid RiImager(RtToken name, ...);
DL_IMPORT RtVoid RiImagerV(RtToken name, RtInt n, RtToken tokens[],
						   RtPointer parms[]);
DL_IMPORT RtVoid RiPixelFilter(RtFilterFunc filterfunc,
							   RtFloat xwidth, RtFloat ywidth);
DL_IMPORT RtVoid RiPixelSamples(RtFloat xsamples, RtFloat ysamples);
DL_IMPORT RtVoid RiPixelVariance(RtFloat variation);
DL_IMPORT RtVoid RiQuantize(RtToken type, RtInt one, RtInt min,
							RtInt max, RtFloat ampl);

DL_IMPORT RtVoid RiConcatTransform(RtMatrix transform);
DL_IMPORT RtVoid RiCoordinateSystem(RtToken space );
DL_IMPORT RtVoid RiCoordSysTransform(RtToken space);
DL_IMPORT RtVoid RiIdentity(void);
DL_IMPORT RtVoid RiPerspective(RtFloat fov);
DL_IMPORT RtVoid RiRotate(RtFloat angle,
						  RtFloat dx, RtFloat dy, RtFloat dz);
DL_IMPORT RtVoid RiScale(RtFloat dx, RtFloat dy, RtFloat dz);
DL_IMPORT RtVoid RiSkew(RtFloat angle,
						RtFloat dx1, RtFloat dy1, RtFloat dz1,
						RtFloat dx2, RtFloat dy2, RtFloat dz2);
DL_IMPORT RtVoid RiTransform(RtMatrix transform);
DL_IMPORT RtVoid RiTransformBegin(void);
DL_IMPORT RtVoid RiTransformEnd(void);
DL_IMPORT RtVoid RiTranslate(RtFloat dx, RtFloat dy, RtFloat dz);
DL_IMPORT RtPoint* RiTransformPoints(RtToken fromspace,
									 RtToken tospace,
									 RtInt n, RtPoint points[]);

DL_IMPORT RtVoid RiAtmosphere(RtToken name, ...);
DL_IMPORT RtVoid RiAtmosphereV(RtToken name, RtInt n,
							   RtToken tokens[], RtPointer parms[]);
DL_IMPORT RtVoid RiDeformation(RtToken name, ...);
DL_IMPORT RtVoid RiDeformationV(RtToken name, RtInt n,
								RtToken tokens[], RtPointer parms[]);
DL_IMPORT RtVoid RiDisplacement(RtToken name, ...);
DL_IMPORT RtVoid RiDisplacementV(RtToken name, RtInt n,
								 RtToken tokens[], RtPointer parms[]);
DL_IMPORT RtVoid RiExterior(RtToken name, ...);
DL_IMPORT RtVoid RiExteriorV(RtToken name, RtInt n,
							 RtToken tokens[], RtPointer parms[]);
DL_IMPORT RtVoid RiIlluminate(RtLightHandle light, RtBoolean onoff);
DL_IMPORT RtVoid RiInterior(RtToken name, ...);
DL_IMPORT RtVoid RiInteriorV(RtToken name, RtInt n,
							 RtToken tokens[], RtPointer parms[]);
DL_IMPORT RtVoid RiMatte(RtBoolean onoff);
DL_IMPORT RtVoid RiShadingRate(RtFloat size);
DL_IMPORT RtVoid RiShadingInterpolation(RtToken type);
DL_IMPORT RtVoid RiSurface(RtToken name, ...);
DL_IMPORT RtVoid RiSurfaceV(RtToken name, RtInt n,
							RtToken tokens[], RtPointer parms[]);

DL_IMPORT RtVoid RiArchiveRecord(RtToken type, char *format, ...);
DL_IMPORT RtVoid RiReadArchive(RtToken name,
							   RtVoid (*callback)(RtToken,char*, ...),
							   ...);
DL_IMPORT RtVoid RiReadArchiveV(RtToken name,
								RtVoid (*callback)(RtToken, char*, ...),
								RtInt n, RtToken tokens[], RtPointer parms[]);

DL_IMPORT RtVoid RiAttribute(RtToken name, ...);
DL_IMPORT RtVoid RiAttributeV(RtToken name, RtInt n, RtToken tokens[],
							  RtPointer parms[]);
DL_IMPORT RtVoid RiAttributeBegin(void);
DL_IMPORT RtVoid RiAttributeEnd(void);
DL_IMPORT RtVoid RiBound( RtBound bound );
DL_IMPORT RtVoid RiColor( RtColor color );
DL_IMPORT RtVoid RiOpacity(RtColor color);
DL_IMPORT RtVoid RiOption(RtToken name, ...);
DL_IMPORT RtVoid RiOptionV(RtToken name, RtInt n,
						   RtToken tokens[], RtPointer parms[]);
DL_IMPORT RtVoid RiReverseOrientation(void);
DL_IMPORT RtVoid RiTextureCoordinates(RtFloat s1, RtFloat t1,
									  RtFloat s2, RtFloat t2,
									  RtFloat s3, RtFloat t3,
									  RtFloat s4, RtFloat t4);
DL_IMPORT RtVoid RiSides(RtInt sides);

DL_IMPORT RtToken RiDeclare(char *name, char *declaration);

DL_IMPORT RtLightHandle RiLightSource(RtToken name, ...);
DL_IMPORT RtLightHandle RiLightSourceV(RtToken name, RtInt n,
									   RtToken tokens[],
									   RtPointer parms[]);
DL_IMPORT RtLightHandle RiAreaLightSource(RtToken name, ...);
DL_IMPORT RtLightHandle RiAreaLightSourceV(RtToken name, RtInt n,
										   RtToken tokens[],
										   RtPointer parms[]);

DL_IMPORT RtVoid RiBasis(RtBasis ubasis, RtInt ustep,
						 RtBasis vbasis, RtInt vstep);
DL_IMPORT RtVoid RiPatch(RtToken type, ...);
DL_IMPORT RtVoid RiPatchMesh(RtToken type, RtInt nu, RtToken uwrap,
							 RtInt nv, RtToken vwrap, ...);
DL_IMPORT RtVoid RiPatchV(RtToken type, RtInt n,
						  RtToken tokens[], RtPointer parms[]);
DL_IMPORT RtVoid RiPatchMeshV(RtToken type, RtInt nu, RtToken uwrap,
							  RtInt nv, RtToken vwrap, RtInt n,
							  RtToken tokens[], RtPointer parms[]);
DL_IMPORT RtVoid RiPoints(RtInt npoints, ...);
DL_IMPORT RtVoid RiPointsV(RtInt npoints, RtInt n,
						   RtToken tokens[], RtPointer parms[]);
DL_IMPORT RtVoid RiCurves(RtToken type, RtInt ncurves,
						  RtInt nvertices[], RtToken wrap, ...);
DL_IMPORT RtVoid RiCurvesV(RtToken type, RtInt ncurves,
						   RtInt nvertices[], RtToken wrap,
						   RtInt n, RtToken tokens[],
						   RtPointer parms[]);

DL_IMPORT RtVoid RiNuPatch(RtInt nu, RtInt uorder, RtFloat uknot[],
						   RtFloat umin, RtFloat umax,
						   RtInt nv, RtInt vorder, RtFloat vknot[],
						   RtFloat vmin, RtFloat vmax, ...);
DL_IMPORT RtVoid RiNuPatchV(RtInt nu, RtInt uorder, RtFloat uknot[],
							RtFloat umin, RtFloat umax,
							RtInt nv, RtInt vorder, RtFloat vknot[],
							RtFloat vmin, RtFloat vmax, RtInt n,
							RtToken tokens[], RtPointer parms[]);
DL_IMPORT RtVoid RiTrimCurve(RtInt nloops, RtInt ncurves[],
							 RtInt order[], RtFloat knot[],
							 RtFloat min[], RtFloat max[], RtInt n[],
							 RtFloat u[], RtFloat v[], RtFloat w[]);

DL_IMPORT RtVoid RiSubdivisionMesh(RtToken scheme,
								   RtInt nfaces, RtInt nvertices[], RtInt vertices[],
								   RtInt ntags, RtToken tags[], RtInt nargs[],
								   RtInt intargs[], RtFloat floatargs[],
								   ...);
DL_IMPORT RtVoid RiSubdivisionMeshV(RtToken scheme,
									RtInt nfaces, RtInt nvertices[], RtInt vertices[],
									RtInt ntags, RtToken tags[], RtInt nargs[],
									RtInt intargs[], RtFloat floatargs[],
									RtInt n, RtToken tokens[], RtPointer parms[]);

DL_IMPORT RtVoid RiCone(RtFloat height, RtFloat radius,
						RtFloat thetamax, ...);
DL_IMPORT RtVoid RiConeV(RtFloat height, RtFloat radius,
						 RtFloat thetamax, RtInt n,
						 RtToken tokens[], RtPointer parms[]);
DL_IMPORT RtVoid RiCylinder(RtFloat radius,RtFloat zmin,RtFloat zmax,
							RtFloat thetamax,...);
DL_IMPORT RtVoid RiCylinderV(RtFloat radius,RtFloat zmin,RtFloat zmax,
							 RtFloat thetamax, RtInt n,
							 RtToken tokens[], RtPointer parms[]);
DL_IMPORT RtVoid RiDisk(RtFloat height, RtFloat radius,
						RtFloat thetamax, ...);
DL_IMPORT RtVoid RiDiskV(RtFloat height, RtFloat radius,
						 RtFloat thetamax, RtInt n,
						 RtToken tokens[], RtPointer parms[]);
DL_IMPORT RtVoid RiHyperboloid(RtPoint point1, RtPoint point2,
							   RtFloat thetamax, ...);
DL_IMPORT RtVoid RiHyperboloidV(RtPoint point1, RtPoint point2,
								RtFloat thetamax, RtInt n,
								RtToken tokens[], RtPointer parms[]);
DL_IMPORT RtVoid RiParaboloid(RtFloat rmax,RtFloat zmin,RtFloat zmax,
							  RtFloat thetamax,...);
DL_IMPORT RtVoid RiParaboloidV(RtFloat rmax,RtFloat zmin,RtFloat zmax,
							   RtFloat thetamax, RtInt n,
							   RtToken tokens[], RtPointer parms[]);
DL_IMPORT RtVoid RiSphere(RtFloat radius,RtFloat zmin,RtFloat zmax,
						  RtFloat thetamax, ...);
DL_IMPORT RtVoid RiSphereV(RtFloat radius,RtFloat zmin,RtFloat zmax,
						   RtFloat thetamax, RtInt n,
						   RtToken tokens[], RtPointer parms[]);
DL_IMPORT RtVoid RiTorus(RtFloat majorradius, RtFloat minorradius,
						 RtFloat phimin, RtFloat phimax,
						 RtFloat thetamax, ...);
DL_IMPORT RtVoid RiTorusV(RtFloat majorradius, RtFloat minorradius,
						  RtFloat phimin, RtFloat phimax,
						  RtFloat thetamax, RtInt n,
						  RtToken tokens[], RtPointer parms[]);

DL_IMPORT RtVoid RiGeneralPolygon(RtInt nloops,
								  RtInt nvertices[], ...);
DL_IMPORT RtVoid RiGeneralPolygonV(RtInt nloops, RtInt nvertices[],
								   RtInt n, RtToken tokens[],
								   RtPointer parms[]);

DL_IMPORT RtVoid RiBlobby(
	RtInt nleaf,
	RtInt nentry, RtInt entry[],
	RtInt nfloat, RtFloat floats[],
	RtInt nstring, RtString strings[],
	RtInt n, RtToken tokens[], RtPointer parms[]);

DL_IMPORT RtVoid RiPointsGeneralPolygons(RtInt npolys, RtInt nloops[],
										 RtInt nvertices[],
										 RtInt vertices[], ...);
DL_IMPORT RtVoid RiPointsGeneralPolygonsV(RtInt npolys, RtInt nloops[],
										  RtInt nvertices[],
										  RtInt vertices[],
										  RtInt n, RtToken tokens[],
										  RtPointer parms[]);
DL_IMPORT RtVoid RiPointsPolygons(RtInt npolys, RtInt nvertices[],
								  RtInt vertices[], ...);
DL_IMPORT RtVoid RiPointsPolygonsV(RtInt npolys, RtInt nvertices[],
								   RtInt vertices[],
								   RtInt n, RtToken tokens[],
								   RtPointer parms[]);
DL_IMPORT RtVoid RiPolygon(RtInt nvertices, ...);
DL_IMPORT RtVoid RiPolygonV(RtInt nvertices,
							RtInt n, RtToken tokens[],
							RtPointer parms[]);

DL_IMPORT RtVoid RiColorSamples(RtInt n, RtFloat nRGB[],
								RtFloat RGBn[]);
DL_IMPORT RtVoid RiHider(RtToken type, ...);
DL_IMPORT RtVoid RiHiderV(RtToken type, RtInt n, RtToken tokens[],
						  RtPointer parms[]);
DL_IMPORT RtVoid RiDetail(RtBound bound);
DL_IMPORT RtVoid RiDetailRange(RtFloat minvisible,
							   RtFloat lowertransition,
							   RtFloat uppertransition,
							   RtFloat maxvisible);
DL_IMPORT RtVoid RiGeometricApproximation(RtToken type, RtFloat value);
DL_IMPORT RtVoid RiGeometry(RtToken type, ...);
DL_IMPORT RtVoid RiGeometryV(RtToken type,
							 RtInt n, RtToken tokens[],
							 RtPointer parms[]);
DL_IMPORT RtVoid RiOrientation(RtToken orientation);
DL_IMPORT RtVoid RiProcedural(RtPointer i_data, RtBound i_bound,
							  RtVoid (*i_Subdivfunc)(RtPointer, RtFloat),
							  RtVoid (*i_Freefunc)(RtPointer));
DL_IMPORT RtVoid RiRelativeDetail(RtFloat relativedetail);
DL_IMPORT RtVoid RiProcDelayedReadArchive(RtPointer data,
										  RtFloat detail);
DL_IMPORT RtVoid RiProcDynamicLoad(RtPointer data, RtFloat detail);
DL_IMPORT RtVoid RiProcRunProgram(RtPointer data, RtFloat detail);
DL_IMPORT RtVoid RiProcFree(RtPointer data);

DL_IMPORT RtVoid RiMakeBump(char *picturename, char *texturename,
							RtToken swrap, RtToken twrap,
							RtFilterFunc filterfunc,
							RtFloat swidth, RtFloat twidth, ...);
DL_IMPORT RtVoid RiMakeBumpV(char *picturename, char *texturename,
							 RtToken swrap, RtToken twrap,
							 RtFilterFunc filterfunc,
							 RtFloat swidth, RtFloat twidth,
							 RtInt n, RtToken tokens[],
							 RtPointer parms[]);
DL_IMPORT RtVoid RiMakeCubeFaceEnvironment(char *px, char *nx,
										   char *py, char *ny,
										   char *pz, char *nz,
										   char *texturename,
										   RtFloat fov,
										   RtFilterFunc filterfunc,
										   RtFloat swidth,
										   RtFloat twidth, ...);
DL_IMPORT RtVoid RiMakeCubeFaceEnvironmentV(char *px, char *nx,
											char *py, char *ny,
											char *pz, char *nz,
											char *texturename,
											RtFloat fov,
											RtFilterFunc filterfunc,
											RtFloat swidth,
											RtFloat twidth,
											RtInt n,
											RtToken tokens[],
											RtPointer parms[]);
DL_IMPORT RtVoid RiMakeLatLongEnvironment(char *picturename,
										  char *texturename,
										  RtFilterFunc filterfunc,
										  RtFloat swidth,
										  RtFloat twidth, ...);
DL_IMPORT RtVoid RiMakeLatLongEnvironmentV(char *picturename,
										   char *texturename,
										   RtFilterFunc filterfunc,
										   RtFloat swidth,
										   RtFloat twidth,
										   RtInt n,
										   RtToken tokens[],
										   RtPointer parms[]);
DL_IMPORT RtVoid RiMakeShadow(char *picturename,
							  char *texturename, ...);
DL_IMPORT RtVoid RiMakeShadowV(char *picturename,
							   char *texturename,
							   RtInt n,
							   RtToken tokens[],
							   RtPointer parms[]);
DL_IMPORT RtVoid RiMakeTexture(char *picturename, char *texturename,
							   RtToken swrap, RtToken twrap,
							   RtFilterFunc filterfunc,
							   RtFloat swidth, RtFloat twidth, ...);
DL_IMPORT RtVoid RiMakeTextureV(char *picturename, char *texturename,
								RtToken swrap, RtToken twrap,
								RtFilterFunc filterfunc,
								RtFloat swidth, RtFloat twidth,
								RtInt n, RtToken tokens[],
								RtPointer parms[]);

DL_IMPORT RtFloat RiBoxFilter(RtFloat x, RtFloat y,
							  RtFloat xwidth, RtFloat ywidth);
DL_IMPORT RtFloat RiTriangleFilter(RtFloat x, RtFloat y,
								   RtFloat xwidth, RtFloat ywidth);
DL_IMPORT RtFloat RiCatmullRomFilter(RtFloat x, RtFloat y,
									 RtFloat xwidth, RtFloat ywidth);
DL_IMPORT RtFloat RiBesselFilter(RtFloat x, RtFloat y,
								 RtFloat xwidth, RtFloat ywidth);
DL_IMPORT RtFloat RiGaussianFilter(RtFloat x, RtFloat y,
								   RtFloat xwidth, RtFloat ywidth);
DL_IMPORT RtFloat RiSincFilter(RtFloat x, RtFloat y,
							   RtFloat xwidth, RtFloat ywidth);

DL_IMPORT RtVoid RiErrorHandler( RtErrorHandler handler);
DL_IMPORT RtVoid RiErrorIgnore(RtInt code, RtInt severity, char *msg);
DL_IMPORT RtVoid RiErrorPrint(RtInt code, RtInt severity, char *msg);
DL_IMPORT RtVoid RiErrorAbort(RtInt code, RtInt severity, char *msg);



/* ===========================================================================
    Error codes
   ===========================================================================

	 1 - 10		System and File Errors
	11 - 20		Program Limitations
	21 - 40		State Errors
	41 - 60		Parameter and Protocol Errors
	61 - 80		Execution Errors
   =========================================================================*/

#define RIE_NOERROR     ((RtInt)0)

#define RIE_NOMEM       ((RtInt)1)	 /* Out of memory */
#define RIE_SYSTEM      ((RtInt)2)	 /* Miscellaneous system error */
#define RIE_NOFILE      ((RtInt)3)	 /* File nonexistent */
#define RIE_BADFILE     ((RtInt)4)	 /* Bad file format */
#define RIE_VERSION     ((RtInt)5)	 /* File version mismatch */
#define RIE_DISKFULL    ((RtInt)6)	 /* Target disk is full */

#define RIE_INCAPABLE   ((RtInt)11)	 /* Optional RI feature */
#define RIE_UNIMPLEMENT ((RtInt)12)	 /* Unimplemented feature */
#define RIE_LIMIT       ((RtInt)13)	 /* Arbitrary program limit */
#define RIE_BUG         ((RtInt)14)	 /* Probably a bug in renderer */

#define RIE_NOTSTARTED  ((RtInt)23)	 /* RiBegin not called */
#define RIE_NESTING     ((RtInt)24)	 /* Bad begin-end nesting */
#define RIE_NOTOPTIONS  ((RtInt)25)	 /* Invalid state for options */
#define RIE_NOTATTRIBS  ((RtInt)26)	 /* Invalid state for attribs */
#define RIE_NOTPRIMS    ((RtInt)27)	 /* Invalid state for primitives */
#define RIE_ILLSTATE    ((RtInt)28)	 /* Other invalid state */
#define RIE_BADMOTION   ((RtInt)29)	 /* Badly formed motion block */
#define RIE_BADSOLID    ((RtInt)30)	 /* Badly formed solid block */

#define RIE_BADTOKEN    ((RtInt)41)	 /* Invalid token for request */
#define RIE_RANGE       ((RtInt)42)	 /* Parameter out of range */
#define RIE_CONSISTENCY ((RtInt)43)	 /* Parameters inconsistent */
#define RIE_BADHANDLE   ((RtInt)44)	 /* Bad object/light handle */
#define RIE_NOSHADER    ((RtInt)45)	 /* Can't load requested shader */
#define RIE_MISSINGDATA ((RtInt)46)	 /* Required parameters not provided */
#define RIE_SYNTAX      ((RtInt)47)	 /* Declare type syntax error */
#define RIE_TOKENREDECLARED ((RtInt)48)	 /* Redeclaration of token */

#define RIE_MATH        ((RtInt)61)	 /* Zerodivide, noninvert matrix, etc. */


#define RIE_BADATTRIB	((RtInt)140)	/* Unsupported attribute */
#define RIE_BADOPTION	((RtInt)141)	/* Unsupported option */
#define RIE_SPACEREDECLARED ((RtInt)142)/* Redeclaration of space */
#define RIE_NODISPLAY   ((RtInt)143)	/* Can't load requested display server */
#define RIE_ERRRERTOOBID  ((RtInt)144)	/* Image resolution too big */
#define RIE_ERRBADSHADERPARAM ((RtInt145)           /* Parameter is not in shader param list */
#define RIE_ERRSHADERPARAMMISMATCH ((RtInt)146)     /* Value for parameter does not match declaration */
#define RIE_ERRBADARRAYACCESSINSHADER ((RtInt)147)  /* Invalid array index done in a shader */

/* ===========================================================================
    Error severity levels
   =========================================================================*/

#define RIE_INFO        ((RtInt)0) /* Rendering stats and other info */
#define RIE_WARNING     ((RtInt)1) /* Something seems wrong, maybe okay */
#define RIE_ERROR       ((RtInt)2) /* Problem. Results may be wrong */
#define RIE_SEVERE      ((RtInt)3) /* So bad you should probably abort */


#ifdef __cplusplus
}
#endif

#endif
