/******************************************************************************/
/*                                                                            */
/*    Copyright (c)The 3Delight Developers.                                   */
/*    All Rights Reserved.                                                    */
/*                                                                            */
/******************************************************************************/
 
// ===========================================================================
// = LIBRARY
//     shaderCompiler
// = FILENAME
//     shadeop.h
// = VERSION
//     $ $
// = DATE RELEASED
//     $Date: 2003/01/10 05:56:25 $
// = RCSID
//     $Id: shadeop.h,v 1.3 2003/01/10 05:56:25 aghiles Exp $
// ===========================================================================

#ifndef __shadeop_H
#define __shadeop_H

#include <string.h>
#include <stdlib.h>

#ifdef WIN32
#define DLEXPORT __declspec(dllexport)
#include <malloc.h>
#else
#define DLEXPORT
#endif

/****************************************************************************
 * How to use this file.
 *
 * This file must be included when we define a shadeop in a DSO.
 *
 * The macro SHADEOP_TABLE creates a table describing the functions 
 * defined in the DSO.  
 *
 * The macro SHADEOP_INIT creates the prototype of the initialization
 * function.
 *
 * The macro SHADEOP creates the prototype of the method function
 *
 * The macro SHADEOP_CLEANUP creates the prototype of the cleanup function.
 *
 *
 * See also end of file for an exemple.
 ****************************************************************************/


/****************************************************************************
  Struct SHADEOP_STRUCT
****************************************************************************/

typedef struct _SHADEOP_SPEC {
	char *declaration;          /* if zero length string, end of table signaled. */
	char *initfunctionname;     /* optional */
	char *cleanupfunctionname;  /* optional */
} SHADEOP_SPEC;


/****************************************************************************
   Macro SHADEOP_TABLE
****************************************************************************/

#define SHADEOP_TABLE(SHADEOP)   SHADEOP_SPEC DLEXPORT SHADEOP ## _shadeops[]


/****************************************************************************
   Macro SHADEOP_INIT
****************************************************************************/

#define SHADEOP_INIT(FUNC)       void DLEXPORT *FUNC(void)


/****************************************************************************
   Macro SHADEOP
****************************************************************************/

#define SHADEOP(FUNC)            int DLEXPORT FUNC(void *initdata, int argc, void **argv)


/****************************************************************************
   Macro SHADEOP_CLEANUP
****************************************************************************/

#define SHADEOP_CLEANUP(FUNC)    void DLEXPORT FUNC(void *initdata)


/****************************************************************************
   Macro ASSIGN_STRING

   For affectations to strings
*****************************************************************************/

#define ASSIGN_STRING(VAR, S)    { if ((VAR)) free((VAR));  (VAR) = strdup(S); }

/****************************************************************************
 * EXAMPLE FOR IMPLEMENTING SHADEOPS INSIDE C++ FILES
 *
 * When compiling a shader using the DSO:
 *    The DSO must be in one of the directorys used to search for include files
 *    (option -I of 'shaderdl')
 *
 * When rendering:
 *    The DSO must be in one of the directorys used to search for the shaders.
 ****************************************************************************/

/*
// #include "shadeop.h"
// #include <iostream.h>

// extern "C" {

//   SHADEOP_TABLE(sqr) =
//   {
// 		{"float sqr(float)", "sqr_init", "sqr_cleanup"},
// 		{"point sqr_p(point)", "sqr_init", "sqr_cleanup"},
// 		{""}
// 	 };
// }


// extern "C" SHADEOP_INIT(sqr_init)
// {
// 	  cout << "Init sqr..." << endl;
		
// 	  return 0x0;
// }


// extern "C" SHADEOP(sqr)
// {
// 	  float *result = (float *)argv[0];

// 	  float f = *((float*)argv[1]);

// 	  *result = f * f;

// 	  //cout << "In sqr..." << endl;

// 	  return 0;
// }

// extern "C" SHADEOP(sqr_p)
// {
// 	  float *result = (float *)argv[0];

// 	  float *f = ((float*)argv[1]);

// 	  result[0] = f[0] * f[0];
// 	  result[1] = f[1] * f[1];
// 	  result[2] = f[2] * f[2];

// 	  cout << "In sqr_p..." << endl;

// 	  return 0;
// }


// extern "C" SHADEOP(sqr_v)
// {
// 	  float *result = (float *)argv[0];

// 	  float *f = ((float*)argv[1]);

// 	  result[0] = f[0] * f[0];
// 	  result[1] = f[1] * f[1];
// 	  result[2] = f[2] * f[2];

// 	  cout << "In sqr_v..." << endl;

// 	  return 0;
// }


// extern "C" SHADEOP_CLEANUP(sqr_cleanup)
// {
// 	  cout << "Cleanup sqr..." << endl;
// }


*/

#endif

