/******************************************************************************/
/*                                                                            */
/*    Copyright (c)The 3Delight Developers.                                   */
/*    All Rights Reserved.                                                    */
/*                                                                            */
/******************************************************************************/

// ===========================================================================
// = LIBRARY
//     3DeLight
// = FILENAME
//     dlDSO.h
// = AUTHOR(S)
//     Patrick Fournier
// = VERSION
//     $Revision: 1.7 $
// = DATE RELEASED
//     $Date: 2001/08/22 17:52:56 $
// = RCSID
//     $Id: dlDSO.h,v 1.7 2001/08/22 17:52:56 aghiles Exp $
// = COPYRIGHT
//     Copyright  1999-2000 The 3Delight Developers.
//     All Rights Reserved.
// ===========================================================================
#ifndef __dlDSO_H
#define __dlDSO_H

#if defined(WIN32)

#include <windows.h>
#include <winbase.h>

/* SGI loading flags */
#define RTLD_LAZY 1
#define RTLD_NOW 2
#define RTLD_GLOBAL 4

typedef HINSTANCE DlDSO;

inline DlDSO DlOpenDSO(const char* i_fullName, int mode)
{
	return LoadLibrary(i_fullName);
}

inline void* DlGetDSOSymbol(DlDSO i_dso, const char* i_name)
{
	return GetProcAddress(i_dso, i_name);
}

inline void DlCloseDSO(DlDSO i_dso)
{
	FreeLibrary(i_dso);
}

inline char* DlDSOError(void)
{
	static char error[256];
	FormatMessage(FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_SYSTEM, 
				  0, GetLastError(), 0, (LPTSTR)error, 255, 0);
	return error;
}

#elif defined(DARWIN)

/* #include <CFBundle.h> */
#include <Carbon/Carbon.h>

typedef void* DlDSO;

/* SGI loading flags */
#define RTLD_LAZY 1
#define RTLD_NOW 2
#define RTLD_GLOBAL 4

inline DlDSO DlOpenDSO(const char* i_fullName, int mode)
{
	CFURLRef bundleURL;
	CFBundleRef myBundle;
	CFStringRef str;
        
	str = CFStringCreateWithCString(NULL, i_fullName, CFStringGetSystemEncoding());

	bundleURL = CFURLCreateWithFileSystemPath( kCFAllocatorDefault,
                                               str, 
                                               kCFURLPOSIXPathStyle, 
                                               true );
	
	myBundle = CFBundleCreate(kCFAllocatorDefault, bundleURL);

	assert(myBundle != 0);

	if( !myBundle )
	{
		return 0;
	}

	if( CFBundleLoadExecutable(myBundle) )
	{
		CFRelease(bundleURL);
		return myBundle;
	}
	else
	{
		CFRelease(bundleURL);
		return 0;
	}
}

inline void* DlGetDSOSymbol(DlDSO i_dso, const char* i_name)
{
	CFStringRef str;
    
	str = CFStringCreateWithCString( NULL, i_name, CFStringGetSystemEncoding() );
	return CFBundleGetFunctionPointerForName((CFBundleRef)i_dso, str);
}

inline void DlCloseDSO(DlDSO i_dso)
{
        CFBundleUnloadExecutable((CFBundleRef)i_dso);
        CFRelease((CFBundleRef)i_dso);
}

inline char* DlDSOError(void)
{
        return "";
}

#else

#include <dlfcn.h>

typedef void* DlDSO;

inline DlDSO DlOpenDSO(const char* i_fullName, int mode)
{
	return dlopen(i_fullName, mode);
}

inline void* DlGetDSOSymbol(DlDSO i_dso, const char* i_name)
{
	return dlsym(i_dso, i_name);
}

inline void DlCloseDSO(DlDSO i_dso)
{
#ifndef LINUX
	dlclose(i_dso);
#endif	
}

inline char* DlDSOError(void)
{
    return dlerror();
}

#endif

#endif // __dlDSO_H

