/******************************************************************************/
/*                                                                            */
/*    Copyright (c)The 3Delight Developers.                                   */
/*    All Rights Reserved.                                                    */
/*                                                                            */
/******************************************************************************/

// ===========================================================================
// = LIBRARY
//		 3delight
// = FILENAME
//		 dlSL.h
// = AUTHOR(S)
//		 Franck Diard, Patrick Fournier, Aghiles Kheffache
// = VERSION
//		 $Revision: 1.93 $
// = DATE RELEASED
//		 $Date: 2003/07/04 18:59:06 $
// = RCSID
//		 $Id: dlSL.h,v 1.93 2003/07/04 18:59:06 aghiles Exp $
// ===========================================================================

#ifndef __dlSL_H
#define __dlSL_H

#include "dlStopAndBool.h"
#include "dlTokenInfo.h"
#include "dlBasicTypes.h"
#include "dlInterface.h"

class DlAttributeSet;
class DlLightInput;
class DlShadingContext;
class DlShaderToken;
class DlRtModule;
class DlShaderSupport;
class DlFrame3D;
class DlTokenRegistry;
class DlSpaceRegistry;
class DlSearchPath;
class DlVar;
class DlRayInfos;
class DlToken;


////////////////////////////////////////////////////////////////////////////
// Spline types
////////////////////////////////////////////////////////////////////////////

enum DlSplineBasis
{
	e_CatmullRomBasis = 0,
	e_BezierBasis = 1,
	e_BSplineBasis = 2,
	e_HermiteBasis = 3,
	e_LinearBasis = 4
};


////////////////////////////////////////////////////////////////////////////
// Message passing shader request
////////////////////////////////////////////////////////////////////////////

enum DlMessageFromShader
{
	e_AtmosphereParam,
	e_DisplacementParam,
	e_LightsourceParam,
	e_SurfaceParam,
	e_IncidentParam,
	e_OppositeParam
};

extern "C" {

	////////////////////////////////////////////////////////////////////////////
	// Shading and lighting functions
	////////////////////////////////////////////////////////////////////////////

	DL_INTERFACE void DlSpecular(DlRGBColor          *o_color, 
								 const DlVector3     *i_normal, 
								 const DlVector3     *i_viewer, 
								 float                i_roughness, 
								 unsigned int         i_numLights, 
								 const DlLightInput  *i_lights, 
								 const DlStopAndBool *i_flag,
								 unsigned int         i_numMp,
								 unsigned int         i_mpIdx);

	DL_INTERFACE void DlSpecularstd(DlRGBColor          *o_color, 
									const DlVector3     *i_normal, 
									const DlVector3     *i_viewer, 
									float                i_roughness, 
									unsigned int         i_numLights, 
									const DlLightInput  *i_lights, 
									const DlStopAndBool *i_flag,
									unsigned int         i_numMp,
									unsigned int         i_mpIdx);

	DL_INTERFACE void DlSpecularbrdf(DlRGBColor           *o_color, 
									 const DlVector3     *i_light,
									 const DlVector3     *i_normal, 
									 const DlVector3     *i_viewer, 
									 float                i_roughness, 
									 const DlStopAndBool *i_flag,
									 unsigned int         i_numMp);

	DL_INTERFACE void DlDiffuse(DlRGBColor          *o_color, 
								const DlVector3     *i_normal, 
								unsigned int         i_numLights, 
								const DlLightInput  *i_lights, 
								const DlStopAndBool *i_flag,
								unsigned int         i_numMp,
								unsigned int         i_mpIdx);


	DL_INTERFACE void DlAmbient(DlRGBColor          *o_color, 
								unsigned int         i_numLights, 
								const DlLightInput  *i_lights,
								const DlStopAndBool *i_flag,
								unsigned int         i_numMp,
								unsigned int         i_mpIdx);

	DL_INTERFACE void DlPhong(DlRGBColor          *o_color, 
							  const DlVector3     *i_N,
							  const DlVector3     *i_V,
							  float                i_size,
							  unsigned int         i_numLights, 
							  const DlLightInput  *i_lights, 
							  const DlStopAndBool *i_flag,
							  unsigned int         i_numMp,
							  unsigned int         i_mpIdx);


	////////////////////////////////////////////////////////////////////////////
	// Test functions for illuminance and illuminate statements
	////////////////////////////////////////////////////////////////////////////


	DL_INTERFACE void DlIlluminate(DlStopAndBool         *o_test,
								   const DlVector3        i_appexPosition, 
								   const DlVector3       *i_axis, 
								   float                  i_angle, 
								   const DlVector3       *i_pos, 
								   DlVector3             *o_L,
								   const DlStopAndBool   *i_flag,
								   unsigned int           i_numMp,
								   bool                   i_shadows,
								   DlRayInfos       *i_rtInfo,
								   const DlShaderSupport *i_sp);

	DL_INTERFACE void DlSolar(DlStopAndBool         *o_res,
							  const DlVector3       *i_pos,
							  const DlVector3        i_Axis, 
							  float                  i_Angle, 
							  DlVector3             *i_L, 
							  const DlStopAndBool   *i_flag,
							  unsigned int           i_numMp,
							  bool                   i_shadows,
							  DlRayInfos       *i_info,
							  const DlShaderSupport *i_sp );

	DL_INTERFACE void DlIlluminance(DlStopAndBool       *o_test,
									const DlVector3     *i_pos, 
									const DlVector3     *i_axis,
									float                i_angle,
									const DlVector3     *i_L, 
									const DlStopAndBool *i_flag,
									unsigned int         i_numMp);


	////////////////////////////////////////////////////////////////////////////
	// Mathematical functions
	////////////////////////////////////////////////////////////////////////////

	DL_INTERFACE void DlSmoothStep(float               *o_res,
								   float                i_min, 
								   float                i_max, 
								   const float         *i_value,
								   const DlStopAndBool *i_flag,
								   unsigned int         i_numMp);

	DL_INTERFACE void DlSpline1D(float      *o_res, 
								 const void *i_args[],
								 int         i_nbArgs);


	DL_INTERFACE void DlSpline3D(DlVector3  *o_res, 
								 const void *i_args[],
								 int         i_nbArgs);


	DL_INTERFACE void DlSpline1DWithBasis(float         *o_res, 
										  DlSplineBasis  i_basis,
										  const void    *i_args[],
										  int            i_nbArgs);


	DL_INTERFACE void DlSpline3DWithBasis(DlVector3     *o_res, 
										  DlSplineBasis  i_basis,
										  const void    *i_args[],
										  int            i_nbArgs);


	DL_INTERFACE void DlRandom1D(float               *o_res,
								 const DlStopAndBool *i_flag,
								 unsigned int         i_numMp );

	DL_INTERFACE void DlRandom3D(DlVector3           *o_res,
								 const DlStopAndBool *i_flag,
								 unsigned int         i_numMp );

	DL_INTERFACE void DlFNoise1D(float               *o_res, 
								 float               *i_f, 
								 const DlStopAndBool *i_test, 
								 unsigned int         i_numMp);

	DL_INTERFACE void DlFNoise2D(float               *o_res, 
								 const float         *i_f1, 
								 const float         *i_f2, 
								 const DlStopAndBool *i_test, 
								 unsigned int         i_numMp);


	DL_INTERFACE void DlFNoise3D(float               *o_res,
								 const DlVector3     *i_p,
								 const DlStopAndBool *i_test, 
								 unsigned int         i_numMp);

	DL_INTERFACE void DlFNoise4D(float               *o_res,
								 const DlVector3     *i_p, 
								 float               *i_f,
								 const DlStopAndBool *i_test, 
								 unsigned int         i_numMp);

	DL_INTERFACE void DlPNoise1D(DlVector3           *o_res, 
								 float               *i_f,
								 const DlStopAndBool *i_test, 
								 unsigned int         i_numMp);

	DL_INTERFACE void DlPNoise2D(DlVector3           *o_res, 
								 float               *i_f1, 
								 float               *i_f2,
								 const DlStopAndBool *i_test, 
								 unsigned int         i_numMp);

	DL_INTERFACE void DlPNoise3D(DlVector3           *res, 
								 const DlVector3     *i_p,
								 const DlStopAndBool *i_test, 
								 unsigned int         i_numMp);

	DL_INTERFACE void DlPNoise4D(DlVector3           *res, 
								 const DlVector3     *i_p,
								 float               *i_f,
								 const DlStopAndBool *i_test, 
								 unsigned int         i_numMp);

	DL_INTERFACE void DlFCellnoise1D(float               *o_res, 
									 float               *i_f, 
									 const DlStopAndBool *i_test, 
									 unsigned int         i_numMp);

	DL_INTERFACE void DlFCellnoise2D(float               *o_res, 
									 const float         *i_f1, 
									 const float         *i_f2, 
									 const DlStopAndBool *i_test, 
									 unsigned int         i_numMp);


	DL_INTERFACE void DlFCellnoise3D(float               *o_res,
									 const DlVector3     *i_p,
									 const DlStopAndBool *i_test, 
									 unsigned int         i_numMp);

	DL_INTERFACE void DlFCellnoise4D(float               *o_res,
									 const DlVector3     *i_p, 
									 float               *i_f,
									 const DlStopAndBool *i_test, 
									 unsigned int         i_numMp);

	DL_INTERFACE void DlPCellnoise1D(DlVector3           *o_res, 
									 float               *i_f,
									 const DlStopAndBool *i_test, 
									 unsigned int         i_numMp);

	DL_INTERFACE void DlPCellnoise2D(DlVector3           *o_res, 
									 float               *i_f1, 
									 float               *i_f2,
									 const DlStopAndBool *i_test, 
									 unsigned int         i_numMp);

	DL_INTERFACE void DlPCellnoise3D(DlVector3           *res, 
									 const DlVector3     *i_p,
									 const DlStopAndBool *i_test, 
									 unsigned int         i_numMp);

	DL_INTERFACE void DlPCellnoise4D(DlVector3           *res, 
									 const DlVector3     *i_p,
									 float               *i_f,
									 const DlStopAndBool *i_test, 
									 unsigned int         i_numMp);


	DL_INTERFACE void DlFPnoise1D(float               *o_res, 
								  float               *i_f, 
								  float                i_period,
								  const DlStopAndBool *i_test, 
								  unsigned int        i_numMp);

	DL_INTERFACE void DlFPnoise2D(float               *o_res, 
								  const float         *i_f1, 
								  const float         *i_f2, 
								  float                i_uperiod,
								  float                i_vperiod,
								  const DlStopAndBool *i_test, 
								  unsigned int         i_numMp);


	DL_INTERFACE void DlFPnoise3D(float               *o_res,
								  const DlVector3     *i_p,
								  const DlVector3      i_period,
								  const DlStopAndBool *i_test, 
								  unsigned int         i_numMp);

	DL_INTERFACE void DlFPnoise4D(float               *o_res,
								  const DlVector3     *i_p, 
								  float               *i_f,
								  const DlVector3      i_pperiod,
								  float                i_tperiod,
								  const DlStopAndBool *i_test, 
								  unsigned int         i_numMp);

	DL_INTERFACE void DlPPnoise1D(DlVector3           *o_res, 
								  float               *i_f,
								  float                i_period,
								  const DlStopAndBool *i_test, 
								  unsigned int         i_numMp);

	DL_INTERFACE void DlPPnoise2D(DlVector3           *o_res, 
								  float               *i_f1, 
								  float               *i_f2,
								  float                i_uperiod,
								  float                i_vepriod,
								  const DlStopAndBool *i_test, 
								  unsigned int         i_numMp);

	DL_INTERFACE void DlPPnoise3D(DlVector3           *res, 
								  const DlVector3     *i_p,
								  const DlVector3      i_period,
								  const DlStopAndBool *i_test, 
								  unsigned int         i_numMp);

	DL_INTERFACE void DlPPnoise4D(DlVector3           *res, 
								  const DlVector3     *i_p,
								  float               *i_f,
								  const DlVector3      i_pperiod,
								  float                i_tperiod,
								  const DlStopAndBool *i_test, 
								  unsigned int         i_numMp);


	DL_INTERFACE void DlDu1D(float               *o_res,
							 const float         *i_value,
							 float                i_du,
							 unsigned int         i_nbMpU,
							 unsigned int         i_nbMpV,
							 const DlStopAndBool *i_flag);

	DL_INTERFACE void DlDu3D(DlVector3           *o_res,
							 const DlVector3     *i_value,
							 float                i_du,
							 unsigned int         i_nbMpU,
							 unsigned int         i_nbMpV,
							 const DlStopAndBool *i_flag);

	DL_INTERFACE void DlDv1D(float               *o_res,
							 const float         *i_value,
							 float                i_dv,
							 unsigned int         i_nbMpU,
							 unsigned int         i_nbMpV,
							 const DlStopAndBool *i_flag);

	DL_INTERFACE void DlDv3D(DlVector3           *o_res,
							 const DlVector3     *i_value,
							 float                i_dv,
							 unsigned int         i_nbMpU,
							 unsigned int         i_nbMpV,
							 const DlStopAndBool *i_flag);

	DL_INTERFACE void DlDeriv1D(float               *o_res,
								const float         *i_num,
								const float         *i_den,
								unsigned int         i_nbMpU,
								unsigned int         i_nbMpV,
								const DlStopAndBool *i_flag);

	DL_INTERFACE void DlDeriv3D(DlVector3           *o_res,
								const DlVector3     *i_num,
								const float         *i_den,
								unsigned int         i_nbMpU,
								unsigned int         i_nbMpV,
								const DlStopAndBool *i_flag);

	DL_INTERFACE void DlMin1D(float      *o_res,
							  const void *i_args[],
							  unsigned    i_nbArgs);

	DL_INTERFACE void DlMin3D(DlVector3  *o_res,
							  const void *i_args[],
							  unsigned    i_nbArgs);

	DL_INTERFACE void DlMax1D(float      *o_res,
							  const void *i_args[],
							  unsigned    i_nbArgs);

	DL_INTERFACE void DlMax3D(DlVector3  *o_res,
							  const void *i_args[],
							  unsigned    i_nbArgs);


	DL_INTERFACE void DlClamp3D(DlVector3           *o_res,
								const DlVector3     *i_p,
								const DlVector3      i_min,
								const DlVector3      i_max,
								const DlStopAndBool *i_flag,
								unsigned int         i_numMp);


	DL_INTERFACE void DlFilterstep(float               *o_res,
								   float               *i_edge,
								   float               *i_s1,
								   float               *i_s2,
								   float                i_width,
								   const char          *i_filter,
								   const DlStopAndBool *i_flag,
								   unsigned int         i_numMp);

	DL_INTERFACE void DlBoxFilterstep(float               *o_res,
									  float               *i_edge,
									  float               *i_s1,
									  float               *i_s2,
									  float                i_width,
									  const DlStopAndBool *i_flag,
									  unsigned int         i_numMp);

	DL_INTERFACE void DlTriangleFilterstep(float               *o_res,
										   float               *i_edge,
										   float               *i_s1,
										   float               *i_s2,
										   float                i_width,
										   const DlStopAndBool *i_flag,
										   unsigned int         i_numMp);

	DL_INTERFACE void DlCatmullRomFilterstep(float               *o_res,
											 float               *i_edge,
											 float               *i_s1,
											 float               *i_s2,
											 float                i_width,
											 const DlStopAndBool *i_flag,
											 unsigned int         i_numMp);

	DL_INTERFACE void DlGaussianFilterstep(float               *o_res,
										   float               *i_edge,
										   float               *i_s1,
										   float               *i_s2,
										   float                i_width,
										   const DlStopAndBool *i_flag,
										   unsigned int         i_numMp);



	////////////////////////////////////////////////////////////////////////////
	// Geometric functions
	////////////////////////////////////////////////////////////////////////////

	DL_INTERFACE void DlLength(
		float               *o_res,
		const DlVector3     *i_v,
		const DlStopAndBool *i_flag,
		unsigned int         i_numMp );

	DL_INTERFACE void DlDistance(
		float               *o_res,
		const DlVector3     *i_p1, 
		const DlVector3     *i_p2,
		const DlStopAndBool *i_flag,
		unsigned int         i_numMp);

	DL_INTERFACE void DlNormalize(
		DlVector3           *o_res, 
		const DlVector3     *i_v, 
		const DlStopAndBool *i_flag,
		unsigned int         i_numMp);

	DL_INTERFACE void DlFaceforward(
		DlVector3           *o_res, 
		const DlVector3     *i_N,
		const DlVector3     *i_I,
		const DlVector3     *i_Ng,
		const DlStopAndBool *i_flag,
		unsigned int         i_numMp);

	DL_INTERFACE void DlTransformSpace(
		DlVector3           *o_res,
		const DlVector3     *i_point,
		const DlFrame3D     *i_space,
		const DlMatrix44    *i_matrix,
		DlTokenInfo::DlType  i_which,
		bool                 i_fromCurrent,
		const DlStopAndBool *i_flag,
		unsigned i_numMp );


	DL_INTERFACE void DlReflect(
		DlVector3           *o_res, 
		const DlVector3     *i_I,
		const DlVector3     *i_N,
		const DlStopAndBool *i_flag,
		unsigned i_numMp );

	DL_INTERFACE void DlRefract(
		DlVector3 *o_res,
		const DlVector3 *i_I,
		const DlVector3 *i_N,
		float i_eta,
		const DlStopAndBool *i_flag,
		unsigned i_numMp );

	DL_INTERFACE void DlFresnel(
		const DlVector3     *i_I,
		const DlVector3     *i_N,
		float                i_eta,
		float               *o_Kr,
		float               *o_Kt,
		DlVector3           *o_R,
		DlVector3           *o_T,
		const DlStopAndBool *i_flag,
		unsigned i_numMp);

	DL_INTERFACE void DlArea(
		float               *o_res,
		const DlVector3     *i_P,
		unsigned int         i_nbMpU,
		unsigned int         i_nbMpV,
		const DlStopAndBool *i_flag );

	DL_INTERFACE void DlCalculateNormal(
		DlVector3 *o_res,
		const DlVector3 *i_P,
		unsigned i_nbMpU,
		unsigned i_nbMpV,
		const DlStopAndBool *i_flag,
		const DlShadingContext  &i_context );

	DL_INTERFACE void DlDepth(
		float *o_z, 
		const DlVector3 *i_p,
		const DlShadingContext &i_context,
		const DlStopAndBool *i_flag,
		unsigned i_numMp );

	DL_INTERFACE void DlTrace(
		DlRGBColor *o_res,
		float *o_dist,
		DlRayInfos *i_info,
		const DlVector3       *i_rayStart,
		const DlVector3       *i_rayDir,
		const DlVector3       *i_I,
		const DlVector3        i_E,
		const DlStopAndBool   *i_flag,
		unsigned i_numMpU, unsigned i_numMpV,
		const DlShaderSupport *i_sp );

	DL_INTERFACE void DlPtlined(
		float               *o_res,
		const DlVector3      i_p0,
		const DlVector3      i_p1,
		const DlVector3     *i_q,
		const DlStopAndBool *i_flag,
		unsigned i_numMp );

	DL_INTERFACE void DlRotatePoint(
		DlVector3           *o_res,
		const DlVector3     *i_q,
		float                i_angle,
		const DlVector3      i_p0,
		const DlVector3      i_p1,
		const DlStopAndBool *i_flag,
		unsigned i_numMp);


	////////////////////////////////////////////////////////////////////////////
	// Matrix functions
	////////////////////////////////////////////////////////////////////////////

	DL_INTERFACE const DlMatrix44 *DlGetToSpaceMatrix(const DlFrame3D *i_space);

	DL_INTERFACE void DlInverse(DlMatrix44       *o_result,
								const DlMatrix44 *i_matrix);

	DL_INTERFACE void DlDeterminant(float            *o_result,
									const DlMatrix44 *i_matrix);

	DL_INTERFACE void DlTranslate(DlMatrix44       *o_result,
								  const DlMatrix44 *i_matrix,
								  const DlVector3  *i_pt);

	DL_INTERFACE void DlRotate(DlMatrix44       *o_result,
							   const DlMatrix44 *i_matrix,
							   const float      *i_angle,
							   const DlVector3  *i_axis);

	DL_INTERFACE void DlScale(DlMatrix44       *o_result,
							  const DlMatrix44 *i_matrix,
							  const DlVector3  *i_point);

	////////////////////////////////////////////////////////////////////////////
	// Color functions
	////////////////////////////////////////////////////////////////////////////

	DL_INTERFACE void DlMix1D(float                *o_res, 
							  const float          *i_color0,
							  const float          *i_color1, 
							  float                 i_value,
							  const DlStopAndBool  *i_flag,
							  unsigned int          i_numMp);

	DL_INTERFACE void DlMix3D(DlRGBColor           *o_res, 
							  const DlRGBColor     *i_color0,
							  const DlRGBColor     *i_color1, 
							  float                 i_value,
							  const DlStopAndBool  *i_flag,
							  unsigned int          i_numMp);

	DL_INTERFACE void DlHSVToRGB(DlRGBColor          *o_rgb, 
								 const DlRGBColor    *i_c, 
								 const DlStopAndBool *i_flag,
								 unsigned int         i_numMp);
  
	DL_INTERFACE void DlRGBToHSV(DlRGBColor          *o_result, 
								 const DlRGBColor    *i_c, 
								 const DlStopAndBool *i_flag,
								 unsigned int         i_numMp);
  
	DL_INTERFACE void DlHSLToRGB(DlRGBColor          *o_rgb, 
								 const DlRGBColor    *i_c, 
								 const DlStopAndBool *i_flag,
								 unsigned int         i_numMp);
  
	DL_INTERFACE void DlRGBToHSL( DlRGBColor          *o_result, 
								  const DlRGBColor    *i_c,
								  const DlStopAndBool *i_flag,
								  unsigned int         i_numMp );

	DL_INTERFACE void DlYIQToRGB( DlRGBColor          *o_rgb, 
								  const DlRGBColor    *i_c, 
								  const DlStopAndBool *i_flag,
								  unsigned int         i_numMp );
  
	DL_INTERFACE void DlRGBToYIQ( DlRGBColor          *o_result, 
								  const DlRGBColor    *i_c, 
								  const DlStopAndBool *i_flag,
								  unsigned int         i_numMp );
  
	DL_INTERFACE void DlXYZToRGB( DlRGBColor          *o_rgb,
								  const DlRGBColor    *i_c,
								  const DlStopAndBool *i_flag,
								  unsigned int         i_numMp );
  
	DL_INTERFACE void DlRGBToXYZ( DlRGBColor          *o_result,
								  const DlRGBColor    *i_c,
								  const DlStopAndBool *i_flag,
								  unsigned int         i_numMp );
  
	DL_INTERFACE void DlTransformColor(
		DlRGBColor *o_res,
		const DlRGBColor *i_c,
		const char *i_space,
		bool i_fromRGB,
		const DlStopAndBool *i_flag,
		unsigned i_numMp );

	/* -------------------------------------------------- R A Y    T R A C I N G
	*/
	DL_INTERFACE void DlTransmission(
		DlRGBColor *o_res,
		const DlVector3 *i_p1,
		const DlVector3 *i_p2,
		const DlStopAndBool *i_flag,
		unsigned i_numMp, unsigned i_numMpV,
		DlRayInfos *i_rt_infos,
		const DlShaderSupport *i_sp );

	DL_INTERFACE void DlIsShadowRay(
		float *o_res, DlRayInfos *i_rt_infos );

	DL_INTERFACE void DlIsIndirectRay(
		float *o_res, DlRayInfos *i_rt_infos );

	DL_INTERFACE void DlRayLevel(
		float *o_res, DlRayInfos *i_rt_infos );

	DL_INTERFACE void DlRayHitTest(
		float *o_res,
		const DlVector3 *i_from,
		const DlVector3 *i_to,
		const DlStopAndBool *i_flag,
		unsigned i_numMpU, unsigned i_numMpV,
		DlRayInfos *i_rt_infos,
		const DlShaderSupport *i_sp );

	DL_INTERFACE void DlOcclusion(
		float *o_res,
		const DlVector3 *i_pos,
		const DlVector3 *i_r,
		float i_angle,
		DlVector3 *o_Nunocc, /* optional */
		float i_samples,
		float i_bias,
		float i_maxdist,
		float i_variance,
		DlVector3 *i_axis,
		const DlStopAndBool *i_flag,
		unsigned i_numMpU, unsigned i_numMpV,
		DlRayInfos *i_rt_infos,
		const DlShaderSupport *i_sp );

	DL_INTERFACE void DlIndirectDiffuse(
		DlRGBColor *o_res,
		const DlVector3 *i_pos,
		const DlVector3 *i_r,
		float i_angle,
		float i_samples,
		float i_bias,
		float  i_maxdist,
		float i_variance,
		const char *i_envmap, /* HDRI map */
		DlVector3 *i_axis,
		DlVector3 *o_Nunocc,
		float *o_occlusion,
		const DlStopAndBool *i_flag,
		unsigned i_numMpU, unsigned i_numMpV,
		DlRayInfos *i_rt_infos,
		const DlShaderSupport *i_sp );

	/* ------------------------------------------------------------- S T R I N G
	*/
	DL_INTERFACE void DlPrintf(const void  *i_args[],
							   unsigned     i_nbArgs);

	DL_INTERFACE void DlConcat(const char **o_result,
							   const void  *i_args[],
							   unsigned     i_nbArgs);

	DL_INTERFACE void DlFormat(const char **o_result,
							   const void  *i_args[],
							   unsigned     i_nbArgs);

	DL_INTERFACE void DlMatch(float      *o_res,
							  const char *i_pattern,
							  const char *i_subject);

	/* --------------------------------------------------------------- O T H E R
	*/
	DL_INTERFACE void DlPack(float               *o_res,
							 const float         *i_byte3,
							 const float         *i_byte2,
							 const float         *i_byte1,
							 const float         *i_byte0,
							 const DlStopAndBool *i_test,
							 unsigned int         i_numMp);

	DL_INTERFACE void DlUnpack(const float         *i_packbyte,
							   float               *o_byte3,
							   float               *o_byte2,
							   float               *o_byte1,
							   float               *o_byte0,
							   const DlStopAndBool *i_test,
							   unsigned int         i_numMp);


	/* -------------------------------------------T E X T U R E    M A P P I N G
	*/
	DL_INTERFACE void DlMakePointsST(
		const float *i_s, const float *i_t,
		float *o_s1, float *o_t1,
		float *o_s2, float *o_t2,
		float *o_s3, float *o_t3,
		float *o_s4, float *o_t4,
		unsigned numMpU, unsigned numMpV,
		const DlStopAndBool *i_flag );

	DL_INTERFACE void DlMakePoints(
		const DlVector3     *i_p0,
		DlVector3           *o_p1,
		DlVector3           *o_p2,
		DlVector3           *o_p3,
		DlVector3           *o_p4,
		unsigned int         numMpU,
		unsigned int         numMpV,
		const DlStopAndBool *i_flag);

	DL_INTERFACE void DlMakeFilterstepPoints(
		const float         *i_p0,
		float               *o_p1,
		unsigned int         numMpU,
		unsigned int         numMpV,
		const DlStopAndBool *i_flag );

	/*
		shadow(...) shadeop
	*/
	DL_INTERFACE void DlShadow(
		float                 *o_res,
		int                    i_dim,
		const char            *i_Name, 
		int                    i_channel, 
		const DlVector3       *i_P1,
		const DlVector3       *i_P2,
		const DlVector3       *i_P3,
		const DlVector3       *i_P4,
		float                  i_variance,
		int                    i_samples,
		float                 *i_sblur,
		float                 *i_tblur,
		float                  i_swidth,
		float                  i_twidth,
		const char            *i_filter,
		float                 *i_bias,
		const DlVector3       *i_lights,
		unsigned int           i_numMpU,
		unsigned int           i_numMpV,
		const DlStopAndBool   *i_flag,
		DlRayInfos       *i_rtInfos,
		const DlShaderSupport *i_sp);

	/*
		area shadow() shadeop
	*/
	DL_INTERFACE void DlAreaShadow(
		float                 *o_res,
		int                    i_dim,
		const char            *i_Name, 
		int                    i_channel, 
		const DlVector3       *i_P1,
		const DlVector3       *i_P2,
		const DlVector3       *i_P3,
		const DlVector3       *i_P4,
		float                  i_variance,
		int                    i_samples,
		float                 *i_sblur,
		float                 *i_tblur,
		float                  i_swidth,
		float                  i_twidth,
		const char            *i_filter,
		float                 *i_bias,
		const DlVector3       *i_lights,
		unsigned int           i_numMpU,
		unsigned int           i_numMpV,
		const DlStopAndBool   *i_flag,
		DlRayInfos       *i_rtInfos,
		const DlShaderSupport *i_sp,
		int                    i_numSourceP,
		const DlVector3        i_sP1,
		const DlVector3        i_sP2,
		const DlVector3        i_sP3,
		const DlVector3        i_sP4 );

	/*
		texture(...) shadeop
	*/
	DL_INTERFACE void DlTexture(
		float                 *o_res,
		int                    i_dim,
		const char            *i_Name, 
		int                    i_channel,
		const float           *i_s1,
		const float           *i_t1,
		const float           *i_s2,
		const float           *i_t2,
		const float           *i_s3,
		const float           *i_t3, 
		const float           *i_s4,
		const float           *i_t4,
		float                  i_variance,
		int                    i_samples,
		float                 *i_sblur,
		float                 *i_tblur,
		float                  i_swidth,
		float                  i_twidth,
		const char            *i_filter,
		float                  i_fill,
		float i_lerp,
		unsigned int           i_numMpU,
		unsigned int           i_numMpV,
		const DlStopAndBool   *i_flag,
		const DlShaderSupport *i_sp );

	/*
		environment(...) shadeop
	*/
	DL_INTERFACE void DlEnvironment(
		float                 *o_res,
		int                    i_dim,
		const char            *i_Name,
		int                    i_channel,
		const DlVector3       *i_direction1,
		const DlVector3       *i_direction2,
		const DlVector3       *i_direction3,
		const DlVector3       *i_direction4,
		float                  i_variance,
		int                    i_samples,
		float                 *i_sblur,
		float                 *i_tblur,
		float                  i_swidth,
		float                  i_twidth,
		const char            *i_filter,
		float                  i_fill,
		float i_lerp,
		unsigned int           i_numMpU,
		unsigned int           i_numMpV,
		const DlStopAndBool   *i_flag,
		const DlShaderSupport *i_sp);

	/*
		textureinfo(...) shadeop
	*/
	DL_INTERFACE void DlTextureInfo(
		float *o_res,
		const char          *i_name,
		const char          *i_data,
		void                *o_info,
		int                  i_isVarying,
		DlTokenInfo::DlType  i_type,
		int                  i_nbElems,
		const DlStopAndBool *i_flag,
		unsigned int         i_numMp,
		const DlShaderSupport *i_sp );


	////////////////////////////////////////////////////////////////////////////
	// Message passing and informations functions
	////////////////////////////////////////////////////////////////////////////

	DL_INTERFACE void DlAttribute(float                  *o_res,
								  const char             *i_name,
								  void                   *o_info,
								  int                     i_isVarying,
								  DlTokenInfo::DlType     i_type,
								  int                     i_nbElems,
								  const DlStopAndBool    *i_flag,
								  unsigned int            i_numMp, 
								  const DlShadingContext &i_context );



	DL_INTERFACE void DlOption(float                  *o_res,
							   const char             *i_name,
							   void                   *o_info,
							   int                     i_isVarying,
							   DlTokenInfo::DlType     i_type,
							   int                     i_nbElems,
							   const DlStopAndBool    *i_flag,
							   unsigned int            i_numMp, 
							   const DlShadingContext &i_context );


	DL_INTERFACE void DlRendererInfo(float               *o_res,
									 const char          *i_name,
									 void                *o_info,
									 int                  i_isVarying,
									 DlTokenInfo::DlType  i_type,
									 int                  i_nbElems,
									 const DlStopAndBool *i_flag,
									 unsigned int         i_numMp);


	DL_INTERFACE void DlMessageParam(float                  *o_res,
									 DlMessageFromShader     i_shader,
									 const char             *i_name,
									 void                   *o_info,
									 int                     i_isVarying,
									 DlTokenInfo::DlType     i_type,
									 int                     i_nbElems,
									 const DlStopAndBool    *i_flag,
									 unsigned int            i_numMp,
									 const DlShadingContext &i_context,
									 unsigned int            i_noLight);


	////////////////////////////////////////////////////////////////////////////
	// Light evaluation
	////////////////////////////////////////////////////////////////////////////

	DL_INTERFACE void DlEvalLightShaders(
		unsigned int           numMpU,
		unsigned int           numMpV,
		const DlVector3       *i_surfacePts,
		DlLightInput          *,
		unsigned               numLights,
		const DlShadingContext &,
		const DlShaderSupport * );


	////////////////////////////////////////////////////////////////////////////
	// Utility functions for shaders
	////////////////////////////////////////////////////////////////////////////

	DL_INTERFACE DlToken *DlFindToken(
		DlTokenRegistry *, const char *i_name );

	DL_INTERFACE DlShaderToken* DlCreateShaderParamsDescr(unsigned int i_nb);

	DL_INTERFACE void DlFreeShaderParamsDescr(DlShaderToken *i_st);

	DL_INTERFACE DlVar* DlCreateVarsInit(DlShaderToken *i_st, int i_nb);
	
	DL_INTERFACE void DlFreeVarsInit(DlVar *i_varsInit);

	DL_INTERFACE void DlInitVarValue(DlVar                &i_var,
									 const char           *i_space,
									 DlTokenInfo::DlClass  i_class,
									 unsigned int          i_numValues,
									 void                 *i_value);

	DL_INTERFACE const DlFrame3D** DlCreateSpaceList(unsigned int i_nbSpaces);

	DL_INTERFACE void DlDeleteSpaceList(const DlFrame3D **i_spaceList);

	DL_INTERFACE const DlFrame3D* FindSpace(const DlSpaceRegistry *i_registry, 
											const char            *i_name);

	DL_INTERFACE const char* FindDSO(const DlSearchPath *i_searchPath, 
									 const char         *i_fileName);

	DL_INTERFACE void ReportBadArrayIndexError(const char *i_dsoFile, 
											   int         i_noLineDso,
											   const char *i_shaderFile,
											   int         i_noLineShader,
											   const char *i_arrayVarName,
											   float       i_index);

	DL_INTERFACE void* DlCalloc(unsigned int i_nbElems, unsigned int i_size);

	DL_INTERFACE void DlFree(void *i_block);

	DL_INTERFACE const char* DlStrdup(const char* i_str);

	DL_INTERFACE void DlFreeStr(const char *i_str);

	DL_INTERFACE bool DlShapeIsLeftHanded(const DlShadingContext &i_context);
}

#endif // __dlSL_H
