/******************************************************************************/
/*                                                                            */
/*    Copyright (c) Centre de traitement en imagerie virtuelle inc.           */
/*    All Rights Reserved.                                                    */
/*                                                                            */
/*    These coded instructions, statements, and computer programs  contain    */
/*    unpublished  proprietary and  confidential information of  Centre de    */
/*    traitement  en  imagerie virtuelle inc. (CTIV), and are protected by    */
/*    Federal copyright law. They may not be disclosed to third parties or    */
/*    or duplicated in any form, in whole or in  part, without  the  prior    */
/*    written consent of CTIV.                                                */
/*                                                                            */
/******************************************************************************/

// ===========================================================================
// = LIBRARY
//     3delight
// = FILENAME
//     dlSearchPath.h
// = AUTHOR(S)
//     Patrick Fournier
// = VERSION
//     $Revision: 1.2 $
// = DATE RELEASED
//     $Date: 2000/06/29 20:08:49 $
// = RCSID
//     $Id: dlSearchPath.h,v 1.2 2000/06/29 20:08:49 ours Exp $
// = COPYRIGHT
//     Copyright  1999-2000 Centre de traitement en imagerie virtuelle inc.
//     All Rights Reserved.
// ===========================================================================

#ifndef __dlSearchPath_H
#define __dlSearchPath_H

#ifdef WIN32
#include "3DeLightSTL.h"
#else
#include <vector.h>
#endif

#ifdef WIN32
#define DL_EXPORT __declspec(dllexport)
#else
#define DL_EXPORT
#endif


class DlSearchPath
{
public:
	typedef enum {
		e_shader,
		e_texture,
		e_display,
		e_archive,
		e_numCategory
	} DlCategory;

	typedef enum
	{
		e_execPerm = 1,
		e_writePerm = 2,
		e_readPerm = 4
	} DlFileAccessPerm;

	DlSearchPath(void);
	~DlSearchPath(void);

	void SetPath(DlCategory i_category, const char* i_pathList,
                 bool eraseOldPaths = true);
	void DeleteCategory( DlCategory i_cat );

	DL_EXPORT const char* Find(DlCategory   i_category, 
							   const char*  i_fileName, 
							   unsigned int i_mode = e_readPerm) const;

	static void Expand(char* outPath, const char* inPath);

protected:
	vector<char*, malloc_alloc> m_paths[e_numCategory];

	//vector<char*> m_paths[e_numCategory];

	static const char* k_pathVarName[e_numCategory];

	static const char *k_directorySeparatorStr;

	static const char k_pathSeparatorChr;
	static const char k_directorySeparatorChr;

	static const char k_envVarBegin;

	static const int k_maxPathLen;

    static const char* k_defaultPaths[e_numCategory];
    
    static void SubstituteEnvVar(char* out, const char* in);
	static void ReplaceDotDot(char* out, const char* in);
};

#endif				 // __dlSearchPath_H
