/******************************************************************************/
/*                                                                            */
/*    Copyright (c)The 3Delight Developers.                                   */
/*    All Rights Reserved.                                                    */
/*                                                                            */
/******************************************************************************/

// ===========================================================================
// = LIBRARY
//     3delight
// = FILENAME
//     dlShadingContext.h
// = AUTHOR(S)
//     Patrick Fournier
// = VERSION
//     $Revision: 1.20 $
// = DATE RELEASED
//     $Date: 2003/06/16 18:34:40 $
// = RCSID
//     $Id: dlShadingContext.h,v 1.20 2003/06/16 18:34:40 aghiles Exp $
// ===========================================================================

#ifndef __dlShadingContext_H
#define __dlShadingContext_H

#include "dlBasicTypes.h"
#include "dlInterface.h"

class DlRIOptions;
class DlShaderSupport;
class DlAttributeSet;
class DlSceneObject;
class DlSpaceRegistry;
class DlFrame3D;
class DlAllVisibleParams;
class DlRayInfos;

class DlShadingContext
{
public:
    DlShadingContext(
		const DlRIOptions* i_options,
		DlShaderSupport* i_sl,
		const DlSpaceRegistry* i_sr,
		DlFrame3D const**i_spaceList,
		const DlSceneObject *,
		DlAllVisibleParams *i_allParams );

    DlShadingContext(
		const DlRIOptions* i_options, DlShaderSupport* i_sl,
		const DlSpaceRegistry* i_sr, DlFrame3D const**i_spaceList,
		const DlSceneObject *i_sceneObject, const DlRayInfos *i_rtInfos,
		const DlVector3& i_eye, DlAllVisibleParams *i_allParams );
    
	DL_INTERFACE const DlRIOptions* Options(void) const
	{
		return m_options;
	}
    
	DL_INTERFACE DlShaderSupport* ShaderSupport(void) const
	{
		return m_sl;
	}
    
	DL_INTERFACE const DlSpaceRegistry* SpaceRegistry(void) const
	{
		return m_spaceRegistry;
	}

	DL_INTERFACE DlFrame3D const* const* SpaceList(void) const
	{
		return m_spaceList;
	}

	DL_INTERFACE const DlSceneObject* SceneObject(void) const
	{
		return m_sceneObject;
	}

	DL_INTERFACE const DlAttributeSet* AttributeSet( void ) const;

	DL_INTERFACE DlRayInfos* RayInfos(void) const
	{
		return (DlRayInfos *)m_rtInfos;
	}

	DL_INTERFACE const DlVector3* Eye(void) const { return &m_eye; }

    DL_INTERFACE bool ShapeIsLeftHanded(void) const;

	DlAllVisibleParams *AllVisibleParams(void) { return m_allParams; }

protected:
    DlShaderSupport *m_sl;
	const DlSpaceRegistry *m_spaceRegistry;
    
	DlFrame3D const**m_spaceList;

	const DlRIOptions *m_options;

	const DlSceneObject *m_sceneObject;
    
	const DlRayInfos *m_rtInfos;
    
    DlVector3 m_eye;

	DlAllVisibleParams *m_allParams;
};

#endif
