/******************************************************************************/
/*                                                                            */
/*    Copyright (c) Centre de traitement en imagerie virtuelle inc.           */
/*    All Rights Reserved.                                                    */
/*                                                                            */
/*    These coded instructions, statements, and computer programs  contain    */
/*    unpublished  proprietary and  confidential information of  Centre de    */
/*    traitement  en  imagerie virtuelle inc. (CTIV), and are protected by    */
/*    Federal copyright law. They may not be disclosed to third parties or    */
/*    or duplicated in any form, in whole or in  part, without  the  prior    */
/*    written consent of CTIV.                                                */
/*                                                                            */
/******************************************************************************/

// ===========================================================================
// = LIBRARY
//     3delight
// = FILENAME
//     dlSpaceRegistry.h
// = AUTHOR(S)
//     Ugo Tremblay
// = VERSION
//     $Revision: 1.1 $
// = DATE RELEASED
//     $Date: 2000/06/21 18:46:36 $
// = RCSID
//     $Id: dlSpaceRegistry.h,v 1.1 2000/06/21 18:46:36 ours Exp $
// = COPYRIGHT
//     Copyright  1999-2000 Centre de traitement en imagerie virtuelle inc.
//     All Rights Reserved.
// ===========================================================================

#ifndef __dlSpaceRegistry_H
#define __dlSpaceRegistry_H

#ifdef WIN32
#include "3DeLightSTL.h"
#else
#include <hash_map.h>
#ifndef __declspec
#define __declspec(x)
#endif
#endif

#include <string.h>


// ===========================================================================
//      C L A S S E S
// ===========================================================================
class DlFrame3D;

// ===========================================================================
//
//      CLASS DlSpaceRegistry
//
// ===========================================================================

/**
  DlSpaceRegistry.

  \begin{description}

  \item[CLASS TYPE]
  --

  \item[DESCRIPTION]
  This class holds a registry of all named spaces (coordinate systems).

  \item[NOTES]
  --

  \end{description}

  @author Ugo Tremblay
  @see --
  */
class DlSpaceRegistry
{
public:
	typedef enum
	{
		e_noErr = 0,
		e_declared,
		e_notDeclared
	} DlErrorCode;

	DlSpaceRegistry();
	~DlSpaceRegistry();

	DlErrorCode RegisterSpace( const char* i_key,
	                           const DlFrame3D& i_space );

	__declspec(dllexport) DlFrame3D* FindSpace( const char* i_name ) const;

protected:
	struct EqKey
	{
		bool operator()(const char* t1, const char* t2) const
		{
			return (strcmp(t1,t2) == 0);
		}
	};

	typedef hash_map< const char*, DlFrame3D*, hash<const char*>, EqKey> DlHashMap;
	DlHashMap m_registry;
};

#endif // __dlSpaceRegistry_H
