/******************************************************************************/
/*                                                                            */
/*    Copyright (c) Centre de traitement en imagerie virtuelle inc.           */
/*    All Rights Reserved.                                                    */
/*                                                                            */
/*    These coded instructions, statements, and computer programs  contain    */
/*    unpublished  proprietary and  confidential information of  Centre de    */
/*    traitement  en  imagerie virtuelle inc. (CTIV), and are protected by    */
/*    Federal copyright law. They may not be disclosed to third parties or    */
/*    or duplicated in any form, in whole or in  part, without  the  prior    */
/*    written consent of CTIV.                                                */
/*                                                                            */
/******************************************************************************/

// ===========================================================================
// = LIBRARY
//     3delight
// = FILENAME
//     dlVector3.h
// = AUTHOR(S)
//     Patrick Fournier
// = VERSION
//     $Revision: 1.1 $
// = DATE RELEASED
//     $Date: 2000/06/21 18:46:36 $
// = RCSID
//     $Id: dlVector3.h,v 1.1 2000/06/21 18:46:36 ours Exp $
// = COPYRIGHT
//     Copyright  1999-2000 Centre de traitement en imagerie virtuelle inc.
//     All Rights Reserved.
// ===========================================================================

#ifndef __dlVector3_H
#define __dlVector3_H

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <assert.h>

typedef float DlVector2[2];
typedef float DlVector3[3];
typedef float DlVector4[4];

typedef double DlVector3Double[3];
typedef double DlVector4Double[4];

typedef float DlVectorUV[2];


extern const DlVector3 k_DlVector3Null;


#define VECTOR2_DIST(a, b) float(sqrt( ((a)[0]-(b)[0])*((a)[0]-(b)[0]) + ((a)[1]-(b)[1])*((a)[1]-(b)[1])) )

// #define VECTOR2_DIST2(a, b, c, d) ((c-a)*(c-a) + (d-b)*(d-b))

#define VECTOR_COPY(V1, V2) \
    { \
        (V1)[0] = (V2)[0]; \
        (V1)[1] = (V2)[1]; \
        (V1)[2] = (V2)[2]; \
    }


#define VECTOR4_COPY(V1, V2) \
    { \
        (V1)[0] = (V2)[0]; \
        (V1)[1] = (V2)[1]; \
        (V1)[2] = (V2)[2]; \
        (V1)[3] = (V2)[3]; \
    }


/**
   Set the values of a vector.

   @param DlVector3* V: Vector to set
   @param DlReal X, Y, Z: Values to give to the vector
*/
#define VECTOR_SET(V, X, Y, Z) \
	{ \
		(V)[0] = X;\
		(V)[1] = Y;\
		(V)[2] = Z;\
	}

/**
  Set a vector to zero

  @param DlVector3* V : vector to set to 0
 */
#define VECTOR_SET_ZERO( V ) \
	{ \
		(V)[0] = 0.0;\
		(V)[1] = 0.0;\
		(V)[2] = 0.0;\
	}

/**
   negate a vector of a vector.

*/
#define VECTOR_OPPOSE(V, S) \
	{ \
		(V)[0] = -(S)[0];\
		(V)[1] = -(S)[1];\
		(V)[2] = -(S)[2];\
	}

/**
   negate a vector of a vector.
*/
#define VECTOR_OPPOSE_IN_PLACE(V) \
	{ \
		(V)[0] = -(V)[0];\
		(V)[1] = -(V)[1];\
		(V)[2] = -(V)[2];\
	}

/**
   Add two vectors and put the result in a third one.

   V = S1 + S2

   @param DlVector3* V.
   @param DlVector3* S1.
   @param DlVector3* S2.
*/
#define VECTOR_ADD(V, S1, S2) \
	{ \
		(V)[0] = (S1)[0] + (S2)[0];\
		(V)[1] = (S1)[1] + (S2)[1];\
		(V)[2] = (S1)[2] + (S2)[2];\
	}

/**
   Add a vector to another.

   V += S1

   @param DlVector3* V.
   @param DlVector3* S1.
*/
#define VECTOR_ADD_IN_PLACE(V, S1) \
	{ \
		(V)[0] += (S1)[0];\
		(V)[1] += (S1)[1];\
		(V)[2] += (S1)[2];\
	}

/**
   Substract two vectors and put the result in a third one.

   V = S1 - S2

   @param DlVector3* V.
   @param DlVector3* S1.
   @param DlVector3* S2.
*/
#define VECTOR_SUB(V, S1, S2) \
	{ \
		(V)[0] = (S1)[0] - (S2)[0];\
		(V)[1] = (S1)[1] - (S2)[1];\
		(V)[2] = (S1)[2] - (S2)[2];\
	}

/**
   constriuct a vector from two vectors and put the result in a third one.

   V = S2 - S1

   @param DlVector3* V.
   @param DlVector3* S1.
   @param DlVector3* S2.
*/
#define VECTOR_MAKE(V, S1, S2) \
	{ \
		(V)[0] = (S2)[0] - (S1)[0];\
		(V)[1] = (S2)[1] - (S1)[1];\
		(V)[2] = (S2)[2] - (S1)[2];\
	}

#define VECTOR_MAKE_INV(S1, S2, V) VECTOR_MAKE(V, S1, S2)

/**
   Substract a vector from another.

   V -= S1

   @param DlVector3* V.
   @param DlVector3* S1.
*/
#define VECTOR_SUB_IN_PLACE(V, S1) \
	{ \
		(V)[0] -= (S1)[0];\
		(V)[1] -= (S1)[1];\
		(V)[2] -= (S1)[2];\
	}

/**
   Multiply a vector by a scalar and put the result in another vector.

   V = S1 * F

   @param DlVector3* V.
   @param DlReal F.
   @param DlVector3* S1.
*/
#define VECTOR_SCALE(V, S1, F) \
	{ \
		(V)[0] = (F) * (S1)[0];\
		(V)[1] = (F) * (S1)[1];\
		(V)[2] = (F) * (S1)[2];\
	}

/**
*/
#define VECTOR_MULT( V, S1, S2 ) \
	{ \
		(V)[0] = (S1)[0] * (S2)[0]; \
		(V)[1] = (S1)[1] * (S2)[1]; \
		(V)[2] = (S1)[2] * (S2)[2]; \
	}
	
/**
   Multiply a vector by a scalar.

   V *= F

   @param DlVector3* V.
   @param DlReal F.
*/
#define VECTOR_SCALE_IN_PLACE(V, F) \
	{ \
		(V)[0] *= (F);\
		(V)[1] *= (F);\
		(V)[2] *= (F);\
	}

/**
   Multiply a vector by a scalar and add it to a point.

   V = I*D+P

   @param DlVector3* V.
   @param DlVector3* D.
   @param DlVector3* P.
   @param DlReal F.
*/
#define VECTOR_SCALE_ADD(V, F , D, P) \
	{ \
		(V)[0] = (F)* (D)[0] + (P)[0] ;\
		(V)[1] = (F)* (D)[1] + (P)[1] ;\
		(V)[2] = (F)* (D)[2] + (P)[2] ;\
	}

/**
   Divide a vector by a scalar and put the result in another vector.

   V = S1 / F

   @param DlVector3* V.
   @param DlReal F.
   @param DlVector3* S1.
*/
#define VECTOR_DIV(V, S1, F) \
	{ \
		assert(F != 0.);\
		VECTOR_SCALE((V), (S1), 1./(F));\
	}

/**
   Divide a vector by a scalar.

   V /= F

   @param DlVector3* V.
   @param DlReal F.
*/
#define VECTOR_DIV_IN_PLACE(V, F) VECTOR_SCALE_IN_PLACE((V), 1.0f/(F))

/**
   Compute the dot product of two vectors.

   D = V1 dot V2

   @param DlVector3* V1.
   @param DlVector3* V2.
   @return DlReal the dot product of V1 and V2
*/
#define VECTOR_DOT(V1, V2) \
	(\
	(V1)[0] * (V2)[0] + (V1)[1] * (V2)[1] + (V1)[2] * (V2)[2]\
	)

/**
  Manhattan distance between 2 points
*/
#define VECTOR_MANHATTAN(V1, V2) \
    fabs( ((V1)[0]-(V2)[0]) + fabs((V1)[1]-(V2)[1]) + fabs((V1)[2]-(V2)[2]) )

/**
  Distance between 2 points
*/
#define VECTOR_DISTANCE(V1, V2) \
	(\
	sqrt ( ((V1)[0]-(V2)[0])*((V1)[0]-(V2)[0]) \
	+ ((V1)[1]-(V2)[1])*((V1)[1]-(V2)[1])\
	+ ((V1)[2]-(V2)[2])*((V1)[2]-(V2)[2]) )\
	)

#define VECTOR_DIST2(V1, V2) \
	(\
	( ((V1)[0]-(V2)[0])*((V1)[0]-(V2)[0]) \
	+ ((V1)[1]-(V2)[1])*((V1)[1]-(V2)[1]) \
	+ ((V1)[2]-(V2)[2])*((V1)[2]-(V2)[2]) )\
	)

/**
   Compute the cross product of two vectors.

   V = V1 cross V2

   @param DlVector3* V1.
   @param DlVector3* V2.
   @return DlReal the cross product of V1 and V2
*/
#define VECTOR_CROSS(V, V1, V2) \
	{ \
		(V)[0] = (V1)[1] * (V2)[2] - (V1)[2] * (V2)[1]; \
		(V)[1] = (V1)[2] * (V2)[0] - (V1)[0] * (V2)[2]; \
		(V)[2] = (V1)[0] * (V2)[1] - (V1)[1] * (V2)[0]; \
	}

#define VECTOR_CROSS_INV(V1, V2, V)  VECTOR_CROSS(V, V1, V2)

/**
   Compute the squared norm of a vector.

   N2 = ||V|| * ||V||

   @param DlVector3* V.
   @return DlReal the squared norm of V
*/
#define VECTOR_NORM2(V) \
	(\
	VECTOR_DOT((V), (V))\
	)

/**
   Compute the norm of a vector.

   N = ||V||

   @param DlVector3* V.
   @return DlReal the norm of V
*/
#define VECTOR_NORM(V) \
	(\
	sqrtf(VECTOR_NORM2((V)))\
	)

/**
   Normalize a vector.

   V = V / ||V||

   @param DlVector3* V.
*/
#define VECTOR_NORMALIZE_IN_PLACE(V) \
	{ \
		float n_=(float)sqrt((V)[0]*(V)[0] + (V)[1]*(V)[1] +(V)[2]*(V)[2]);\
		if ( n_ > 0.0) \
		    VECTOR_DIV_IN_PLACE(V, n_);\
	}

#define VECTOR_NORMALIZE(D ,S) \
	{ \
		VECTOR_COPY((D),(S));\
		VECTOR_NORMALIZE_IN_PLACE(D);\
	}

#define VECTOR_NORMALIZE_CHECK_ZERO(D ,S) \
	{ \
		VECTOR_COPY((D),(S));\
        if (VECTOR_IS_NOT_ZERO(S)) \
        { \
		    VECTOR_NORMALIZE_IN_PLACE(D);\
        } \
	}

#define VECTOR_PRINT(S, Comment) \
	{ \
		printf("\n \t VECTOR3: %s , %5f %5f %5f", Comment, (S)[0], (S)[1], (S)[2]);\
	}
#define VECTOR_IS_EPSILON(V, E) ( fabs((V)[0])<E &&  fabs((V)[1])<E && fabs((V)[2])<E )

#define VECTOR_IS_ZERO(S) ( ((S)[0]  == 0.0) &&  ((S)[1]  == 0.0) && ((S)[2]  == 0.0))
#define VECTOR_IS_NOT_ZERO(S) ( ((S)[0]  != 0.0) ||  ((S)[1]  != 0.0) || ((S)[2]  != 0.0))

#define VECTOR_DIFFER(S1, S2) ( ((S1)[0]  != (S2)[0]) ||  ((S1)[1]  != (S2)[1]) || ((S1)[2]  != (S2)[2]))

#define VECTOR_EQUAL(S1, S2) ( ((S1)[0]  == (S2)[0]) &&  ((S1)[1]  == (S2)[1]) && ((S1)[2]  == (S2)[2]))


#endif				 // __dlVector3_H
