/******************************************************************************/
/*                                                                            */
/*    Copyright (c)The 3Delight Developers.                                   */
/*    All Rights Reserved.                                                    */
/*                                                                            */
/******************************************************************************/

// ===========================================================================
// = LIBRARY
//     shaderCompiler
// = FILENAME
//     slcShaderCCodeOp.h
// = AUTHOR(S)
//     Marco Jacques
// = VERSION
//     $Revision: 1.99 $
// = DATE RELEASED
//     $Date: 2003/07/07 17:54:56 $
// = RCSID
//     $Id: slcShaderCCodeOp.h,v 1.99 2003/07/07 17:54:56 aghiles Exp $
// ===========================================================================

////////////////////////////////////////////////////////////////////////////
// This file is used only by the generated C file of the shader, not by the
// compiler itself
////////////////////////////////////////////////////////////////////////////

#ifndef __slcShaderCCodeOp_H
#define __slcShaderCCodeOp_H

#include <float.h>

////////////////////////////////////////////////////////////////////////////
// Dll initialization function needed for cygwin
////////////////////////////////////////////////////////////////////////////

#if defined(DL_COMPILING_SHADER) && defined(CYGWIN32)

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#undef WIN32_LEAN_AND_MEAN
#include <stdio.h>

extern "C" {
	BOOL APIENTRY
	DllMainCRTStartup (
		HANDLE hInst /* Library instance handle. */ ,
		DWORD reason /* Reason this function is being called. */ ,
		void *reserved /* Not used. */ );
	
	// defined in minwg32 library...
	//
	BOOL _CRT_INIT (
		HANDLE hInst /* Library instance handle. */ ,
		DWORD reason /* Reason this function is being called. */ ,
		void *reserved /* Not used. */ );
}

// var needed by the cygwin libraries...
//
struct _reent *_impure_ptr = 0;


// Enter point of the dll needed by Visual-C++...
//
BOOL APIENTRY DllMainCRTStartup (
	 HANDLE hInst /* Library instance handle. */ ,
	 DWORD reason /* Reason this function is being called. */ ,
	 void *reserved /* Not used. */ )
{

	_CRT_INIT(hInst, reason, reserved);

	// I'm always returning true here.  CRT_INIT seems to return
	// 0, which something has gone wrong.  But, after testing
	// everything seems to be fine...
	//
	return 1;
}


#endif


#include "dlSL.h"
#include "dlBasicTypes.h"
#include "dlShaderInterface.h"
#include "dlDSO.h"
#include "dlVar.h"
#include "dlLightHidden.h"
#include "dlLightInput.h"
#include "dlShadingContext.h"
#include "dlSurfaceHidden.h"
#include "dlImagerHidden.h"
#include "dlInterface.h"

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#ifdef WIN32
#define DL_SHADER_FUNC __declspec(dllexport)
#else
#define DL_SHADER_FUNC
#endif

////////////////////////////////////////////////////////////////////////////
// EXTERN VARIABLES OF THE 3DELIGHT LIBRARY
////////////////////////////////////////////////////////////////////////////


extern "C" {
	// NOTE: if you add tokens here, rememeber to put a __declspec(dllexport)
	// to the token declaration in dlRtToken.h
	DL_INTERFACE extern char* RI_P;
	DL_INTERFACE extern char* RI_N;
	DL_INTERFACE extern char* RI_CS;
	DL_INTERFACE extern char* RI_OS;
	DL_INTERFACE extern char* RI_S;
	DL_INTERFACE extern char* RI_T;
	DL_INTERFACE extern char* RI_PS;
	DL_INTERFACE extern char* RI_E;
	DL_INTERFACE extern char* RI_CI;
	DL_INTERFACE extern char* RI_OI;
	DL_INTERFACE extern char* RI_ALPHA;
	DL_INTERFACE extern char* RI_NCOMPS;
	DL_INTERFACE extern char* RI_TIME;
	DL_INTERFACE extern char* RI_L;
	DL_INTERFACE extern char* RI_CL;
	DL_INTERFACE extern char* RI_OL;
	DL_INTERFACE extern char* RI_I;
	DL_INTERFACE extern char* RI_U;
	DL_INTERFACE extern char* RI_V;
	DL_INTERFACE extern char* RI_DU;
	DL_INTERFACE extern char* RI_DV;
	DL_INTERFACE extern char* RI_NG;
	DL_INTERFACE extern char* RI_DPDU;
	DL_INTERFACE extern char* RI_DPDV;
}

/////////////////////////////////////////////////////////////////////////////
// Code for reading null references. Null references are default values
// for optional varying and output parameters. Normally, a null reference
// should be 0. Unfortunately, the "0" constant as a parameter will cause
// some trouble with the macros in this file, since "&0" is not valid.
// Since the interpreter always passes references, this is not a problem.
// But for C genrerated code, it is. Therefore, we declare this "nullRef"
// constant. All C generated code will use the "nullRef" keyword in the case
// of a null reference.
// The following macros are meant to detect such a case, put back "0" and be
// sure that the compiler doesn't panic.
/////////////////////////////////////////////////////////////////////////////

extern void *sg_nullRef;

#define READ_POINT_REF(Ref) \
(((DlVector3 *)&Ref == (DlVector3 *)&sg_nullRef) ? 0 : (DlVector3 *)(&(Ref)))

#define READ_POINT_REF_SIMD(Ref) \
(((DlVector3 *)Ref == (DlVector3 *)sg_nullRef) ? 0 : (DlVector3 *)Ref)

#define READ_FLOAT_REF(Ref) \
(((float *)&Ref == (float *)&sg_nullRef) ? 0 : (float *)(&(Ref)))

#define READ_FLOAT_REF_SIMD(Ref) \
(((float *)Ref == (float *)sg_nullRef) ? 0 : (float *)Ref)

/////////////////////////////////////////////////////////////////////////////
// For functions defined in DSO
/////////////////////////////////////////////////////////////////////////////

struct SlcDSOInfo {
	typedef void *(*InitFunction)(void);
	typedef int   (*MethodFunction)(void *initdata, int argc, void **argv);
	typedef void  (*CleanupFunction)(void *initdata);

	InitFunction     m_initFunc;
	MethodFunction   m_methodFunc;
	CleanupFunction  m_cleanupFunc;	
	void            *m_initdata;
};


////////////////////////////////////////////////////////////////////////////
// MACROS FOR ASSIGNMENTS
////////////////////////////////////////////////////////////////////////////

#define ASSIGN_FLOAT_FLOAT(F1, F2)  (F1) = (F2)

#define ASSIGN_COLOR_COLOR(C1, C2) ((C1)[0] = (C2)[0]), ((C1)[1] = (C2)[1]), ((C1)[2] = (C2)[2])

#define ASSIGN_POINT_POINT(P1, P2) ((P1)[0] = (P2)[0]), ((P1)[1] = (P2)[1]), ((P1)[2] = (P2)[2])

#define ASSIGN_MATRIX_MATRIX(M1, M2)    memcpy((M1), (M2), sizeof(DlMatrix44))

#define ASSIGN_COLOR_FLOAT(C1, F2) ((C1)[0] = F2), ((C1)[1] = F2), ((C1)[2] = F2)

#define ASSIGN_COLOR_POINT(C1, P2)  ASSIGN_COLOR_COLOR((C1), (P2))

#define ASSIGN_POINT_COLOR(P1, C2)  ASSIGN_POINT_POINT((P1), (C2))

#define ASSIGN_POINT_FLOAT(P1, F2) ((P1)[0] = F2), ((P1)[1] = F2), ((P1)[2] = F2)

#define ASSIGN_MATRIX_FLOAT(M1, F2)  \
   ((M1)[0]  = (F2),  (M1)[1]  = 0,    (M1)[2]  = 0,     (M1)[3]  = 0, \
    (M1)[4]  = 0,     (M1)[5]  = (F2), (M1)[6]  = 0,     (M1)[7]  = 0, \
    (M1)[8]  = 0,     (M1)[9]  = 0,    (M1)[10] = (F2),  (M1)[11] = 0, \
    (M1)[12] = 0,     (M1)[13] = 0,    (M1)[14] = 0,     (M1)[15] = (F2))

#define ASSIGN_BOOL_BOOL(B1, B2)  ((B1) = (B2))

#define ASSIGN_BOOL_BOOL_SIMD(B1, B2)  memcpy((B1), (B2), numMp * sizeof(DlStopAndBool))

#define INIT_POINT(P1, F2, F3, F4) ((P1)[0] = (F2), (P1)[1] = (F3), (P1)[2] = (F4))

#define INIT_COLOR(C1, F2, F3, F4) ((C1)[0] = (F2), (C1)[1] = (F3), (C1)[2] = (F4))

#define INIT_MATRIX(M, E0, E1, E2, E3, E4, E5, E6, E7, E8, E9, E10, E11, E12, E13, E14, E15) \
   ((M)[0] = (E0), (M)[1] = (E1), (M)[2] = (E2), (M)[3] = (E3), \
    (M)[4] = (E4), (M)[5] = (E5), (M)[6] = (E6), (M)[7] = (E7), \
    (M)[8] = (E8), (M)[9] = (E9), (M)[10] = (E10), (M)[11] = (E11), \
    (M)[12] = (E12), (M)[13] = (E13), (M)[14] = (E14), (M)[15] = (E15))
 

#define ASSIGN_FLOAT_FLOAT_SIMD(F1, F2) memcpy((F1), (F2), numMp * sizeof(float))

#define ASSIGN_COLOR_COLOR_SIMD(C1, C2) memcpy((C1), (C2), numMp * sizeof(DlRGBColor))

#define ASSIGN_POINT_POINT_SIMD(P1, P2) memcpy((P1), (P2), numMp * sizeof(DlVector3))

#define ASSIGN_MATRIX_MATRIX_SIMD(M1, M2) memcpy((M1), (M2), numMp * sizeof(DlMatrix44))

#define FREE_STRING(S) DlFreeStr((S))

#define ASSIGN_STRING_STRING(S1, S2) \
if( (S1) != (S2) ) \
{ \
	if( (S1) ) \
	{ \
		FREE_STRING( (S1) ); \
	} \
	(S1) = DlStrdup( (S2) ); \
}

#define ASSIGN_TO_ZERO_FLOAT(F1) \
	memset((F1), 0, numMp * sizeof(float))

#define ASSIGN_TO_ZERO_COLOR(C1) \
	memset((C1), 0, numMp * sizeof(DlRGBColor))

#define ASSIGN_TO_ZERO_POINT(P1) \
	memset((P1), 0, numMp * sizeof(DlVector3))

#define ASSIGN_TO_ZERO_MATRIX(M1) \
	memset((M1), 0, numMp * sizeof(DlMatrix44))


////////////////////////////////////////////////////////////////////////////
// MACROS FOR COLOR OR POINT SPACE TRANSFORMERS OPERATORS
////////////////////////////////////////////////////////////////////////////

#define CONVERT_COLOR_HSV_TO_RGB(C1, C2) \
	DlHSVToRGB(&(C1), &(C2), 0, 1)

#define CONVERT_COLOR_HSV_TO_RGB_SIMD(C1, C2) \
	DlHSVToRGB((C1), (C2), 0, numMp)

#define CONVERT_COLOR_HSV_TO_RGB_SIMD_IF_TRUE(C1, B, C2) \
	DlHSVToRGB((C1), (C2), (B),  numMp)

////////////////////////////////////////////////////////////////////////////
// POINTS, VECTOR AND NORMALS TRANSFORMATION
//
// Spaces are known at compile-time.
////////////////////////////////////////////////////////////////////////////

#define CALL_PRIM_TRANSFORM_TO_CURRENT(P1, P2, SPACE)  \
	DlTransformSpace(&(P1), &(P2), i_context.SpaceList()[SPACE], 0,\
		DlTokenInfo::e_RtPoint, false, 0, 1)

#define CALL_PRIM_VTRANSFORM_TO_CURRENT(P1, P2, SPACE)  \
	DlTransformSpace(&(P1), &(P2), i_context.SpaceList()[SPACE], 0, \
		DlTokenInfo::e_RtVector, false, 0, 1)

#define CALL_PRIM_NTRANSFORM_TO_CURRENT(P1, P2, SPACE)  \
	DlTransformSpace(&(P1), &(P2), i_context.SpaceList()[SPACE], 0, \
		DlTokenInfo::e_RtNormal, false, 0, 1)

#define CALL_PRIM_TRANSFORM_FROM_CURRENT(P1, P2, SPACE)  \
	DlTransformSpace(&(P1), &(P2), i_context.SpaceList()[SPACE], 0, \
		DlTokenInfo::e_RtPoint, true, 0, 1)

#define CALL_PRIM_VTRANSFORM_FROM_CURRENT(P1, P2, SPACE)  \
	DlTransformSpace(&(P1), &(P2), i_context.SpaceList()[SPACE], 0, \
		DlTokenInfo::e_RtVector, true, 0, 1)

#define CALL_PRIM_NTRANSFORM_FROM_CURRENT(P1, P2, SPACE)  \
	DlTransformSpace(&(P1), &(P2), i_context.SpaceList()[SPACE], 0, \
		DlTokenInfo::e_RtNormal, true, 0, 1)


#define CALL_PRIM_TRANSFORM_TO_CURRENT_SIMD(P1, P2, SPACE)  \
	DlTransformSpace((P1), (P2), i_context.SpaceList()[SPACE], 0, \
		DlTokenInfo::e_RtPoint, false, 0, numMp)

#define CALL_PRIM_VTRANSFORM_TO_CURRENT_SIMD(P1, P2, SPACE)  \
	DlTransformSpace((P1), (P2), i_context.SpaceList()[SPACE], 0, \
		DlTokenInfo::e_RtVector, false, 0, numMp)

#define CALL_PRIM_NTRANSFORM_TO_CURRENT_SIMD(P1, P2, SPACE)  \
	DlTransformSpace((P1), (P2), i_context.SpaceList()[SPACE], 0, \
		DlTokenInfo::e_RtNormal, false, 0, numMp)

#define CALL_PRIM_TRANSFORM_FROM_CURRENT_SIMD(P1, P2, SPACE)  \
	DlTransformSpace((P1), (P2), i_context.SpaceList()[SPACE], 0, \
		DlTokenInfo::e_RtPoint, true, 0, numMp)

#define CALL_PRIM_VTRANSFORM_FROM_CURRENT_SIMD(P1, P2, SPACE)  \
	DlTransformSpace((P1), (P2), i_context.SpaceList()[SPACE], 0, \
		DlTokenInfo::e_RtVector, true, 0, numMp)

#define CALL_PRIM_NTRANSFORM_FROM_CURRENT_SIMD(P1, P2, SPACE)  \
	DlTransformSpace((P1), (P2), i_context.SpaceList()[SPACE], 0, \
		DlTokenInfo::e_RtNormal, true, 0, numMp)

#define CALL_PRIM_TRANSFORM_TO_CURRENT_SIMD_IF_TRUE(P1, B, P2, SPACE)  \
	DlTransformSpace((P1), (P2), i_context.SpaceList()[SPACE], 0, \
		DlTokenInfo::e_RtPoint, false, (B),  \
		numMp)

#define CALL_PRIM_VTRANSFORM_TO_CURRENT_SIMD_IF_TRUE(P1, B, P2, SPACE)  \
	DlTransformSpace((P1), (P2), i_context.SpaceList()[SPACE], 0, \
		DlTokenInfo::e_RtVector, false, (B),  \
		numMp)

#define CALL_PRIM_NTRANSFORM_TO_CURRENT_SIMD_IF_TRUE(P1, B, P2, SPACE)  \
	DlTransformSpace((P1), (P2), i_context.SpaceList()[SPACE], 0, \
		DlTokenInfo::e_RtNormal, false, (B),  \
		numMp)

#define CALL_PRIM_TRANSFORM_FROM_CURRENT_SIMD_IF_TRUE(P1, B, P2, SPACE)  \
	DlTransformSpace((P1), (P2), i_context.SpaceList()[SPACE], 0, \
		DlTokenInfo::e_RtPoint, true, (B),  \
		numMp)

#define CALL_PRIM_VTRANSFORM_FROM_CURRENT_SIMD_IF_TRUE(P1, B, P2, SPACE)  \
	DlTransformSpace((P1), (P2), i_context.SpaceList()[SPACE], 0, \
		DlTokenInfo::e_RtVector, true, (B),  \
		numMp)

#define CALL_PRIM_NTRANSFORM_FROM_CURRENT_SIMD_IF_TRUE(P1, B, P2, SPACE)  \
	DlTransformSpace((P1), (P2), i_context.SpaceList()[SPACE], 0, \
		DlTokenInfo::e_RtNormal, true, (B),  \
		numMp)

////////////////////////////////////////////////////////////////////////////
// MACROS FOR ARITHMETIC OPERATORS
////////////////////////////////////////////////////////////////////////////

#define ADD_FLOAT_FLOAT(F1, F2, F3)  (F1) = (F2) + (F3)

#define ADD_FLOAT_COLOR(C1, F2, C3)  ((C1)[0] = (F2) + (C3)[0]), \
                                     ((C1)[1] = (F2) + (C3)[1]), \
                                     ((C1)[2] = (F2) + (C3)[2])

#define ADD_FLOAT_POINT(P1, F2, P3)  ((P1)[0] = (F2) + (P3)[0]), \
                                     ((P1)[1] = (F2) + (P3)[1]), \
                                     ((P1)[2] = (F2) + (P3)[2])

#define ADD_COLOR_FLOAT(C1, C2, F3)  ((C1)[0] = (C2)[0] + (F3)), \
                                     ((C1)[1] = (C2)[1] + (F3)), \
                                     ((C1)[2] = (C2)[2] + (F3))

#define ADD_COLOR_COLOR(C1, C2, C3)  ((C1)[0] = (C2)[0] + (C3)[0]), \
                                     ((C1)[1] = (C2)[1] + (C3)[1]), \
                                     ((C1)[2] = (C2)[2] + (C3)[2])

#define ADD_POINT_FLOAT(P1, P2, F3)  ((P1)[0] = (P2)[0] + (F3)), \
                                     ((P1)[1] = (P2)[1] + (F3)), \
                                     ((P1)[2] = (P2)[2] + (F3))

#define ADD_POINT_POINT(P1, P2, P3)  ((P1)[0] = (P2)[0] + (P3)[0]), \
                                     ((P1)[1] = (P2)[1] + (P3)[1]), \
                                     ((P1)[2] = (P2)[2] + (P3)[2])


#define SUB_FLOAT_FLOAT(F1, F2, F3)  (F1) = (F2) - (F3)

#define SUB_FLOAT_COLOR(F1, F2, C3)  ((F1)[0] = (F2) - (C3)[0]), \
                                     ((F1)[1] = (F2) - (C3)[1]), \
                                     ((F1)[2] = (F2) - (C3)[2])

#define SUB_FLOAT_POINT(F1, F2, P3)  ((F1)[0] = (F2) - (P3)[0]), \
                                     ((F1)[1] = (F2) - (P3)[1]), \
                                     ((F1)[2] = (F2) - (P3)[2])
 
#define SUB_COLOR_FLOAT(C1, C2, F3)  ((C1)[0] = (C2)[0] - (F3)), \
                                     ((C1)[1] = (C2)[1] - (F3)), \
                                     ((C1)[2] = (C2)[2] - (F3))

#define SUB_COLOR_COLOR(C1, C2, C3)  ((C1)[0] = (C2)[0] - (C3)[0]), \
                                     ((C1)[1] = (C2)[1] - (C3)[1]), \
                                     ((C1)[2] = (C2)[2] - (C3)[2])

#define SUB_POINT_FLOAT(P1, P2, F3)  ((P1)[0] = (P2)[0] - (F3)), \
                                     ((P1)[1] = (P2)[1] - (F3)), \
                                     ((P1)[2] = (P2)[2] - (F3))

#define SUB_POINT_POINT(P1, P2, P3)  ((P1)[0] = (P2)[0] - (P3)[0]), \
                                     ((P1)[1] = (P2)[1] - (P3)[1]), \
                                     ((P1)[2] = (P2)[2] - (P3)[2])


#define MULT_FLOAT_FLOAT(F1, F2, F3)  (F1) = (F2) * (F3)

#define MULT_FLOAT_COLOR(C1, F2, C3)  ((C1)[0] = (F2) * (C3)[0]), \
                                      ((C1)[1] = (F2) * (C3)[1]), \
                                      ((C1)[2] = (F2) * (C3)[2])

#define MULT_FLOAT_POINT(P1, F2, P3)  ((P1)[0] = (F2) * (P3)[0]), \
                                      ((P1)[1] = (F2) * (P3)[1]), \
                                      ((P1)[2] = (F2) * (P3)[2])

#define MULT_FLOAT_MATRIX(M1, F2, M3) \
{ \
    (M1)[0] = (F2) * (M3)[0]; \
    (M1)[1] = (F2) * (M3)[1]; \
    (M1)[2] = (F2) * (M3)[2]; \
    (M1)[3] = (F2) * (M3)[3]; \
    (M1)[4] = (F2) * (M3)[4]; \
    (M1)[5] = (F2) * (M3)[5]; \
    (M1)[6] = (F2) * (M3)[6]; \
    (M1)[7] = (F2) * (M3)[7]; \
    (M1)[8] = (F2) * (M3)[8]; \
    (M1)[9] = (F2) * (M3)[9]; \
    (M1)[10] = (F2) * (M3)[10]; \
    (M1)[11] = (F2) * (M3)[11]; \
    (M1)[12] = (F2) * (M3)[12]; \
    (M1)[13] = (F2) * (M3)[13]; \
    (M1)[14] = (F2) * (M3)[14]; \
    (M1)[15] = (F2) * (M3)[15]; \
}    


#define MULT_COLOR_FLOAT(C1, C2, F3)  ((C1)[0] = (C2)[0] * (F3)), \
                                      ((C1)[1] = (C2)[1] * (F3)), \
                                      ((C1)[2] = (C2)[2] * (F3))
  
#define MULT_COLOR_COLOR(C1, C2, C3)  ((C1)[0] = (C2)[0] * (C3)[0]), \
                                      ((C1)[1] = (C2)[1] * (C3)[1]), \
                                      ((C1)[2] = (C2)[2] * (C3)[2])

#define MULT_POINT_FLOAT(P1, P2, F3)  ((P1)[0] = (P2)[0] * (F3)), \
                                      ((P1)[1] = (P2)[1] * (F3)), \
                                      ((P1)[2] = (P2)[2] * (F3))

#define MULT_POINT_POINT(P1, P2, P3)  ((P1)[0] = (P2)[0] * (P3)[0]), \
                                      ((P1)[1] = (P2)[1] * (P3)[1]), \
                                      ((P1)[2] = (P2)[2] * (P3)[2])

#define MULT_MATRIX_FLOAT(M1, M3, F2)  \
{ \
    (M1)[0] = (F2) * (M3)[0]; \
    (M1)[1] = (F2) * (M3)[1]; \
    (M1)[2] = (F2) * (M3)[2]; \
    (M1)[3] = (F2) * (M3)[3]; \
    (M1)[4] = (F2) * (M3)[4]; \
    (M1)[5] = (F2) * (M3)[5]; \
    (M1)[6] = (F2) * (M3)[6]; \
    (M1)[7] = (F2) * (M3)[7]; \
    (M1)[8] = (F2) * (M3)[8]; \
    (M1)[9] = (F2) * (M3)[9]; \
    (M1)[10] = (F2) * (M3)[10]; \
    (M1)[11] = (F2) * (M3)[11]; \
    (M1)[12] = (F2) * (M3)[12]; \
    (M1)[13] = (F2) * (M3)[13]; \
    (M1)[14] = (F2) * (M3)[14]; \
    (M1)[15] = (F2) * (M3)[15]; \
}    

#define MULT_MATRIX_MATRIX(M1, M2, M3) \
{ \
    DlMatrix44 t; \
\
    t[0] = (M2)[0] * (M3[0]) + (M2)[1] * (M3)[4] + (M2)[2] * (M3)[8]  + (M2)[3] * (M3)[12];  \
    t[1] = (M2)[0] * (M3[1]) + (M2)[1] * (M3)[5] + (M2)[2] * (M3)[9]  + (M2)[3] * (M3)[13];  \
    t[2] = (M2)[0] * (M3[2]) + (M2)[1] * (M3)[6] + (M2)[2] * (M3)[10] + (M2)[3] * (M3)[14];  \
    t[3] = (M2)[0] * (M3[3]) + (M2)[1] * (M3)[7] + (M2)[2] * (M3)[11] + (M2)[3] * (M3)[15];  \
\
    t[4] = (M2)[4] * (M3[0]) + (M2)[5] * (M3)[4] + (M2)[6] * (M3)[8]  + (M2)[7] * (M3)[12];  \
    t[5] = (M2)[4] * (M3[1]) + (M2)[5] * (M3)[5] + (M2)[6] * (M3)[9]  + (M2)[7] * (M3)[13];  \
    t[6] = (M2)[4] * (M3[2]) + (M2)[5] * (M3)[6] + (M2)[6] * (M3)[10] + (M2)[7] * (M3)[14];  \
    t[7] = (M2)[4] * (M3[3]) + (M2)[5] * (M3)[7] + (M2)[6] * (M3)[11] + (M2)[7] * (M3)[15];  \
\
    t[8]  = (M2)[8] * (M3[0]) + (M2)[9] * (M3)[4] + (M2)[10] * (M3)[8]  + (M2)[11] * (M3)[12];  \
    t[9]  = (M2)[8] * (M3[1]) + (M2)[9] * (M3)[5] + (M2)[10] * (M3)[9]  + (M2)[11] * (M3)[13];  \
    t[10] = (M2)[8] * (M3[2]) + (M2)[9] * (M3)[6] + (M2)[10] * (M3)[10] + (M2)[11] * (M3)[14];  \
    t[11] = (M2)[8] * (M3[3]) + (M2)[9] * (M3)[7] + (M2)[10] * (M3)[11] + (M2)[11] * (M3)[15];  \
\
    t[12] = (M2)[12] * (M3[0]) + (M2)[13] * (M3)[4] + (M2)[14] * (M3)[8]  + (M2)[15] * (M3)[12];  \
    t[13] = (M2)[12] * (M3[1]) + (M2)[13] * (M3)[5] + (M2)[14] * (M3)[9]  + (M2)[15] * (M3)[13];  \
    t[14] = (M2)[12] * (M3[2]) + (M2)[13] * (M3)[6] + (M2)[14] * (M3)[10] + (M2)[15] * (M3)[14];  \
    t[15] = (M2)[12] * (M3[3]) + (M2)[13] * (M3)[7] + (M2)[14] * (M3)[11] + (M2)[15] * (M3)[15];  \
\
    ASSIGN_MATRIX_MATRIX((M1), t); \
}

#define CONCAT_SPACE_MATRIX(M1, M2, SPACE) \
{ \
    const DlMatrix44 *tempM = DlGetToSpaceMatrix(i_context.SpaceList()[SPACE]); \
\
    if (tempM) { \
       MULT_MATRIX_MATRIX((M1), (M2), (*tempM)); \
    } \
    else { \
	   ASSIGN_MATRIX_MATRIX((M1), (M2)); \
    } \
} \


/* ------------------------------------------ D I V I D E    P R I M I T I V E S

	NOTE:
	- We always check if the devider is 0. If it is, we set the result to
	FLT_MAX, this should be the correct behaviour for most applications. To
	be perfectly right, we should also consider the 0/0 case, which is undefied,
	but we let this check to the user inside the shader...

	- Divide operations performred on points reuse the macros defined for
	colors (since they have both 3 componenets).
*/

#define DIV_FLOAT_FLOAT(F1, F2, F3) \
{ \
	if( (F3) == 0.0f ) \
	{ \
		(F1) = FLT_MAX;	\
	} \
	else \
	{ \
		(F1) = (F2) / (F3); \
	} \
}

#define DIV_FLOAT_COLOR(C1, F2, C3) \
{ \
	if( (C3)[0] != 0.0f ) \
		(C1)[0] = (F2) / (C3)[0]; \
	else \
		(C1)[0]  = FLT_MAX; \
	\
	if( (C3)[1] != 0.0f ) \
		(C1)[1] = (F2) / (C3)[1]; \
	else \
		(C1)[1]  = FLT_MAX; \
	\
	if( (C3)[2] != 0.0f ) \
		(C1)[2] = (F2) / (C3)[2]; \
	else \
		(C1)[2] = FLT_MAX; \
}

/* WARNING: Uses same code as DIV_FLOAT_COLOR */
#define DIV_FLOAT_POINT(P1, F2, P3)  \
	DIV_FLOAT_COLOR(P1, F2, P3)

#define DIV_FLOAT_MATRIX(M1, F2, M3)  \
{ \
    DlMatrix44 temp; DlInverse(&temp, &(M3)); \
  \
    MULT_FLOAT_MATRIX((M1), (F2), temp);  \
}

#define DIV_COLOR_FLOAT(C1, C2, F3) \
{ \
	if( (F3) == 0.0f ) \
	{ \
		(C1)[0] = (C1)[1] = (C1)[2] = FLT_MAX; \
	} \
	else \
	{ \
		float inv = 1.0f / (F3); \
		(C1)[0] = (C2)[0] * inv; \
		(C1)[1] = (C2)[1] * inv; \
		(C1)[2] = (C2)[2] * inv; \
	} \
}

#define DIV_COLOR_COLOR(C1, C2, C3) \
{ \
	if( (C3)[0] != 0.0f ) \
		(C1)[0] = (C2)[0] / (C3)[0]; \
	else \
		(C1)[0] = FLT_MAX; \
	\
	if( (C3)[1] != 0.0f ) \
		(C1)[1] = (C2)[1] / (C3)[1]; \
	else \
		(C1)[1] = FLT_MAX; \
	\
	if( (C3)[2] != 0.0f ) \
		(C1)[2] = (C2)[2] / (C3)[2]; \
	else \
		(C1)[2] = FLT_MAX; \
}

#define DIV_POINT_FLOAT(P1, P2, F3) \
	DIV_COLOR_FLOAT(P1, P2, F3)

#define DIV_POINT_POINT(P1, P2, P3) \
	DIV_COLOR_COLOR(P1, P2, P3)

#define DIV_MATRIX_FLOAT(M1, M2, F3)  \
{ \
	float temp; \
	DIV_FLOAT_FLOAT(temp, 1.0f, (F3)); \
	MULT_MATRIX_FLOAT((M1), (M2), temp);  \
}

#define DIV_MATRIX_MATRIX(M1, M2, M3) \
{  \
	DlMatrix44 temp; \
	DlInverse( &temp, &(M3) );  \
	MULT_MATRIX_MATRIX((M1), (M2), temp); \
}

#define DOT_PROD_COLOR_COLOR(F1, C2, C3) \
   (F1) = ((C2)[0] * (C3)[0]) + ((C2)[1] * (C3)[1]) + ((C2)[2] * (C3)[2])

#define DOT_PROD_FLOAT_COLOR(F1, F2, C3) \
   (F1) = (F2) * ((C3)[0] + (C3)[1] + (C3)[2])

#define DOT_PROD_COLOR_FLOAT(F1, C2, F3) \
   (F1) = (F3) * ((C2)[0] + (C2)[1] + (C2)[2])

#define DOT_PROD_POINT_POINT(F1, P2, P3) \
   (F1) = ((P2)[0] * (P3)[0]) + ((P2)[1] * (P3)[1]) + ((P2)[2] * (P3)[2])

#define DOT_PROD_FLOAT_POINT(F1, F2, P3) \
   (F1) = (F2) * ((P3)[0] + (P3)[1] + (P3)[2])

#define DOT_PROD_POINT_FLOAT(F1, P2, F3) \
   (F1) = (F3) * ((P2)[0] + (P2)[1] + (P2)[2])


#define CROSS_PROD_POINT_POINT(P1, P2, P3) \
{ \
	float temp1, temp2, temp3; \
	temp1 = (P2)[1] * (P3)[2] - (P2)[2] * (P3)[1]; \
	temp2 = (P2)[2] * (P3)[0] - (P2)[0] * (P3)[2]; \
	temp3 = (P2)[0] * (P3)[1] - (P2)[1] * (P3)[0]; \
	(P1)[0] = temp1; (P1)[1] = temp2; (P1)[2] = temp3; \
}

#define CROSS_PROD_FLOAT_POINT(P1, F2, P3) \
{ \
	float temp1, temp2, temp3; \
	temp1 = (F2) * ((P3)[2] - (P3)[1]); \
	temp2 = (F2) * ((P3)[0] - (P3)[2]); \
	temp3 = (F2) * ((P3)[1] - (P3)[0]); \
	(P1)[0] = temp1; (P1)[1] = temp2; (P1)[2] = temp3; \
}

#define CROSS_PROD_POINT_FLOAT(P1, P2, F3) \
{ \
	float temp1, temp2, temp3; \
	temp1 = (F3) * ((P2)[1] - (P2)[2]); \
	temp2 = (F3) * ((P2)[2] - (P2)[0]); \
	temp3 = (F3) * ((P2)[0] - (P2)[1]); \
	(P1)[0] = temp1; (P1)[1] = temp2; (P1)[2] = temp3; \
}

#define INVERT_FLOAT(F1, F2) (F1) = -(F2)

#define INVERT_COLOR(C1, C2)     ((C1)[0] = -(C2)[0]), \
                                 ((C1)[1] = -(C2)[1]), \
                                 ((C1)[2] = -(C2)[2])

#define INVERT_POINT(P1, P2)     ((P1)[0] = -(P2)[0]), \
                                 ((P1)[1] = -(P2)[1]), \
                                 ((P1)[2] = -(P2)[2])

//#define INVERT_MATRIX(M1, M2)    invert_matrix_to_be_implemented


////////////////////////////////////////////////////////////////////////////
// MACROS FOR RELATIONS RETURNING BOOLEAN RESULTS (Version 2)
////////////////////////////////////////////////////////////////////////////

#define BOOL_LESS_THEN_FLOAT_FLOAT(B1, F1, F2)     SET_BOOL((B1), ((F1) < (F2)))

#define BOOL_LESS_EQUAL_FLOAT_FLOAT(B1, F1, F2)    SET_BOOL((B1), ((F1) <= (F2)))

#define BOOL_GREATER_THEN_FLOAT_FLOAT(B1, F1, F2)  SET_BOOL((B1), ((F1) > (F2)))

#define BOOL_GREATER_EQUAL_FLOAT_FLOAT(B1, F1, F2) SET_BOOL((B1), ((F1) >= (F2)))

#define BOOL_EQUAL_FLOAT_FLOAT(B1, F1, F2)         SET_BOOL((B1), ((F1) == (F2)))

#define BOOL_NOT_EQUAL_FLOAT_FLOAT(B1, F1, F2)     SET_BOOL((B1), ((F1) != (F2)))

#define BOOL_EQUAL_STRING_STRING(B1, S1, S2)       SET_BOOL((B1), (!strcmp((S1), (S2))))

#define BOOL_NOT_EQUAL_STRING_STRING(B1, S1, S2)   SET_BOOL((B1), (strcmp((S1), (S2))))

#define BOOL_EQUAL_COLOR_COLOR(B1, C1, C2)         SET_BOOL((B1), ((C1)[0] == (C2)[0] && (C1)[1] == (C2)[1] && (C1)[2] == (C2)[2]))

#define BOOL_NOT_EQUAL_COLOR_COLOR(B1, C1, C2)     SET_BOOL((B1), ((C1)[0] != (C2)[0] || (C1)[1] != (C2)[1] || (C1)[2] != (C2)[2]))

#define BOOL_EQUAL_POINT_POINT(B1, P1, P2)         SET_BOOL((B1), ((P1)[0] == (P2)[0] && (P1)[1] == (P2)[1] && (P1)[2] == (P2)[2]))

#define BOOL_NOT_EQUAL_POINT_POINT(B1, P1, P2)     SET_BOOL((B1), ((P1)[0] != (P2)[0] || (P1)[1] != (P2)[1] || (P1)[2] != (P2)[2]))

#define BOOL_NOT_EQUAL_FLOAT_POINT(B1, F1, P2)     SET_BOOL((B1), ((F1) != (P2)[0] || (F1) != (P2)[1] || (F1) != (P2)[2]))

#define BOOL_NOT_EQUAL_POINT_FLOAT(B1, P1, F2)     SET_BOOL((B1), ((P1)[0] != (F2) || (P1)[1] != (F2) || (P1)[2] != (F2)))

#define BOOL_NOT_EQUAL_FLOAT_COLOR(B1, F1, C2)     SET_BOOL((B1), ((F1) != (C2)[0] || (F1) != (C2)[1] || (F1) != (C2)[2]))

#define BOOL_NOT_EQUAL_COLOR_FLOAT(B1, C1, F2)     SET_BOOL((B1), ((C1)[0] != (F2) || (C1)[1] != (F2) || (C1)[2] != (F2)))

#define BOOL_EQUAL_FLOAT_POINT(B1, F1, P2)     SET_BOOL((B1), ((F1) == (P2)[0] && (F1) == (P2)[1] && (F1) == (P2)[2]))

#define BOOL_EQUAL_POINT_FLOAT(B1, P1, F2)     SET_BOOL((B1), ((P1)[0] == (F2) && (P1)[1] == (F2) && (P1)[2] == (F2)))

#define BOOL_EQUAL_FLOAT_COLOR(B1, F1, C2)     SET_BOOL((B1), ((F1) == (C2)[0] && (F1) == (C2)[1] && (F1) == (C2)[2]))

#define BOOL_EQUAL_COLOR_FLOAT(B1, C1, F2)     SET_BOOL((B1), ((C1)[0] == (F2) && (C1)[1] == (F2) && (C1)[2] == (F2)))

#define BOOL_EQUAL_MATRIX_MATRIX(B, M1, M2)    SET_BOOL((B), !memcmp((M1), (M2), sizeof(DlMatrix44)))
  
#define BOOL_EQUAL_FLOAT_MATRIX(B, F, M) \
{ \
    DlMatrix44 t;   ASSIGN_MATRIX_FLOAT(t, (F));  \
\
    SET_BOOL((B), !memcmp((M), t, sizeof(DlMatrix44))); \
}

#define BOOL_EQUAL_MATRIX_FLOAT(B, M, F)  BOOL_EQUAL_FLOAT_MATRIX(B, F, M)

#define BOOL_NOT_EQUAL_MATRIX_MATRIX(B, M1, M2)    SET_BOOL((B), memcmp((M1), (M2), sizeof(DlMatrix44)))
  
#define BOOL_NOT_EQUAL_FLOAT_MATRIX(B, F, M) \
{ \
    DlMatrix44 t;   ASSIGN_MATRIX_FLOAT(t, (F));  \
\
    SET_BOOL((B), memcmp((M), t, sizeof(DlMatrix44))); \
}

#define BOOL_NOT_EQUAL_MATRIX_FLOAT(B, M, F)  BOOL_NOT_EQUAL_FLOAT_MATRIX(B, F, M)



/* ------------------------------------------------ M A T H    F U N C T I O N S
*/
#define DL_PI 3.1415926536f

#define CALL_PRIM_RADIANS(F1, F2)  (F1) = (F2) * (DL_PI / 180)

#define CALL_PRIM_DEGREES(F1, F2)  (F1) = (F2) * (180 / DL_PI)

#define CALL_PRIM_SIN(F1, F2) (F1) = sin(F2)

#define CALL_PRIM_ASIN(F1, F2) (F1) = (((F2) < -1.0) ? DL_PI : ((F2) > 1.0) ? 0 : asin(F2))

#define CALL_PRIM_COS(F1, F2) (F1) = cos(F2)

#define CALL_PRIM_ACOS(F1, F2) (F1) = (((F2) < -1.0) ? DL_PI : ((F2) > 1.0) ? 0 : acos(F2))

#define CALL_PRIM_TAN(F1, F2) (F1) = tan(F2)

#define CALL_PRIM_ATAN(F1, F2) (F1) = atan(F2)

#define CALL_PRIM_ATAN2(F1, F2, F3) (F1) = atan2((F2), (F3))

#define CALL_PRIM_POW(F1, F2, F3) \
{ \
	if( (F2)==0.f ) \
	{ \
		(F1)=0.f; \
	} \
	else if( (F2)<0.f ) \
	{ \
		/* if F2 is negative, then F3 must be an integer */ \
		if( ceil(F3)!=F3 ) \
		{ \
			(F1) = .0f; \
		} \
		else \
		{ \
			(F1) = pow((F2), (F3)); \
		} \
	} \
	else \
	{ \
		(F1) = pow((F2), (F3)); \
	} \
}

#define CALL_PRIM_EXP(F1, F2) (F1) = exp(F2)

#define CALL_PRIM_SQRT(F1, F2) \
{ \
	if ((F2) <= 0.0) \
		(F1) = 0.0; \
	else \
		(F1) = sqrt(F2); \
}

#define CALL_PRIM_INVERSESQRT(F1, F2) \
{ \
	if ((F2) <= 0.0) \
		(F1) = 0.0; \
	else \
		(F1) = 1.0 / sqrt(F2); \
}

#define CALL_PRIM_LOG(F1, F2) { if ((F2) <= 0.0) (F1) = 0.0; else (F1) = log(F2); }

#define CALL_PRIM_LOG2(F1, F2, F3) { if ((F2) <= 0.0 || ((F3) <= 0.0)) (F1) = 0; else (F1) = (log(F2)) / (log(F3));}

#define CALL_PRIM_MOD(F1, F2, F3)  { int n = (int)((F2)/(F3)); \
                                     (F1) = (F2) - n * (F3); \
                                     if ((F1) < 0) (F1) += fabs(F3); }

#define CALL_PRIM_ABS(F1, F2)      (F1) = fabs(F2)

#define CALL_PRIM_SIGN(F1, F2)     (F1) = ((F2) > 0) ? 1.0 : (((F2) < 0) ? -1.0 : 0.0)

#define CALL_PRIM_MAX2_FLOAT(F1, F2, F3)      ((F1) = ((F2) > (F3)) ? (F2) : (F3))

#define CALL_PRIM_MIN2_FLOAT(F1, F2, F3)      ((F1) = ((F2) < (F3)) ? (F2) : (F3))

#define CALL_PRIM_CLAMP_1D(F1, F2, F3, F4)  \
      (F1) = ((F2) < (F3)) ? (F3) : (((F2) > (F4)) ? (F4) : (F2))

#define CALL_PRIM_CLAMP_3D(P1, P2, P3, P4)  DlClamp3D(&(P1), &(P2), (P3), (P4), 0, 1);

#define CALL_PRIM_CLAMP_3D_SIMD(P1, P2, P3, P4)  \
           DlClamp3D((P1), (P2), (P3), (P4), 0, numMp);

#define CALL_PRIM_CLAMP_3D_SIMD_IF_TRUE(P1, B, P2, P3, P4)  \
           DlClamp3D((P1), (P2), (P3), (P4), (B), numMp);

#define CALL_PRIM_MIN_1D(F1, ARGS, NB)  DlMin1D(&(F1), (ARGS), (NB))

#define CALL_PRIM_MIN_3D(P1, ARGS, NB)  DlMin3D(&(P1), (ARGS), (NB))

#define CALL_PRIM_MAX_1D(F1, ARGS, NB)  DlMax1D(&(F1), (ARGS), (NB))

#define CALL_PRIM_MAX_3D(P1, ARGS, NB)  DlMax3D(&(P1), (ARGS), (NB))

#define CALL_PRIM_FLOOR(F1, F2)     (F1) = floor(F2)

#define CALL_PRIM_CEIL(F1, F2)      (F1) = ceil(F2)

#define CALL_PRIM_ROUND(F1, F2)     (F1) = floor((F2) + 0.5)

#define CALL_PRIM_STEP(F1, F2, F3)  (F1) = ((F3) < (F2)) ? 0 : 1

#define CALL_PRIM_SMOOTHSTEP(F1, F2, F3, F4)  \
{ \
    float temp = (F4); \
    DlSmoothStep(&(F1), (F2), (F3), &temp, 0, 1); \
}

#define CALL_PRIM_SMOOTHSTEP_SIMD(F1, F2, F3, F4) \
    DlSmoothStep((F1), (F2), (F3), (F4), 0, numMp)

#define CALL_PRIM_SMOOTHSTEP_SIMD_IF_TRUE(F1, B2, F2, F3, F4) \
    DlSmoothStep((F1), (F2), (F3), (F4), (B2),  numMp)


#define CALL_PRIM_SPLINE_1D(F1, ARGS, NB)  \
    DlSpline1DWithBasis(&(F1), e_CatmullRomBasis, (ARGS), (NB))

#define CALL_PRIM_SPLINE_3D(P1, ARGS, NB)  \
    DlSpline3DWithBasis(&(P1), e_CatmullRomBasis, (ARGS), (NB))

#define CALL_PRIM_SPLINE_BASIS_1D(F1, ARGS, NB)  DlSpline1D(&(F1), (ARGS), (NB))

#define CALL_PRIM_SPLINE_BASIS_3D(F1, ARGS, NB)  DlSpline3D(&(F1), (ARGS), (NB))


#define CALL_PRIM_DU_FLOAT_SIMD(F1, F2, DU)    DlDu1D((F1), (F2), (DU), numMpU, numMpV, 0)

#define CALL_PRIM_DU_FLOAT_SIMD_IF_TRUE(F1, B, F2, DU)  DlDu1D((F1), (F2), (DU), numMpU, numMpV, (B))


#define CALL_PRIM_DV_FLOAT_SIMD(F1, F2, DV)    DlDv1D((F1), (F2), (DV), numMpU, numMpV, 0)

#define CALL_PRIM_DV_FLOAT_SIMD_IF_TRUE(F1, B, F2, DV)  DlDv1D((F1), (F2), (DV), numMpU, numMpV, (B))


#define CALL_PRIM_DERIV_FLOAT_SIMD(F1, F2, F3) DlDeriv1D((F1), (F2), (F3), numMpU, numMpV, 0)

#define CALL_PRIM_DERIV_FLOAT_SIMD_IF_TRUE(F1, B, F2, F3) \
                                          DlDeriv1D((F1), (F2), (F3), numMpU, numMpV, (B))



#define CALL_PRIM_DU_COLOR_SIMD(C1, C2, DU)    DlDu3D((C1), (C2), (DU), numMpU, numMpV, 0) 

#define CALL_PRIM_DU_COLOR_SIMD_IF_TRUE(C1, B, C2, DU)    \
                                          DlDu3D((C1), (C2), (DU), numMpU, numMpV, (B)) 



#define CALL_PRIM_DV_COLOR_SIMD(C1, C2, DV)    DlDv3D((C1), (C2), (DV), numMpU, numMpV, 0) 

#define CALL_PRIM_DV_COLOR_SIMD_IF_TRUE(C1, B, C2, DV)    \
                                          DlDv3D((C1), (C2), (DV), numMpU, numMpV, (B)) 



#define CALL_PRIM_DERIV_COLOR_SIMD(C1, C2, F3) DlDeriv3D((C1), (c2), (F3), numMpU, numMpV, 0)

#define CALL_PRIM_DERIV_COLOR_SIMD_IF_TRUE(C1, B, C2, F3) \
                                          DlDeriv3D((C1), (c2), (F3), numMpU, numMpV, (B))



#define CALL_PRIM_DU_POINT_SIMD(P1, P2, DU)  DlDu3D((P1), (P2), (DU), numMpU, numMpV, 0)

#define CALL_PRIM_DU_POINT_SIMD_IF_TRUE(P1, B, P2, DU)    \
                                          DlDu3D((P1), (P2), (DU), numMpU, numMpV, (B))



#define CALL_PRIM_DV_POINT_SIMD(P1, P2, DV)  DlDv3D((P1), (P2), (DV), numMpU, numMpV, 0)

#define CALL_PRIM_DV_POINT_SIMD_IF_TRUE(P1, B, P2, DV)    \
                                          DlDv3D((P1), (P2), (DV), numMpU, numMpV, (B))



#define CALL_PRIM_DERIV_POINT_SIMD(P1, P2, F3) DlDeriv3D((P1), (P2), (F3), numMpU, numMpV, 0)

#define CALL_PRIM_DERIV_POINT_SIMD_IF_TRUE(P1, B, P2, F3) \
                                          DlDeriv3D((P1), (P2), (F3), numMpU, numMpV, (B))




#define CALL_PRIM_RANDOM_FLOAT(F1)                DlRandom1D(&(F1), 0, 1)

#define CALL_PRIM_RANDOM_FLOAT_SIMD(F1)           DlRandom1D((F1), 0, numMp)

#define CALL_PRIM_RANDOM_FLOAT_SIMD_IF_TRUE(F1, B)   DlRandom1D((F1), (B),  numMp)


#define CALL_PRIM_RANDOM_COLOR(C1)                DlRandom3D(&(C1), 0, 1)

#define CALL_PRIM_RANDOM_COLOR_SIMD(C1)           DlRandom3D((C1), 0, numMp)

#define CALL_PRIM_RANDOM_COLOR_SIMD_IF_TRUE(C1, B)   DlRandom3D((C1), (B),  numMp)


#define CALL_PRIM_RANDOM_POINT(C1)                DlRandom3D(&(C1), 0, 1)

#define CALL_PRIM_RANDOM_POINT_SIMD(C1)           DlRandom3D((C1), 0, numMp)

#define CALL_PRIM_RANDOM_POINT_SIMD_IF_TRUE(C1, B)   DlRandom3D((C1), (B),  numMp)


#define CALL_PRIM_NOISE_FLOAT_1D(F1, F2) \
{ \
    float temp = (F2); \
    DlFNoise1D(&(F1), &temp, 0, 1); \
}

#define CALL_PRIM_NOISE_FLOAT_2D(F1, F2, F3) \
{ \
    float temp1 = (F2), temp2 = (F3);  \
    DlFNoise2D(&(F1), &temp1, &temp2, 0, 1);  \
} 

#define CALL_PRIM_NOISE_FLOAT_3D(F1, P2)  \
{ \
     DlFNoise3D(&(F1), &(P2), 0, 1); \
}

#define CALL_PRIM_NOISE_FLOAT_4D(F1, P2, F3)  \
{ \
     float temp = (F3);  \
     DlFNoise4D(&(F1), &(P2), &temp, 0, 1);  \
}

#define CALL_PRIM_NOISE_POINT_1D(P1, F2) \
{ \
    float temp = (F2); \
    DlPNoise1D(&(P1), &temp, 0, 1); \
}

#define CALL_PRIM_NOISE_POINT_2D(P1, F2, F3) \
{ \
    float temp1 = (F2), temp2 = (F3);  \
    DlPNoise2D(&(P1), &temp1, &temp2, 0, 1);  \
} 

#define CALL_PRIM_NOISE_POINT_3D(P1, P2)  \
{ \
     DlPNoise3D(&(P1), &(P2), 0, 1); \
}

#define CALL_PRIM_NOISE_POINT_4D(P1, P2, F3)  \
{ \
     float temp = (F3);  \
     DlPNoise4D(&(P1), &(P2), &temp, 0, 1);  \
}



#define CALL_PRIM_NOISE_FLOAT_1D_SIMD(F1, F2) \
{ \
    DlFNoise1D((F1), (F2), 0, numMp); \
}

#define CALL_PRIM_NOISE_FLOAT_2D_SIMD(F1, F2, F3) \
{ \
    DlFNoise2D((F1), (F2), (F3), 0, numMp);  \
} 

#define CALL_PRIM_NOISE_FLOAT_3D_SIMD(F1, P2)  \
{ \
    DlFNoise3D((F1), (P2), 0, numMp); \
}

#define CALL_PRIM_NOISE_FLOAT_4D_SIMD(F1, P2, F3)  \
{ \
    DlFNoise4D((F1), (P2), (F3), 0, numMp);  \
}

#define CALL_PRIM_NOISE_POINT_1D_SIMD(P1, F2) \
{ \
    DlPNoise1D((P1), (F2), 0, numMp); \
}

#define CALL_PRIM_NOISE_POINT_2D_SIMD(P1, F2, F3) \
{ \
    DlPNoise2D((P1), (F2), (F3), 0, numMp);  \
} 

#define CALL_PRIM_NOISE_POINT_3D_SIMD(P1, P2)  \
{ \
     DlPNoise3D((P1), (P2), 0, numMp); \
}

#define CALL_PRIM_NOISE_POINT_4D_SIMD(P1, P2, F3)  \
{ \
     DlPNoise4D((P1), (P2), (F3), 0, numMp);  \
}


#define CALL_PRIM_NOISE_FLOAT_1D_SIMD_IF_TRUE(F1, B, F2) \
{ \
    DlFNoise1D((F1), (F2), (B),  numMp); \
}

#define CALL_PRIM_NOISE_FLOAT_2D_SIMD_IF_TRUE(F1, B, F2, F3) \
{ \
    DlFNoise2D((F1), (F2), (F3), (B),  numMp);  \
} 

#define CALL_PRIM_NOISE_FLOAT_3D_SIMD_IF_TRUE(F1, B, P2)  \
{ \
    DlFNoise3D((F1), (P2), (B),  numMp); \
}

#define CALL_PRIM_NOISE_FLOAT_4D_SIMD_IF_TRUE(F1, B, P2, F3)  \
{ \
    DlFNoise4D((F1), (P2), (F3), (B),  numMp);  \
}

#define CALL_PRIM_NOISE_POINT_1D_SIMD_IF_TRUE(P1, B, F2) \
{ \
    DlPNoise1D((P1), (F2), (B),  numMp); \
}

#define CALL_PRIM_NOISE_POINT_2D_SIMD_IF_TRUE(P1, B, F2, F3) \
{ \
    DlPNoise2D((P1), (F2), (F3), (B),  numMp);  \
} 

#define CALL_PRIM_NOISE_POINT_3D_SIMD_IF_TRUE(P1, B, P2)  \
{ \
     DlPNoise3D((P1), (P2), (B),  numMp); \
}

#define CALL_PRIM_NOISE_POINT_4D_SIMD_IF_TRUE(P1, B, P2, F3)  \
{ \
     DlPNoise4D((P1), (P2), (F3), (B),  numMp);  \
}




#define CALL_PRIM_CELLNOISE_FLOAT_1D(F1, F2) \
{ \
    float temp = (F2); \
    DlFCellnoise1D(&(F1), &temp, 0, 1); \
}

#define CALL_PRIM_CELLNOISE_FLOAT_2D(F1, F2, F3) \
{ \
    float temp1 = (F2), temp2 = (F3);  \
    DlFCellnoise2D(&(F1), &temp1, &temp2, 0, 1);  \
} 

#define CALL_PRIM_CELLNOISE_FLOAT_3D(F1, P2)  \
{ \
     DlFCellnoise3D(&(F1), &(P2), 0, 1); \
}

#define CALL_PRIM_CELLNOISE_FLOAT_4D(F1, P2, F3)  \
{ \
     float temp = (F3);  \
     DlFCellnoise4D(&(F1), &(P2), &temp, 0, 1);  \
}

#define CALL_PRIM_CELLNOISE_POINT_1D(P1, F2) \
{ \
    float temp = (F2); \
    DlPCellnoise1D(&(P1), &temp, 0, 1); \
}

#define CALL_PRIM_CELLNOISE_POINT_2D(P1, F2, F3) \
{ \
    float temp1 = (F2), temp2 = (F3);  \
    DlPCellnoise2D(&(P1), &temp1, &temp2, 0, 1);  \
} 

#define CALL_PRIM_CELLNOISE_POINT_3D(P1, P2)  \
{ \
     DlPCellnoise3D(&(P1), &(P2), 0, 1); \
}

#define CALL_PRIM_CELLNOISE_POINT_4D(P1, P2, F3)  \
{ \
     float temp = (F3);  \
     DlPCellnoise4D(&(P1), &(P2), &temp, 0, 1);  \
}


#define CALL_PRIM_CELLNOISE_FLOAT_1D_SIMD(F1, F2) \
{ \
    DlFCellnoise1D((F1), (F2), 0, numMp); \
}

#define CALL_PRIM_CELLNOISE_FLOAT_2D_SIMD(F1, F2, F3) \
{ \
    DlFCellnoise2D((F1), (F2), (F3), 0, numMp);  \
} 

#define CALL_PRIM_CELLNOISE_FLOAT_3D_SIMD(F1, P2)  \
{ \
     DlFCellnoise3D((F1), (P2), 0, numMp); \
}

#define CALL_PRIM_CELLNOISE_FLOAT_4D_SIMD(F1, P2, F3)  \
{ \
     DlFCellnoise4D((F1), (P2), (F3), 0, numMp);  \
}

#define CALL_PRIM_CELLNOISE_POINT_1D_SIMD(P1, F2) \
{ \
    DlPCellnoise1D((P1), (F2), 0, numMp); \
}

#define CALL_PRIM_CELLNOISE_POINT_2D_SIMD(P1, F2, F3) \
{ \
    DlPCellnoise2D((P1), (F2), (F3), 0, numMp);  \
} 

#define CALL_PRIM_CELLNOISE_POINT_3D_SIMD(P1, P2)  \
{ \
     DlPCellnoise3D((P1), (P2), 0, numMp); \
}

#define CALL_PRIM_CELLNOISE_POINT_4D_SIMD(P1, P2, F3)  \
{ \
     DlPCellnoise4D((P1), (P2), (F3), 0, numMp);  \
}


#define CALL_PRIM_CELLNOISE_FLOAT_1D_SIMD_IF_TRUE(F1, B, F2) \
{ \
    DlFCellnoise1D((F1), (F2), (B), numMp); \
}

#define CALL_PRIM_CELLNOISE_FLOAT_2D_SIMD_IF_TRUE(F1, B, F2, F3) \
{ \
    DlFCellnoise2D((F1), (F2), (F3), (B), numMp);  \
} 

#define CALL_PRIM_CELLNOISE_FLOAT_3D_SIMD_IF_TRUE(F1, B, P2)  \
{ \
     DlFCellnoise3D((F1), (P2), (B), numMp); \
}

#define CALL_PRIM_CELLNOISE_FLOAT_4D_SIMD_IF_TRUE(F1, B, P2, F3)  \
{ \
     DlFCellnoise4D((F1), (P2), (F3), (B), numMp);  \
}

#define CALL_PRIM_CELLNOISE_POINT_1D_SIMD_IF_TRUE(P1, B, F2) \
{ \
    DlPCellnoise1D((P1), (F2), (B), numMp); \
}

#define CALL_PRIM_CELLNOISE_POINT_2D_SIMD_IF_TRUE(P1, B, F2, F3) \
{ \
    DlPCellnoise2D((P1), (F2), (F3), (B), numMp);  \
} 

#define CALL_PRIM_CELLNOISE_POINT_3D_SIMD_IF_TRUE(P1, B, P2)  \
{ \
     DlPCellnoise3D((P1), (P2), (B), numMp); \
}

#define CALL_PRIM_CELLNOISE_POINT_4D_SIMD_IF_TRUE(P1, B, P2, F3)  \
{ \
     DlPCellnoise4D((P1), (P2), (F3), (B), numMp);  \
}



#define CALL_PRIM_PNOISE_FLOAT_1D(F1, F2, P) \
{ \
    float temp = (F2); \
    DlFPnoise1D(&(F1), &temp, (P), 0, 1); \
}

#define CALL_PRIM_PNOISE_FLOAT_2D(F1, F2, F3, UP, VP) \
{ \
    float temp1 = (F2), temp2 = (F3);  \
    DlFPnoise2D(&(F1), &temp1, &temp2, (UP), (VP), 0, 1);  \
} 

#define CALL_PRIM_PNOISE_FLOAT_3D(F1, P2, P)  \
{ \
     DlFPnoise3D(&(F1), &(P2), (P), 0, 1); \
}

#define CALL_PRIM_PNOISE_FLOAT_4D(F1, P2, F3, PP, TP)  \
{ \
     float temp = (F3);  \
     DlFPnoise4D(&(F1), &(P2), &temp, (PP), (TP), 0, 1);  \
}

#define CALL_PRIM_PNOISE_POINT_1D(P1, F2, P) \
{ \
    float temp = (F2); \
    DlPPnoise1D(&(P1), &temp, (P), 0, 1); \
}

#define CALL_PRIM_PNOISE_POINT_2D(P1, F2, F3, UP, VP) \
{ \
    float temp1 = (F2), temp2 = (F3);  \
    DlPPnoise2D(&(P1), &temp1, &temp2, (UP), (VP), 0, 1);  \
} 

#define CALL_PRIM_PNOISE_POINT_3D(P1, P2, P)  \
{ \
     DlPPnoise3D(&(P1), &(P2), (P), 0, 1); \
}

#define CALL_PRIM_PNOISE_POINT_4D(P1, P2, F3, PP, TP)  \
{ \
     float temp = (F3);  \
     DlPPnoise4D(&(P1), &(P2), &temp, (PP), (TP), 0, 1);  \
}


#define CALL_PRIM_PNOISE_FLOAT_1D_SIMD(F1, F2, P) \
{ \
    DlFPnoise1D((F1), (F2), (P), 0, numMp); \
}

#define CALL_PRIM_PNOISE_FLOAT_2D_SIMD(F1, F2, F3, UP, VP) \
{ \
    DlFPnoise2D((F1), (F2), (F3), (UP), (VP), 0, numMp);  \
} 

#define CALL_PRIM_PNOISE_FLOAT_3D_SIMD(F1, P2, P)  \
{ \
     DlFPnoise3D((F1), (P2), (P), 0, numMp); \
}

#define CALL_PRIM_PNOISE_FLOAT_4D_SIMD(F1, P2, F3, PP, TP)  \
{ \
     DlFPnoise4D((F1), (P2), (F3), (PP), (TP), 0, numMp);  \
}

#define CALL_PRIM_PNOISE_POINT_1D_SIMD(P1, F2, P) \
{ \
    DlPPnoise1D((P1), (F2), (P), 0, numMp); \
}

#define CALL_PRIM_PNOISE_POINT_2D_SIMD(P1, F2, F3, UP, VP) \
{ \
    DlPPnoise2D((P1), (F2), (F3), (UP), (VP), 0, numMp);  \
} 

#define CALL_PRIM_PNOISE_POINT_3D_SIMD(P1, P2, P)  \
{ \
     DlPPnoise3D((P1), (P2), (P), 0, numMp); \
}

#define CALL_PRIM_PNOISE_POINT_4D_SIMD(P1, P2, F3, PP, TP)  \
{ \
     DlPPnoise4D((P1), (P2), (F3), (PP), (TP), 0, numMp);  \
}


#define CALL_PRIM_PNOISE_FLOAT_1D_SIMD_IF_TRUE(F1, B, F2, P) \
{ \
    DlFPnoise1D((F1), (F2), (P), (B), numMp); \
}

#define CALL_PRIM_PNOISE_FLOAT_2D_SIMD_IF_TRUE(F1, B, F2, F3, UP, VP) \
{ \
    DlFPnoise2D((F1), (F2), (F3), (UP), (VP), (B), numMp);  \
} 

#define CALL_PRIM_PNOISE_FLOAT_3D_SIMD_IF_TRUE(F1, B, P2, P)  \
{ \
     DlFPnoise3D((F1), (P2), (P), (B), numMp); \
}

#define CALL_PRIM_PNOISE_FLOAT_4D_SIMD_IF_TRUE(F1, B, P2, F3, PP, TP)  \
{ \
     DlFPnoise4D((F1), (P2), (F3), (PP), (TP), (B), numMp);  \
}

#define CALL_PRIM_PNOISE_POINT_1D_SIMD_IF_TRUE(P1, B, F2, P) \
{ \
    DlPPnoise1D((P1), (F2), (P), (B), numMp); \
}

#define CALL_PRIM_PNOISE_POINT_2D_SIMD_IF_TRUE(P1, B, F2, F3, UP, VP) \
{ \
    DlPPnoise2D((P1), (F2), (F3), (UP), (VP), (B), numMp);  \
} 

#define CALL_PRIM_PNOISE_POINT_3D_SIMD_IF_TRUE(P1, B, P2, P)  \
{ \
     DlPPnoise3D((P1), (P2), (P), (B), numMp); \
}

#define CALL_PRIM_PNOISE_POINT_4D_SIMD_IF_TRUE(P1, B, P2, F3, PP, TP)  \
{ \
     DlPPnoise4D((P1), (P2), (F3), (PP), (TP), (B), numMp);  \
}


////////////////////////////////////////////////////////////////////////////
// MACROS FOR GEOMETRIC FUNCTIONS
////////////////////////////////////////////////////////////////////////////
#define CALL_PRIM_XCOMP(F1, P2)     (F1) = (P2)[0]

#define CALL_PRIM_YCOMP(F1, P2)     (F1) = (P2)[1]

#define CALL_PRIM_ZCOMP(F1, P2)     (F1) = (P2)[2]

#define CALL_PRIM_SETXCOMP(P1, F2)  (P1)[0] = (F2)

#define CALL_PRIM_SETYCOMP(P1, F2)  (P1)[1] = (F2)

#define CALL_PRIM_SETZCOMP(P1, F2)  (P1)[2] = (F2)

/* length shadeop */
#define CALL_PRIM_LENGTH(F1, P2) \
	DlLength(&(F1), &(P2), 0, 1)
#define CALL_PRIM_LENGTH_SIMD(F1, P2) \
	DlLength((F1), (P2), 0, numMp)
#define CALL_PRIM_LENGTH_SIMD_IF_TRUE(F1, B2, P2) \
	DlLength((F1), (P2), (B2),  numMp)

#define CALL_PRIM_DISTANCE(F1, P2, P3)       DlDistance(&(F1), &(P2), &(P3), 0, 1)

#define CALL_PRIM_DISTANCE_SIMD(F1, P2, P3)  DlDistance((F1), (P2), (P3), 0, numMp)

#define CALL_PRIM_DISTANCE_SIMD_IF_TRUE(F1, B2, P2, P3)   \
                                          DlDistance((F1), (P2), (P3), (B2),  numMp)


#define CALL_PRIM_AREA_SIMD(F1, P2)      DlArea((F1), (P2), numMpU, numMpV, 0)

#define CALL_PRIM_AREA_SIMD_IF_TRUE(F1, B, P2)      DlArea((F1), (P2), numMpU, numMpV, (B))

/* normalize shadeop */
#define CALL_PRIM_NORMALIZE(P1, P2)      DlNormalize(&(P1), &(P2), 0, 1)

#define CALL_PRIM_NORMALIZE_SIMD(P1, P2) DlNormalize((P1), (P2), 0, numMp)

#define CALL_PRIM_NORMALIZE_SIMD_IF_TRUE(P1, B2, P2) \
                                         DlNormalize((P1), (P2), (B2),  numMp)


/* ptlined shadeop */
#define CALL_PRIM_PTLINED(F, P0, P1, Q) \
	DlPtlined(&(F), (P0), (P1), &(Q), 0, 1)
#define CALL_PRIM_PTLINED_SIMD(F, P0, P1, Q) \
	DlPtlined((F), (P0), (P1), Q, 0, numMp)
#define CALL_PRIM_PTLINED_SIMD_IF_TRUE(F, B, P0, P1, Q) \
	DlPtlined((F), (P0), (P1), Q, (B), numMp)

/* rotate shadeop */
#define CALL_PRIM_ROTATE_POINT(P, Q, ANGLE, P0, P1)  DlRotatePoint(&(P), &(Q), (ANGLE), (P0), (P1), 0, 1)

/* filterstep */
#define CALL_PRIM_FILTERSTEP(RES, E, S1, S2, W, FTR) \
{  \
   float temp1 = (E);  \
   float temp2 = (S1); \
   float temp3 = (S2); \
   DlFilterstep(&(RES),&(temp1),&(temp2),&(temp3),(W),(FTR), 0, 1); \
}

#define CALL_PRIM_FILTERSTEP_SIMD(RES, E, S1, S2, W, FTR)  \
   DlFilterstep((RES),(E),(S1),(S2),(W),(FTR),0,numMp)

#define CALL_PRIM_FILTERSTEP_SIMD_IF_TRUE(RES, B, E, S1, S2, W, FTR)  \
   DlFilterstep((RES),(E),(S1),(S2),(W),(FTR),(B),numMp)

#define COMPUTE_FILTERSTEP_POINTS(S1, S2)  \
   DlMakeFilterstepPoints((S1), (S2), numMpU, numMpV, 0)

#define COMPUTE_FILTERSTEP_POINTS_IF_TRUE(S1, S2, B)  \
   DlMakeFilterstepPoints((S1), (S2), numMpU, numMpV, B)


////////////////////////////////////////////////////////////////////////////
// Faceforward
////////////////////////////////////////////////////////////////////////////

#define CALL_PRIM_FACEFORWARD(P1, P2, P3)      \
	DlFaceforward(&(P1), &(P2), &(P3), \
            	   &(i_hiddens.GetNg()[polyCnt]), 0, 1)

#define CALL_PRIM_FACEFORWARD_SIMD(P1, P2, P3) \
	DlFaceforward((P1), (P2), (P3), \
		i_hiddens.GetNg(), 0, numMp)

#define CALL_PRIM_FACEFORWARD_SIMD_IF_TRUE(P1, B2, P2, P3) \
	DlFaceforward((P1), (P2), (P3), \
		i_hiddens.GetNg(), (B2),  numMp)



#define CALL_PRIM_FACEFORWARD2(P1, P2, P3, P4)      \
	DlFaceforward(&(P1), &(P2), &(P3), \
		&(P4), 0, 1)

#define CALL_PRIM_FACEFORWARD2_SIMD(P1, P2, P3, P4) \
	DlFaceforward((P1), (P2), (P3), \
		(P4), 0, numMp)

#define CALL_PRIM_FACEFORWARD2_SIMD_IF_TRUE(P1, B2, P2, P3, P4) \
	DlFaceforward((P1), (P2), (P3), \
		(P4), (B2),  numMp)


#define CALL_PRIM_REFLECT(P1, P2, P3) DlReflect(&(P1), &(P2), &(P3), 0, 1)

#define CALL_PRIM_REFLECT_SIMD(P1, P2, P3) DlReflect((P1), (P2), (P3), 0, numMp)

#define CALL_PRIM_REFLECT_SIMD_IF_TRUE(P1, B, P2, P3) \
                                      DlReflect((P1), (P2), (P3), (B),  numMp)



#define CALL_PRIM_REFRACT(P1, P2, P3, F4) DlRefract(&(P1), &(P2), &(P3), (F4), 0, 1)

#define CALL_PRIM_REFRACT_SIMD(P1, P2, P3, F4) \
                               DlRefract((P1), (P2), (P3), (F4), 0, numMp)

#define CALL_PRIM_REFRACT_SIMD_IF_TRUE(P1, B, P2, P3, F4) \
                               DlRefract((P1), (P2), (P3), (F4), (B),  numMp)



#define CALL_PRIM_FRESNEL5(P1, P2, F3, F4, F5) \
                               DlFresnel(&(P1), &(P2), (F3), &(F4), &(F5), 0, 0, 0, 1)

#define CALL_PRIM_FRESNEL5_SIMD(P1, P2, F3, F4, F5) \
                               DlFresnel((P1), (P2), (F3), (F4), (F5), 0, 0, 0, numMp)

#define CALL_PRIM_FRESNEL5_SIMD_IF_TRUE(B, P1, P2, F3, F4, F5) \
                         DlFresnel((P1), (P2), (F3), (F4), (F5), 0, 0, (B),  numMp)



#define CALL_PRIM_FRESNEL7(P1, P2, F3, F4, F5, P6, P7) \
                          DlFresnel(&(P1), &(P2), (F3), &(F4), &(F5), &(P6), &(P7), 0, 1)

#define CALL_PRIM_FRESNEL7_SIMD(P1, P2, F3, F4, F5, P6, P7) \
                          DlFresnel((P1), (P2), (F3), (F4), (F5), (P6), (P7), 0, numMp)

#define CALL_PRIM_FRESNEL7_SIMD_IF_TRUE(B, P1, P2, F3, F4, F5, P6, P7) \
                    DlFresnel((P1), (P2), (F3), (F4), (F5), (P6), (P7), (B),  numMp)



////////////////////////////////////////////////////////////////////////////
// transform, vtransform and ntransform
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////
// Version 4
//
// Spaces are not known at compile-time.
////////////////////////////////////////////////////////

#define FIND_SPACE(S)  \
     (!strcmp(S, "object") ? i_context.SpaceList()[MATRIX_object] : \
        (!strcmp(S, "shader") ? i_context.SpaceList()[MATRIX_shader] : FindSpace(i_context.SpaceRegistry(), S)))

#define CALL_PRIM_TRANSFORM(P1, S2, S3, P4) \
     { \
         const DlFrame3D *fromSpace = FIND_SPACE(S2); \
         const DlFrame3D *toSpace = FIND_SPACE(S3); \
         DlTransformSpace(&(P1), &(P4), fromSpace, 0, DlTokenInfo::e_RtPoint, false, 0, 1); \
         DlTransformSpace(&(P1), &(P1), toSpace, 0, DlTokenInfo::e_RtPoint, true, 0, 1); \
     }

#define CALL_PRIM_VTRANSFORM(P1, S2, S3, P4) \
     { \
         const DlFrame3D *fromSpace = FIND_SPACE(S2); \
         const DlFrame3D *toSpace = FIND_SPACE(S3); \
         DlTransformSpace(&(P1), &(P4), fromSpace, 0, DlTokenInfo::e_RtVector, false, 0, 1); \
         DlTransformSpace(&(P1), &(P1), toSpace, 0, DlTokenInfo::e_RtVector, true, 0, 1); \
     }

#define CALL_PRIM_NTRANSFORM(P1, S2, S3, P4) \
     { \
         const DlFrame3D *fromSpace = FIND_SPACE(S2); \
         const DlFrame3D *toSpace = FIND_SPACE(S3); \
         DlTransformSpace(&(P1), &(P4), fromSpace, 0, DlTokenInfo::e_RtNormal, false, 0, 1); \
         DlTransformSpace(&(P1), &(P1), toSpace, 0, DlTokenInfo::e_RtNormal, true, 0, 1); \
     }

#define CALL_PRIM_TRANSFORM_SIMD(P1, S2, S3, P4) \
     { \
         const DlFrame3D *fromSpace = FIND_SPACE(S2); \
         const DlFrame3D *toSpace = FIND_SPACE(S3); \
         DlTransformSpace((P1), (P4), fromSpace, 0, DlTokenInfo::e_RtPoint, false, 0, numMp); \
         DlTransformSpace((P1), (P1), toSpace, 0, DlTokenInfo::e_RtPoint, true, 0, numMp); \
     }

#define CALL_PRIM_VTRANSFORM_SIMD(P1, S2, S3, P4) \
     { \
         const DlFrame3D *fromSpace = FIND_SPACE(S2); \
         const DlFrame3D *toSpace = FIND_SPACE(S3); \
         DlTransformSpace((P1), (P4), fromSpace, 0, DlTokenInfo::e_RtVector, false, 0, numMp); \
         DlTransformSpace((P1), (P1), toSpace, 0, DlTokenInfo::e_RtVector, true, 0, numMp); \
     }

#define CALL_PRIM_NTRANSFORM_SIMD(P1, S2, S3, P4) \
     { \
         const DlFrame3D *fromSpace = FIND_SPACE(S2); \
         const DlFrame3D *toSpace = FIND_SPACE(S3); \
         DlTransformSpace((P1), (P4), fromSpace, 0, DlTokenInfo::e_RtNormal, false, 0, numMp); \
         DlTransformSpace((P1), (P1), toSpace, 0, DlTokenInfo::e_RtNormal, true, 0, numMp); \
     }

#define CALL_PRIM_TRANSFORM_SIMD_IF_TRUE(P1, B2, S2, S3, P4) \
     { \
         const DlFrame3D *fromSpace = FIND_SPACE(S2); \
         const DlFrame3D *toSpace = FIND_SPACE(S3); \
         DlTransformSpace((P1), (P4), fromSpace, 0, DlTokenInfo::e_RtPoint, false,  (B2), numMp); \
         DlTransformSpace((P1), (P1), toSpace, 0, DlTokenInfo::e_RtPoint, true,  (B2), numMp); \
     }

#define CALL_PRIM_VTRANSFORM_SIMD_IF_TRUE(P1, B2, S2, S3, P4) \
     { \
         const DlFrame3D *fromSpace = FIND_SPACE(S2); \
         const DlFrame3D *toSpace = FIND_SPACE(S3); \
         DlTransformSpace((P1), (P4), fromSpace, 0, DlTokenInfo::e_RtVector, false,  (B2), numMp); \
         DlTransformSpace((P1), (P1), toSpace, 0, DlTokenInfo::e_RtVector, true,  (B2), numMp); \
     }

#define CALL_PRIM_NTRANSFORM_SIMD_IF_TRUE(P1, B2, S2, S3, P4) \
     { \
         const DlFrame3D *fromSpace = FIND_SPACE(S2); \
         const DlFrame3D *toSpace = FIND_SPACE(S3); \
         DlTransformSpace((P1), (P4), fromSpace, 0, DlTokenInfo::e_RtNormal, false,  (B2), numMp); \
         DlTransformSpace((P1), (P1), toSpace, 0, DlTokenInfo::e_RtNormal, true,  (B2), numMp); \
     }


#define CALL_PRIM_TRANSFORM2(P1, S3, P4) \
     { \
         const DlFrame3D *toSpace = FIND_SPACE(S3); \
         DlTransformSpace(&(P1), &(P4), toSpace, 0, DlTokenInfo::e_RtPoint, true, 0, 1); \
     }

#define CALL_PRIM_VTRANSFORM2(P1, S3, P4) \
     { \
         const DlFrame3D *toSpace = FIND_SPACE(S3); \
         DlTransformSpace(&(P1), &(P4), toSpace, 0, DlTokenInfo::e_RtVector, true, 0, 1); \
     }

#define CALL_PRIM_NTRANSFORM2(P1, S3, P4) \
     { \
         const DlFrame3D *toSpace = FIND_SPACE(S3); \
         DlTransformSpace(&(P1), &(P4), toSpace, 0, DlTokenInfo::e_RtNormal, true, 0, 1); \
     }

#define CALL_PRIM_TRANSFORM2_SIMD(P1, S3, P4) \
     { \
         const DlFrame3D *toSpace = FIND_SPACE(S3); \
         DlTransformSpace((P1), (P4), toSpace, 0, DlTokenInfo::e_RtPoint, true, 0, numMp); \
     }

#define CALL_PRIM_VTRANSFORM2_SIMD(P1, S3, P4) \
     { \
         const DlFrame3D *toSpace = FIND_SPACE(S3); \
         DlTransformSpace((P1), (P4), toSpace, 0, DlTokenInfo::e_RtVector, true, 0, numMp); \
     }

#define CALL_PRIM_NTRANSFORM2_SIMD(P1, S3, P4) \
     { \
         const DlFrame3D *toSpace = FIND_SPACE(S3); \
         DlTransformSpace((P1), (P4), toSpace, 0, DlTokenInfo::e_RtNormal, true, 0, numMp); \
     }

#define CALL_PRIM_TRANSFORM2_SIMD_IF_TRUE(P1, B2, S3, P4) \
     { \
         const DlFrame3D *toSpace = FIND_SPACE(S3); \
         DlTransformSpace((P1), (P4), toSpace, 0, DlTokenInfo::e_RtPoint, true,  (B2), numMp); \
     }

#define CALL_PRIM_VTRANSFORM2_SIMD_IF_TRUE(P1, B2, S3, P4) \
     { \
         const DlFrame3D *toSpace = FIND_SPACE(S3); \
         DlTransformSpace((P1), (P4), toSpace, 0, DlTokenInfo::e_RtVector, true,  (B2), numMp); \
     }

#define CALL_PRIM_NTRANSFORM2_SIMD_IF_TRUE(P1, B2, S3, P4) \
     { \
         const DlFrame3D *toSpace = FIND_SPACE(S3); \
         DlTransformSpace((P1), (P4), toSpace, 0, DlTokenInfo::e_RtNormal, true,  (B2), numMp); \
     }


////////////////////////////////////////////////////////////////////////////
// Transform functions with matrices
////////////////////////////////////////////////////////////////////////////

#define CALL_PRIM_TRANSFORM_MATRIX(P1, S2, M, P4) \
     { \
         const DlFrame3D *fromSpace = FIND_SPACE(S2); \
         DlTransformSpace(&(P1), &(P4), fromSpace, 0, DlTokenInfo::e_RtPoint, false, 0, 1); \
         DlTransformSpace(&(P1), &(P1), 0, &(M), DlTokenInfo::e_RtPoint, true, 0, 1); \
     }

#define CALL_PRIM_VTRANSFORM_MATRIX(P1, S2, M, P4) \
     { \
         const DlFrame3D *fromSpace = FIND_SPACE(S2); \
         DlTransformSpace(&(P1), &(P4), fromSpace, 0, DlTokenInfo::e_RtVector, false, 0, 1); \
         DlTransformSpace(&(P1), &(P1), 0, &(M), DlTokenInfo::e_RtVector, true, 0, 1); \
     }

#define CALL_PRIM_NTRANSFORM_MATRIX(P1, S2, M, P4) \
     { \
         const DlFrame3D *fromSpace = FIND_SPACE(S2); \
         DlTransformSpace(&(P1), &(P4), fromSpace, 0, DlTokenInfo::e_RtNormal, false, 0, 1); \
         DlTransformSpace(&(P1), &(P1), 0, &(M), DlTokenInfo::e_RtNormal, true, 0, 1); \
     }

#define CALL_PRIM_TRANSFORM_MATRIX_SIMD(P1, S2, M, P4) \
     { \
         const DlFrame3D *fromSpace = FIND_SPACE(S2); \
         DlTransformSpace((P1), (P4), fromSpace, 0, DlTokenInfo::e_RtPoint, false, 0, numMp); \
         DlTransformSpace((P1), (P1), 0, &(M), DlTokenInfo::e_RtPoint, true, 0, numMp); \
     }

#define CALL_PRIM_VTRANSFORM_MATRIX_SIMD(P1, S2, M, P4) \
     { \
         const DlFrame3D *fromSpace = FIND_SPACE(S2); \
         DlTransformSpace((P1), (P4), fromSpace, 0, DlTokenInfo::e_RtVector, false, 0, numMp); \
         DlTransformSpace((P1), (P1), 0, &(M), DlTokenInfo::e_RtVector, true, 0, numMp); \
     }

#define CALL_PRIM_NTRANSFORM_MATRIX_SIMD(P1, S2, M, P4) \
     { \
         const DlFrame3D *fromSpace = FIND_SPACE(S2); \
         DlTransformSpace((P1), (P4), fromSpace, 0, DlTokenInfo::e_RtNormal, false, 0, numMp); \
         DlTransformSpace((P1), (P1), 0, &(M), DlTokenInfo::e_RtNormal, true, 0, numMp); \
     }

#define CALL_PRIM_TRANSFORM_MATRIX_SIMD_IF_TRUE(P1, B, S2, M, P4) \
     { \
         const DlFrame3D *fromSpace = FIND_SPACE(S2); \
         DlTransformSpace((P1), (P4), fromSpace, 0, DlTokenInfo::e_RtPoint, false, (B),  numMp); \
         DlTransformSpace((P1), (P1), 0, &(M), DlTokenInfo::e_RtPoint, true, (B),  numMp); \
     }

#define CALL_PRIM_VTRANSFORM_MATRIX_SIMD_IF_TRUE(P1, B, S2, M, P4) \
     { \
         const DlFrame3D *fromSpace = FIND_SPACE(S2); \
         DlTransformSpace((P1), (P4), fromSpace, 0, DlTokenInfo::e_RtVector, false, (B),  numMp); \
         DlTransformSpace((P1), (P1), 0, &(M), DlTokenInfo::e_RtVector, true, (B),  numMp); \
     }

#define CALL_PRIM_NTRANSFORM_MATRIX_SIMD_IF_TRUE(P1, B, S2, M, P4) \
     { \
         const DlFrame3D *fromSpace = FIND_SPACE(S2); \
         DlTransformSpace((P1), (P4), fromSpace, 0, DlTokenInfo::e_RtNormal, false, (B),  numMp); \
         DlTransformSpace((P1), (P1), 0, &(M), DlTokenInfo::e_RtNormal, true, (B),  numMp); \
     }



#define CALL_PRIM_TRANSFORM2_MATRIX(P1, M, P2) \
     { \
         DlTransformSpace(&(P1), &(P2), 0, &(M), DlTokenInfo::e_RtPoint, true, 0, 1); \
     }

#define CALL_PRIM_VTRANSFORM2_MATRIX(P1, M, P2) \
     { \
         DlTransformSpace(&(P1), &(P2), 0, &(M), DlTokenInfo::e_RtVector, true, 0, 1); \
     }

#define CALL_PRIM_NTRANSFORM2_MATRIX(P1, M, P2) \
     { \
         DlTransformSpace(&(P1), &(P2), 0, &(M), DlTokenInfo::e_RtNormal, true, 0, 1); \
     }

#define CALL_PRIM_TRANSFORM2_MATRIX_SIMD(P1, M, P2) \
     { \
         DlTransformSpace((P1), (P2), 0, &(M), DlTokenInfo::e_RtPoint, true, 0, numMp); \
     }

#define CALL_PRIM_VTRANSFORM2_MATRIX_SIMD(P1, M, P2) \
     { \
         DlTransformSpace((P1), (P2), 0, &(M), DlTokenInfo::e_RtVector, true, 0, numMp); \
     }

#define CALL_PRIM_NTRANSFORM2_MATRIX_SIMD(P1, M, P2) \
     { \
         DlTransformSpace((P1), (P2), 0, &(M), DlTokenInfo::e_RtNormal, true, 0, numMp); \
     }

#define CALL_PRIM_TRANSFORM2_MATRIX_SIMD_IF_TRUE(P1, B, M, P2) \
     { \
         DlTransformSpace((P1), (P2), 0, &(M), DlTokenInfo::e_RtPoint, true, (B),  numMp); \
     }

#define CALL_PRIM_VTRANSFORM2_MATRIX_SIMD_IF_TRUE(P1, B, M, P2) \
     { \
         DlTransformSpace((P1), (P2), 0, &(M), DlTokenInfo::e_RtVector, true, (B),  numMp); \
     }

#define CALL_PRIM_NTRANSFORM2_MATRIX_SIMD_IF_TRUE(P1, B, M, P2) \
     { \
         DlTransformSpace((P1), (P2), 0, &(M), DlTokenInfo::e_RtNormal, true, (B),  numMp); \
     }



////////////////////////////////////////////////////////////////////////////
// 
////////////////////////////////////////////////////////////////////////////

#define CALL_PRIM_DEPTH(F1, P2)   DlDepth(&(F1), &(P2), i_context, 0, 1)

#define CALL_PRIM_DEPTH_SIMD(F1, P2)   DlDepth((F1), (P2), i_context, 0, numMp)

#define CALL_PRIM_DEPTH_SIMD_IF_TRUE(F1, B, P2)   \
	DlDepth((F1), (P2), i_context, (B),  numMp)

#define CALL_PRIM_CALCULATENORMAL_SIMD(P1, P2)  \
	DlCalculateNormal( \
		(P1), (P2), numMpU, numMpV, 0, i_context )

#define CALL_PRIM_CALCULATENORMAL_SIMD_IF_TRUE( P1, B, P2 )  \
		DlCalculateNormal((P1), (P2), numMpU, numMpV, (B), i_context )


////////////////////////////////////////////////////////////////////////////
// MACROS FOR COMPONENT ACCESS FUNCTIONS (matrices, colors, points)
////////////////////////////////////////////////////////////////////////////
/*
	NOTE:
	There is no need to do the following:
	"
	int index = int(F2);
	if( index>=0 && index<=2 )
		...
	"

	Since we can do a more optimal:

	"
	unsigned index = unsigned(F2);
	if( index <= 2 )
		...
	"

	This is valid since negative floats will 'wrap' around and become large
	unsigned values.
*/
#define CALL_PRIM_COMP(F1, C2, F3) \
{ \
	unsigned index = unsigned(F3); \
	if( index <= 2 ) \
	{ \
		(F1) = *(((float*)(&(C2))) + index); \
	} \
}

#define CALL_PRIM_SETCOMP(C1, F2, F3) \
{ \
    unsigned index = unsigned(F2); \
	if( index <= 2 ) \
	{ \
		*(((float*)(&(C1))) + index) = (F3); \
	} \
}

#define CALL_PRIM_COMP_MATRIX(F1, M, R, C)  \
{ \
    unsigned row = unsigned(R); \
	unsigned col = unsigned(C); \
    if( row <= 3 && col <= 3) \
    { \
        (F1) = (M)[ row*4 + col ]; \
    } \
}

#define CALL_PRIM_SETCOMP_MATRIX(M, R, C, F) \
{ \
    unsigned row = unsigned( R ); \
	unsigned col = unsigned( C ); \
    if( row <= 3 && col <= 3)  \
    { \
        (M)[row*4+col] = (F); \
    } \
}

#define CALL_PRIM_MIX_1D(C1, F2, F3, F4)  \
{ \
    float temp1 = (F2);  float temp2 = (F3);  \
    DlMix1D(&(C1), &(temp1), &(temp2), (F4), 0, 1); \
}

#define CALL_PRIM_MIX_1D_SIMD(C1, C2, C3, F4)  \
	DlMix1D((C1), (C2), (C3), (F4), 0, numMp)

#define CALL_PRIM_MIX_1D_SIMD_IF_TRUE(C1, B, C2, C3, F4)  \
	DlMix1D((C1), (C2), (C3), (F4), (B),  numMp)

#define CALL_PRIM_MIX_3D(C1, C2, C3, F4) \
	DlMix3D(&(C1), &(C2), &(C3), (F4), 0, 1)

#define CALL_PRIM_MIX_3D_SIMD(C1, C2, C3, F4)  \
	DlMix3D((C1), (C2), (C3), (F4), 0, numMp)

#define CALL_PRIM_MIX_3D_SIMD_IF_TRUE(C1, B, C2, C3, F4)  \
	DlMix3D((C1), (C2), (C3), (F4), (B),  numMp)


#define CALL_PRIM_CTRANSFORM_FROM_RGB(C1, C2, S) \
	DlTransformColor(&(C1), &(C2), (S), true, 0, 1)

#define CALL_PRIM_CTRANSFORM_FROM_RGB_SIMD(C1, C2, S) \
	DlTransformColor((C1), (C2), (S), true, 0, numMp)

#define CALL_PRIM_CTRANSFORM_FROM_RGB_SIMD_IF_TRUE(C1, B, C2, S) \
	DlTransformColor((C1), (C2), (S), true, (B),  numMp)



#define CALL_PRIM_CTRANSFORM_TO_RGB(C1, C2, S) \
	DlTransformColor(&(C1), &(C2), (S), false, 0, 1)

#define CALL_PRIM_CTRANSFORM_TO_RGB_SIMD(C1, C2, S) \
	DlTransformColor((C1), (C2), (S), false, 0, numMp)

#define CALL_PRIM_CTRANSFORM_TO_RGB_SIMD_IF_TRUE(C1, B, C2, S) \
	DlTransformColor((C1), (C2), (S), false, (B),  numMp)


#define CALL_PRIM_CTRANSFORM(C1, S1, S2, C2) \
{ \
     CALL_PRIM_CTRANSFORM_TO_RGB(C1, C2, S1); \
     CALL_PRIM_CTRANSFORM_FROM_RGB(C1, C1, S2);  \
}

#define CALL_PRIM_CTRANSFORM_SIMD(C1, S1, S2, C2) \
{ \
     CALL_PRIM_CTRANSFORM_TO_RGB_SIMD(C1, C2, S1); \
     CALL_PRIM_CTRANSFORM_FROM_RGB_SIMD(C1, C1, S2);  \
}

#define CALL_PRIM_CTRANSFORM_SIMD_IF_TRUE(C1, B, S1, S2, C2) \
{ \
     CALL_PRIM_CTRANSFORM_TO_RGB_SIMD_IF_TRUE(C1, B, C2, S1); \
     CALL_PRIM_CTRANSFORM_FROM_RGB_SIMD_IF_TRUE(C1, B, C1, S2);  \
}


#define CALL_PRIM_CTRANSFORM2(C1, S1, C2) \
{ \
     CALL_PRIM_CTRANSFORM_FROM_RGB(C1, C2, S1);  \
}

#define CALL_PRIM_CTRANSFORM2_SIMD(C1, S1, C2) \
{ \
     CALL_PRIM_CTRANSFORM_FROM_RGB_SIMD(C1, C2, S1);  \
}

#define CALL_PRIM_CTRANSFORM2_SIMD_IF_TRUE(C1, B, S1, C2) \
{ \
     CALL_PRIM_CTRANSFORM_FROM_RGB_SIMD_IF_TRUE(C1, B, C2, S1);  \
}



////////////////////////////////////////////////////////////////////////////
// MACROS FOR SHADING AND LIGHTING FUNCTIONS
////////////////////////////////////////////////////////////////////////////

#define CALL_PRIM_AMBIENT(C1) \
	DlAmbient(&(C1), i_numLights, i_lights, 0, 1, polyCnt )

#define CALL_PRIM_AMBIENT_SIMD(C1) \
	DlAmbient((C1),i_numLights, i_lights, 0, numMp, 0 )

#define CALL_PRIM_AMBIENT_SIMD_IF_TRUE(C1, B2) \
	DlAmbient((C1), i_numLights, i_lights, (B2), numMp, 0)

#define CALL_PRIM_DIFFUSE(C1, P2) \
	DlDiffuse(&(C1), &(P2), i_numLights, i_lights, 0, 1, polyCnt)

#define CALL_PRIM_DIFFUSE_SIMD(C1, P2) \
	DlDiffuse((C1), (P2), i_numLights, i_lights, 0, numMp, 0)

#define CALL_PRIM_DIFFUSE_SIMD_IF_TRUE(C1, B3, P2) \
	DlDiffuse((C1), (P2), i_numLights, i_lights, (B3), numMp, 0 )


#define CALL_PRIM_SPECULAR(C1, P2, P3, F4) \
	DlSpecular(&(C1), &(P2),\
		&(P3), (F4), \
		i_numLights, \
		i_lights,    \
		0, 1, polyCnt)

#define CALL_PRIM_SPECULAR_SIMD(C1, P2, P3, F4) \
	DlSpecular((C1), (P2),\
		(P3), (F4), \
		i_numLights, \
		i_lights,    \
		0, numMp, 0)

#define CALL_PRIM_SPECULAR_SIMD_IF_TRUE(C1, B2, P2, P3, F4) \
	DlSpecular((C1), (P2),\
		(P3), (F4), \
		i_numLights, \
		i_lights,    \
		(B2),  \
		numMp, 0)

#define CALL_PRIM_SPECULARSTD(C1, P2, P3, F4) \
	DlSpecularstd(&(C1), &(P2),\
	&(P3), (F4), \
	i_numLights, \
	i_lights,    \
	0, 1, polyCnt)

#define CALL_PRIM_SPECULARSTD_SIMD(C1, P2, P3, F4) \
	DlSpecularstd((C1), (P2),\
	(P3), (F4), \
	i_numLights, \
	i_lights,    \
	0, numMp, 0)

#define CALL_PRIM_SPECULARSTD_SIMD_IF_TRUE(C1, B2, P2, P3, F4) \
	DlSpecularstd((C1), (P2),\
		(P3), (F4), \
		i_numLights, \
		i_lights,    \
		(B2),  \
		numMp, 0)


#define CALL_PRIM_SPECULARBRDF(C, L, N, V, R) \
	DlSpecularbrdf(&(C), &(L), \
		&(N), &(V), \
		(R),       \
		0, 1)

#define CALL_PRIM_SPECULARBRDF_SIMD(C, L, N, V, R) \
	DlSpecularbrdf((C), (L), \
		(N), (V), \
		(R),      \
		0, numMp)

#define CALL_PRIM_SPECULARBRDF_SIMD_IF_TRUE(C, B, L, N, V, R) \
	DlSpecularbrdf((C), (L), \
		(N), (V), \
		(R),      \
		(B),  numMp )



#define CALL_PRIM_PHONG(C1, P2, P3, F4)  DlPhong(&(C1), &(P2),\
                                                              &(P3), (F4), \
                                                              i_numLights, \
                                                              i_lights,    \
                                                              0, 1, polyCnt)

#define CALL_PRIM_PHONG_SIMD(C1, P2, P3, F4)  DlPhong((C1), (P2),\
                                                                   (P3), (F4), \
                                                                   i_numLights, \
                                                                   i_lights,    \
                                                                   0, numMp, 0)


#define CALL_PRIM_PHONG_SIMD_IF_TRUE(C1, B, P2, P3, F4)  DlPhong((C1), (P2),\
                                                                              (P3), (F4), \
                                                                              i_numLights, \
                                                                              i_lights,    \
                                                                              (B),  numMp, 0)



/* ------------------------------------------------------ R A Y    T R A C I N G
*/
#define CALL_PRIM_TRACE(C1, P2, P3, o_dist )                      \
	DlTrace(&(C1), READ_FLOAT_REF(o_dist),        \
		i_context.RayInfos(),       \
		&(P2), &(P3),               \
		0x0, \
		*(i_context.Eye()),         \
		0, 1, 1, shaderSupport) 

#define CALL_PRIM_TRACE_SIMD(C1, P2, P3, o_dist )       \
	DlTrace((C1), READ_FLOAT_REF_SIMD(o_dist),          \
		i_context.RayInfos(),        \
		(P2), (P3),                 \
		0x0,           \
		*(i_context.Eye()),         \
		0, numMpU, numMpV, shaderSupport )


#define CALL_PRIM_TRACE_SIMD_IF_TRUE(C1, B, P2, P3, o_dist ) \
	DlTrace((C1), READ_FLOAT_REF_SIMD(o_dist), \
		i_context.RayInfos(), \
		(P2), (P3), \
		0x0, \
		*(i_context.Eye()), \
		(B), numMpU, numMpV, shaderSupport )

#define CALL_PRIM_TRACE2(C1, P2, P3, o_dist, I ) \
	DlTrace(&(C1), READ_FLOAT_REF(o_dist),        \
		i_context.RayInfos(),        \
		&(P2), &(P3),               \
		&(I),                       \
		*(i_context.Eye()),         \
		0, 1, 1, shaderSupport)

#define CALL_PRIM_TRACE2_SIMD( C1, P2, P3, o_dist, I ) \
	DlTrace((C1), READ_FLOAT_REF_SIMD(o_dist), \
		i_context.RayInfos(), \
		(P2), (P3), \
		(I), \
		*(i_context.Eye()), \
		0, numMpU, numMpV, shaderSupport )

#define CALL_PRIM_TRACE2_SIMD_IF_TRUE(C1, B, P2, P3, o_dist, I )  \
	DlTrace((C1), READ_FLOAT_REF_SIMD(o_dist), \
		i_context.RayInfos(), \
		(P2), (P3),                 \
		(I),                        \
		*(i_context.Eye()),         \
		(B), numMpU, numMpV, shaderSupport )

#define CALL_PRIM_OCCLUSION(Ret, P, R, Samples, A, Bias, MaxHitDist, Variance, Axis, Nun ) \
	DlOcclusion( \
		&(Ret),			\
		&(P), &(R), (A), READ_POINT_REF(Nun), (Samples), (Bias), (MaxHitDist), \
		(Variance), READ_POINT_REF(Axis), \
		0, 1, 1, i_context.RayInfos(), shaderSupport)

#define CALL_PRIM_OCCLUSION_SIMD(Ret, P, R, Samples, A, Bias, MaxHitDist, Variance, Axis, Nun ) \
		DlOcclusion( \
			(Ret),			\
			(P), (R), (A), READ_POINT_REF_SIMD(Nun), (Samples), (Bias), (MaxHitDist), (Variance), READ_POINT_REF_SIMD(Axis), \
			0, numMpU, numMpV, i_context.RayInfos(), shaderSupport)

#define CALL_PRIM_OCCLUSION_SIMD_IF_TRUE(Ret, B, P, R, Samples, A, Bias, MaxHitDist, Variance, Axis, Nun ) \
		DlOcclusion( \
			(Ret), \
			(P), (R), (A), READ_POINT_REF_SIMD(Nun), (Samples), (Bias), (MaxHitDist), (Variance), READ_POINT_REF_SIMD(Axis), \
			(B), numMpU, numMpV, i_context.RayInfos(), shaderSupport )


#define CALL_PRIM_INDIRECTDIFFUSE( \
	Ret, P, R, Samples, A, Bias, MaxHitDist, Variance, EnvMap, Axis, Nun, Occ ) \
	DlIndirectDiffuse( \
		&(Ret), \
		&(P), &(R), (A), (Samples), (Bias), (MaxHitDist), (Variance), (EnvMap), \
		READ_POINT_REF(Axis), READ_POINT_REF(Nun), READ_FLOAT_REF(Occ), \
		0, 1, 1, i_context.RayInfos(), shaderSupport)

#define CALL_PRIM_INDIRECTDIFFUSE_SIMD( \
	Ret, P, R, Samples, A, Bias, MaxHitDist, Variance, EnvMap, Axis, Nun, Occ ) \
	DlIndirectDiffuse( \
		(Ret), \
		(P), (R), (A), (Samples), (Bias), (MaxHitDist), (Variance), (EnvMap), \
		READ_POINT_REF_SIMD(Axis), READ_POINT_REF_SIMD(Nun),  READ_FLOAT_REF_SIMD(Occ), \
		0, numMpU, numMpV, i_context.RayInfos(), shaderSupport)

#define CALL_PRIM_INDIRECTDIFFUSE_SIMD_IF_TRUE( \
	Ret, B, P, R, Samples, A, Bias, MaxHitDist, Variance, EnvMap, Axis, Nun, Occ ) \
	DlIndirectDiffuse( \
		(Ret), \
		(P), (R), (A), (Samples), (Bias), (MaxHitDist), (Variance), (EnvMap), \
		READ_POINT_REF_SIMD(Axis),  READ_POINT_REF_SIMD(Nun), READ_FLOAT_REF_SIMD(Occ), \
		(B), numMpU, numMpV, i_context.RayInfos(), shaderSupport )

////////////////////////////////////////////////////////////////////////////
// MACROS FOR FUNCTIONS ON STRINGS
////////////////////////////////////////////////////////////////////////////

#define CALL_PRIM_PRINTF(ARGS, NB)    DlPrintf(ARGS, NB)

#define CALL_PRIM_CONCAT(S, ARGS, NB)  DlConcat(&(S), ARGS, NB)

#define CALL_PRIM_FORMAT(S, ARGS, NB)  DlFormat(&(S), ARGS, NB)

#define CALL_PRIM_MATCH(F, S1, S2)     DlMatch(&(F), (S1), (S2))

////////////////////////////////////////////////////////////////////////////
// MACROS FOR LIGHT STATEMENTS
////////////////////////////////////////////////////////////////////////////

#define EVAL_LIGHTS(P1) \
        DlEvalLightShaders(numMpU, numMpV, \
        (P1), i_lights, i_numLights, i_context, shaderSupport )

#define ILLUMINATE(RES, P, AXIS, ANGLE, PS, L)  \
        DlIlluminate(&(RES), (P), &(AXIS), (ANGLE), &(PS), &(L), 0, 1, i_shadows, \
                       i_context.RayInfos(), shaderSupport )

#define ILLUMINATE_SIMD(RES, P, AXIS, ANGLE, PS, L)  \
        DlIlluminate((RES), (P), &(AXIS), (ANGLE), (PS), (L), 0, numMp, i_shadows, \
                       i_context.RayInfos(), shaderSupport )

#define ILLUMINATE_SIMD_IF_TRUE(RES, B, P, AXIS, ANGLE, PS, L)  \
        DlIlluminate((RES), (P), &(AXIS), (ANGLE), (PS), (L), (B),  numMp, i_shadows, \
                       i_context.RayInfos(), shaderSupport)

#define ILLUMINATE_NO_CONE(RES, P, PS, L)  \
        DlIlluminate(&(RES), (P), 0, 0, &(PS), &(L), 0, 1, i_shadows, \
                       i_context.RayInfos(), shaderSupport )

#define ILLUMINATE_NO_CONE_SIMD(RES, P, PS, L)  \
        DlIlluminate((RES), (P), 0, 0, (PS), (L), 0, numMp, i_shadows, \
                       i_context.RayInfos(), shaderSupport )

#define ILLUMINATE_NO_CONE_SIMD_IF_TRUE(RES, B, P, PS, L)  \
        DlIlluminate((RES), (P), 0, 0, (PS), (L), (B),  numMp, i_shadows, \
                       i_context.RayInfos(), shaderSupport )

#define SOLAR(IS_ILLU, AXIS, ANGLE, LIGHT)  \
        DlSolar(&(IS_ILLU), &(i_hiddens.GetPs()[polyCnt]), (AXIS), (ANGLE), &(LIGHT), \
                  0, 1, i_shadows, i_context.RayInfos(), shaderSupport )

#define SOLAR_SIMD(IS_ILLU, AXIS, ANGLE, LIGHT)  \
        DlSolar((IS_ILLU), (i_hiddens.GetPs()), (AXIS), (ANGLE), (LIGHT), 0, numMp, \
                  i_shadows, i_context.RayInfos(), shaderSupport )

#define SOLAR_SIMD_IF_TRUE(IS_ILLU, B, AXIS, ANGLE, LIGHT)  \
        DlSolar((IS_ILLU), (i_hiddens.GetPs()), (AXIS), (ANGLE), (LIGHT), (B),  numMp, \
                  i_shadows, i_context.RayInfos(), shaderSupport )		


#define SOLAR2(IS_ILLU, AXIS, ANGLE, LIGHT, PS)  \
        DlSolar(&(IS_ILLU), &(PS), (AXIS), (ANGLE), &(LIGHT), 0, 1, i_shadows, \
                  i_context.RayInfos(), shaderSupport )

#define SOLAR2_SIMD(IS_ILLU, AXIS, ANGLE, LIGHT, PS)  \
        DlSolar((IS_ILLU), (PS), (AXIS), (ANGLE), (LIGHT), 0, numMp, i_shadows, \
                  i_context.RayInfos(), shaderSupport )

#define SOLAR2_SIMD_IF_TRUE(IS_ILLU, B, AXIS, ANGLE, LIGHT, PS)  \
        DlSolar((IS_ILLU), (PS), (AXIS), (ANGLE), (LIGHT), (B),  numMp, i_shadows, \
                  i_context.RayInfos(), shaderSupport )		



/*
#define SOLAR(IS_ILLU, AXIS, ANGLE, LIGHT)  \
        DlSolar(&(IS_ILLU), (AXIS), (ANGLE), &(LIGHT), 0, 1)
#define SOLAR_SIMD(IS_ILLU, AXIS, ANGLE, LIGHT)  \
        DlSolar((IS_ILLU), (AXIS), (ANGLE), (LIGHT), 0, numMp)
#define SOLAR_SIMD_IF_TRUE(IS_ILLU, B, AXIS, ANGLE, LIGHT)  \
        DlSolar((IS_ILLU), (AXIS), (ANGLE), (LIGHT), (B),  numMp)		
*/



#define ILLUMINANCE(B0, P1, P2, F3)  \
        DlIlluminance(&(B0), &(P1), &(P2), (F3), &(i_lights[lightCnt].L[polyCnt]), 0, 1 )

#define ILLUMINANCE_SIMD(B0, P1, P2, F3)  \
        DlIlluminance((B0), (P1), (P2), (F3), i_lights[lightCnt].L, 0, numMp )

#define ILLUMINANCE_SIMD_IF_TRUE(B0, P1, P2, F3, B5)  \
        DlIlluminance((B0), (P1), (P2), (F3), i_lights[lightCnt].L, (B5),  numMp )


////////////////////////////////////////////////////////////////////////////
// MACROS FOR TEXTURE FUNCTIONS
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
// COMPUTE_ST_POINTS
////////////////////////////////////////////////////////////////////////////

#define COMPUTE_ST_POINTS(S0, T0, S1, T1, S2, T2, S3, T3, S4, T4)  \
        DlMakePointsST((S0), (T0), (S1), (T1), (S2), (T2), (S3), (T3), (S4), (T4), \
                                    numMpU, numMpV, 0)

#define COMPUTE_ST_POINTS_IF_TRUE(S0, T0, S1, T1, S2, T2, S3, T3, S4, T4, B)  \
        DlMakePointsST((S0), (T0), (S1), (T1), (S2), (T2), (S3), (T3), (S4), (T4), \
                                    numMpU, numMpV, (B))


////////////////////////////////////////////////////////////////////////////
// COMPUTE_POINTS
////////////////////////////////////////////////////////////////////////////

#define COMPUTE_POINTS(P0, P1, P2, P3, P4)  \
        DlMakePoints((P0), (P1), (P2), (P3), (P4), numMpU, numMpV, 0)

#define COMPUTE_POINTS_IF_TRUE(P0, P1, P2, P3, P4, B)  \
        DlMakePoints((P0), (P1), (P2), (P3), (P4), numMpU, numMpV, (B))


////////////////////////////////////////////////////////////////////////////
// CALL_TEXTURE_1D
//
// R                               : return result
// TX                              : name of the texture
// CH                              : channel
// S1, T1, S2, T2, S3, T3, S4, T4  : texture points
// V                               : variance
// SPLES                           : samples
// SB, TB                          : blur on 's' and 't' directions
// SW, TW                          : width on 's' and 't' directions
// FLTR                            : filter
// FIL                             : filler
// LERP                            : lerp (1=yes)
////////////////////////////////////////////////////////////////////////////

#define CALL_TEXTURE_1D( R, TX, CH, S1, T1, S2, T2, S3, T3, S4, T4, V, SPLES, SB, TB, SW, TW, FLTR, FIL, LERP ) \
	DlTexture( \
		&(R), 1, (TX), int(CH), &(S1), &(T1), &(S2), &(T2), &(S3), &(T3), \
		&(S4), &(T4), (V), int(SPLES), &(SB), &(TB), (SW), (TW), (FLTR), (FIL), \
		(LERP), 1, 1, 0, shaderSupport)

#define CALL_TEXTURE_1D_SIMD( R, TX, CH, S1, T1, S2, T2, S3, T3, S4, T4, V, SPLES, SB, TB, SW, TW, FLTR, FIL, LERP ) \
	DlTexture( \
		(R), 1, (TX), int(CH), (S1), (T1), (S2), (T2), (S3), (T3), (S4), (T4), \
		(V), int(SPLES), (SB), (TB), (SW), (TW), (FLTR), (FIL), (LERP), \
		numMpU, numMpV, 0, shaderSupport)

#define CALL_TEXTURE_1D_SIMD_IF_TRUE( R, B, TX, CH, S1, T1, S2, T2, S3, T3, S4, T4, V, SPLES, SB, TB, SW, TW, FLTR, FIL, LERP ) \
	DlTexture( \
		(R), 1, (TX), int(CH), (S1), (T1), (S2), (T2), (S3), (T3), \
		(S4), (T4), (V), int(SPLES), (SB), (TB), (SW), (TW), (FLTR), (FIL), (LERP), \
		numMpU, numMpV, (B), shaderSupport )


////////////////////////////////////////////////////////////////////////////
// CALL_TEXTURE_3D
//
// C1                              : return result
// TX                              : name of the texture
// CH                              : channel
// S1, T1, S2, T2, S3, T3, S4, T4  : texture points
// V                               : variance
// SPLES                           : samples
// SB, TB                          : blur on 's' and 't' directions
// SW, TW                          : width on 's' and 't' directions
// FLTR                            : filter
// FIL                             : filler
////////////////////////////////////////////////////////////////////////////

#define CALL_TEXTURE_3D( R, TX, CH, S1, T1, S2, T2, S3, T3, S4, T4, V, SPLES, SB, TB, SW, TW, FLTR, FIL, LERP ) \
	DlTexture( \
		(float*)(&(R)), 3, (TX), int(CH), &(S1), &(T1), &(S2), &(T2), &(S3), &(T3), &(S4), &(T4), \
		(V), int(SPLES), &(SB), &(TB), (SW), (TW), (FLTR), (FIL), (LERP), \
            	1, 1, 0, shaderSupport)

#define CALL_TEXTURE_3D_SIMD( R, TX, CH, S1, T1, S2, T2, S3, T3, S4, T4, V, SPLES, SB, TB, SW, TW, FLTR, FIL, LERP ) \
	DlTexture( \
		(float*)(R), 3, (TX), int(CH), (S1), (T1), (S2), (T2), (S3), (T3), (S4), (T4), \
		(V), int(SPLES), (SB), (TB), (SW), (TW), (FLTR), (FIL), (LERP), \
		numMpU, numMpV, 0, shaderSupport )

#define CALL_TEXTURE_3D_SIMD_IF_TRUE( R, B, TX, CH, S1, T1, S2, T2, S3, T3, S4, T4, V, SPLES, SB, TB, SW, TW, FLTR, FIL, LERP ) \
	DlTexture( \
		(float*)(R), 3, (TX), int(CH), (S1), (T1), (S2), (T2), (S3), (T3), (S4), (T4), \
		(V), int(SPLES), (SB), (TB), (SW), (TW), (FLTR), (FIL), (LERP), \
		numMpU, numMpV, (B), shaderSupport )


////////////////////////////////////////////////////////////////////////////
// CALL_ENVIRONEMENT_1D
//
// R              : return result (float)
// TX             : name of the texture
// CH             : channel
// P1, P2, P3, P4 : points of the texture
// V              : variance
// SPLES          : samples
// SB, TB         : blur on 's' and 't' directions
// SW, TW         : width on 's' and 't' directions
// FLTR           : filter
// FIL            : filler
////////////////////////////////////////////////////////////////////////////

#define CALL_ENVIRONMENT_1D( R, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, FIL, LERP ) \
	DlEnvironment(&(R), 1, (TX), int(CH), &(P1), &(P2), &(P3), &(P4), \
		(V), int(SPLES), &(SB), &(TB), (SW), (TW), (FLTR), (FIL), (LERP), \
		1, 1, 0, shaderSupport )

#define CALL_ENVIRONMENT_1D_SIMD(R, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, FIL, LERP) \
	DlEnvironment((R), 1, (TX), int(CH), (P1), (P2), (P3), (P4), \
		(V), int(SPLES), (SB), (TB), (SW), (TW), (FLTR), (FIL), (LERP), \
		numMpU, numMpV, 0, shaderSupport)

#define CALL_ENVIRONMENT_1D_SIMD_IF_TRUE(R, B, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, FIL, LERP) \
	DlEnvironment((R), 1, (TX), int(CH), (P1), (P2), (P3), (P4), \
		(V), int(SPLES), (SB), (TB), (SW), (TW), (FLTR), (FIL), (LERP), \
		numMpU, numMpV, (B), shaderSupport)


////////////////////////////////////////////////////////////////////////////
// CALL_ENVIRONEMENT_3D
//
// R              : return result (color)
// TX             : name of the texture
// CH             : channel
// P1, P2, P3, P4 : points of the texture
// V              : variance
// SPLES          : samples
// SB, TB         : blur on 's' and 't' directions
// SW, TW         : width on 's' and 't' directions
// FLTR           : filter
// FIL            : filler
////////////////////////////////////////////////////////////////////////////

#define CALL_ENVIRONMENT_3D( R, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, FIL, LERP ) \
	DlEnvironment((float*)(&(R)), 3, (TX), int(CH), &(P1), &(P2), &(P3), &(P4), \
		(V), (int)(SPLES), &(SB), &(TB), (SW), (TW), (FLTR), (FIL), (LERP), \
		1, 1, 0, shaderSupport)

#define CALL_ENVIRONMENT_3D_SIMD( R, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, FIL, LERP ) \
	DlEnvironment((float*)(R), 3, (TX), int(CH), (P1), (P2), (P3), (P4), \
		(V), (int)(SPLES), (SB), (TB), (SW), (TW), (FLTR), (FIL), (LERP), \
		numMpU, numMpV, 0, shaderSupport)

#define CALL_ENVIRONMENT_3D_SIMD_IF_TRUE( R, B, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, FIL, LERP ) \
	DlEnvironment((float*)(R), 3, (TX), int(CH), (P1), (P2), (P3), (P4), \
		(V), (int)(SPLES), (SB), (TB), (SW), (TW), (FLTR), (FIL), (LERP), \
		numMpU, numMpV, (B), shaderSupport)


////////////////////////////////////////////////////////////////////////////
// CALL_SHADOW_1D
//
// R              : return result
// TX             : name of the texture
// CH             : channel
// P1, P2, P3, P4 : points of the texture
// V              : variance
// SPLES          : samples
// SB, TB         : blur on 's' and 't' directions
// SW, TW         : width on 's' and 't' directions
// FLTR           : filter
// BIAS           : bias
////////////////////////////////////////////////////////////////////////////

#define CALL_SHADOW_1D(R, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, BIAS) \
        DlShadow(&(R), 1, (TX), int(CH), &(P1), &(P2), &(P3), &(P4), \
                   (V), (int)(SPLES), &(SB), &(TB), (SW), (TW), (FLTR), &(BIAS), 0,\
                   1, 1, 0, i_context.RayInfos(), shaderSupport)

#define CALL_SHADOW_1D_SIMD(R, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, BIAS) \
        DlShadow((R), 1, (TX), int(CH), (P1), (P2), (P3), (P4), \
                   (V), (int)(SPLES), (SB), (TB), (SW), (TW), (FLTR), (BIAS), 0,\
                   numMpU, numMpV, 0, i_context.RayInfos(), shaderSupport)

#define CALL_SHADOW_1D_SIMD_IF_TRUE(R, B, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, BIAS) \
        DlShadow((R), 1, (TX), int(CH), (P1), (P2), (P3), (P4), \
                   (V), (int)(SPLES), (SB), (TB), (SW), (TW), (FLTR), (BIAS), 0,\
                   numMpU, numMpV, 0, i_context.RayInfos(), shaderSupport)
/*-----------*/

#define CALL_AREASHADOW_1D(R, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, BIAS, NS, sP1, sP2, sP3, sP4) \
        DlAreaShadow(&(R), 1, (TX), int(CH), &(P1), &(P2), &(P3), &(P4), \
                   (V), (int)(SPLES), &(SB), &(TB), (SW), (TW), (FLTR), &(BIAS), 0,\
                   1, 1, 0, i_context.RayInfos(), shaderSupport, \
				   NS, sP1, sP2, sP3, sP4)

#define CALL_AREASHADOW_1D_SIMD(R, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, BIAS, NS, sP1, sP2, sP3, sP4) \
        DlAreaShadow((R), 1, (TX), int(CH), (P1), (P2), (P3), (P4), \
                   (V), (int)(SPLES), (SB), (TB), (SW), (TW), (FLTR), (BIAS), 0,\
                   numMpU, numMpV, 0, i_context.RayInfos(), shaderSupport,\
				   NS, sP1, sP2, sP3, sP4)

#define CALL_AREASHADOW_1D_SIMD_IF_TRUE(R, B, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, BIAS, NS, sP1, sP2, sP3, sP4) \
        DlAreaShadow((R), 1, (TX), int(CH), (P1), (P2), (P3), (P4), \
                   (V), (int)(SPLES), (SB), (TB), (SW), (TW), (FLTR), (BIAS), 0,\
                   numMpU, numMpV, 0, i_context.RayInfos(), shaderSupport, \
				   NS, sP1, sP2, sP3, sP4)
/*-----------*/

#define CALL_SHADOW_1D_LIGHT(R, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, BIAS) \
        DlShadow(&(R), 1, (TX), int(CH), &(P1), &(P2), &(P3), &(P4), \
                   (V), int(SPLES), &(SB), &(TB), (SW), (TW), (FLTR), &(BIAS),\
                   &((static_cast<const DlLightHidden&>(i_hiddens).GetL())[polyCnt]),\
                   1, 1, 0, i_context.RayInfos(), shaderSupport)

#define CALL_SHADOW_1D_LIGHT_SIMD(R, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, BIAS) \
        DlShadow((R), 1, (TX), int(CH), (P1), (P2), (P3), (P4), \
                      (V), int(SPLES), (SB), (TB), (SW), (TW), (FLTR), (BIAS),\
                      static_cast<const DlLightHidden&>(i_hiddens).GetL(),\
                      numMpU, numMpV, 0, i_context.RayInfos(), shaderSupport)

#define CALL_SHADOW_1D_LIGHT_SIMD_IF_TRUE(R, B, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, BIAS) \
        DlShadow((R), 1, (TX), int(CH), (P1), (P2), (P3), (P4), \
                      (V), int(SPLES), (SB), (TB), (SW), (TW), (FLTR), (BIAS),\
                      static_cast<const DlLightHidden&>(i_hiddens).GetL(),\
                      numMpU, numMpV, (B), i_context.RayInfos(), shaderSupport)

/*-----------*/

#define CALL_AREASHADOW_1D_LIGHT(R, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, BIAS, NS, sP1, sP2, sP3, sP4) \
        DlAreaShadow(&(R), 1, (TX), int(CH), &(P1), &(P2), &(P3), &(P4), \
                   (V), int(SPLES), &(SB), &(TB), (SW), (TW), (FLTR), &(BIAS),\
                   &((static_cast<const DlLightHidden&>(i_hiddens).GetL())[polyCnt]),\
                   1, 1, 0, i_context.RayInfos(), shaderSupport, \
				   NS, sP1, sP2, sP3, sP4)

#define CALL_AREASHADOW_1D_LIGHT_SIMD(R, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, BIAS, NS, sP1, sP2, sP3, sP4) \
        DlAreaShadow((R), 1, (TX), int(CH), (P1), (P2), (P3), (P4), \
                      (V), int(SPLES), (SB), (TB), (SW), (TW), (FLTR), (BIAS),\
                      static_cast<const DlLightHidden&>(i_hiddens).GetL(),\
                      numMpU, numMpV, 0, i_context.RayInfos(), shaderSupport, \
					  NS, sP1, sP2, sP3, sP4)

#define CALL_AREASHADOW_1D_LIGHT_SIMD_IF_TRUE(R, B, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, BIAS, NS, sP1, sP2, sP3, sP4) \
        DlAreaShadow((R), 1, (TX), int(CH), (P1), (P2), (P3), (P4), \
                      (V), int(SPLES), (SB), (TB), (SW), (TW), (FLTR), (BIAS),\
                      static_cast<const DlLightHidden&>(i_hiddens).GetL(),\
                      numMpU, numMpV, (B), i_context.RayInfos(), shaderSupport, \
					  NS, sP1, sP2, sP3, sP4)

////////////////////////////////////////////////////////////////////////////
// CALL_SHADOW_3D
//
// R              : return result
// TX             : name of the texture
// CH             : channel
// P1, P2, P3, P4 : points of the texture
// V              : variance
// SPLES          : samples
// SB, TB         : blur on 's' and 't' directions
// SW, TW         : width on 's' and 't' directions
// FLTR           : filter
// BIAS           : bias
////////////////////////////////////////////////////////////////////////////

#define CALL_SHADOW_3D(R, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, BIAS) \
        DlShadow((float*)(&(R)), 3, (TX), int(CH), &(P1), &(P2), &(P3), &(P4), \
                      (V), int(SPLES), &(SB), &(TB), (SW), (TW), (FLTR), &(BIAS), 0,\
                      1, 1, 0, i_context.RayInfos(), shaderSupport)

#define CALL_SHADOW_3D_SIMD(R, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, BIAS) \
        DlShadow((float*)(R), 3, (TX), int(CH), (P1), (P2), (P3), (P4), \
                      (V), int(SPLES), (SB), (TB), (SW), (TW), (FLTR), (BIAS), 0,\
                      numMpU, numMpV, 0, i_context.RayInfos(), shaderSupport)

#define CALL_SHADOW_3D_SIMD_IF_TRUE(R, B, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, BIAS) \
        DlShadow((float*)(R), 3, (TX), int(CH), (P1), (P2), (P3), (P4), \
                      (V), int(SPLES), (SB), (TB), (SW), (TW), (FLTR), (BIAS), 0,\
                      numMpU, numMpV, 0, i_context.RayInfos(), shaderSupport)

/*-----------*/
#define CALL_AREASHADOW_3D(R, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, BIAS, NS, sP1, sP2, sP3, sP4 ) \
        DlAreaShadow((float*)(&(R)), 3, (TX), int(CH), &(P1), &(P2), &(P3), &(P4), \
                      (V), int(SPLES), &(SB), &(TB), (SW), (TW), (FLTR), &(BIAS), 0,\
                      1, 1, 0, i_context.RayInfos(), shaderSupport, \
					  NS, sP1, sP2, sP3, sP4 )

#define CALL_AREASHADOW_3D_SIMD(R, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, BIAS, NS, sP1, sP2, sP3, sP4) \
        DlAreaShadow((float*)(R), 3, (TX), int(CH), (P1), (P2), (P3), (P4), \
                      (V), int(SPLES), (SB), (TB), (SW), (TW), (FLTR), (BIAS), 0,\
                      numMpU, numMpV, 0, i_context.RayInfos(), shaderSupport, \
					  NS, sP1, sP2, sP3, sP4 )

#define CALL_AREASHADOW_3D_SIMD_IF_TRUE(R, B, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, BIAS, NS, sP1, sP2, sP3, sP4) \
        DlAreaShadow((float*)(R), 3, (TX), int(CH), (P1), (P2), (P3), (P4), \
                      (V), int(SPLES), (SB), (TB), (SW), (TW), (FLTR), (BIAS), 0,\
                      numMpU, numMpV, 0, i_context.RayInfos(), shaderSupport, \
					  NS, sP1, sP2, sP3, sP4 )
/*-----------*/
#define CALL_SHADOW_3D_LIGHT(R, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, BIAS) \
        DlShadow((float*)(&(R)), 3, (TX), int(CH), &(P1), &(P2), &(P3), &(P4), \
                      (V), int(SPLES), &(SB), &(TB), (SW), (TW), (FLTR), &(BIAS),\
                      &((static_cast<const DlLightHidden&>(i_hiddens).GetL())[polyCnt]),\
                      1, 1, 0, i_context.RayInfos(), shaderSupport)

#define CALL_SHADOW_3D_LIGHT_SIMD(R, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, BIAS) \
        DlShadow((float*)(R), 3, (TX), int(CH), (P1), (P2), (P3), (P4), \
                      (V), int(SPLES), (SB), (TB), (SW), (TW), (FLTR), (BIAS),\
                      static_cast<const DlLightHidden&>(i_hiddens).GetL(),\
                      numMpU, numMpV, 0, i_context.RayInfos(), shaderSupport)

#define CALL_SHADOW_3D_LIGHT_SIMD_IF_TRUE(R, B, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, BIAS) \
        DlShadow((float*)(R), 3, (TX), int(CH), (P1), (P2), (P3), (P4), \
                      (V), int(SPLES), (SB), (TB), (SW), (TW), (FLTR), (BIAS),\
                      static_cast<const DlLightHidden&>(i_hiddens).GetL(),\
                      numMpU, numMpV, (B), i_context.RayInfos(), shaderSupport)

/*-----------*/
#define CALL_AREASHADOW_3D_LIGHT(R, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, BIAS, NS, sP1, sP2, sP3, sP4) \
        DlAreaShadow((float*)(&(R)), 3, (TX), int(CH), &(P1), &(P2), &(P3), &(P4), \
                      (V), int(SPLES), &(SB), &(TB), (SW), (TW), (FLTR), &(BIAS),\
                      &((static_cast<const DlLightHidden&>(i_hiddens).GetL())[polyCnt]),\
                      1, 1, 0, i_context.RayInfos(), shaderSupport, \
					  NS, sP1, sP2, sP3, sP4 )

#define CALL_AREASHADOW_3D_LIGHT_SIMD(R, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, BIAS, NS, sP1, sP2, sP3, sP4) \
        DlAreaShadow((float*)(R), 3, (TX), int(CH), (P1), (P2), (P3), (P4), \
                      (V), int(SPLES), (SB), (TB), (SW), (TW), (FLTR), (BIAS),\
                      static_cast<const DlLightHidden&>(i_hiddens).GetL(),\
                      numMpU, numMpV, 0, i_context.RayInfos(), shaderSupport, \
					  NS, sP1, sP2, sP3, sP4 )

#define CALL_AREASHADOW_3D_LIGHT_SIMD_IF_TRUE(R, B, TX, CH, P1, P2, P3, P4, V, SPLES, SB, TB, SW, TW, FLTR, BIAS, NS, sP1, sP2, sP3, sP4) \
        DlAreaShadow((float*)(R), 3, (TX), int(CH), (P1), (P2), (P3), (P4), \
                      (V), int(SPLES), (SB), (TB), (SW), (TW), (FLTR), (BIAS),\
                      static_cast<const DlLightHidden&>(i_hiddens).GetL(),\
                      numMpU, numMpV, (B), i_context.RayInfos(), shaderSupport, \
					  NS, sP1, sP2, sP3, sP4 )


////////////////////////////////////////////////////////////////////////////
// MACROS FOR MATRICES FUNCTIONS
////////////////////////////////////////////////////////////////////////////

#define CALL_PRIM_INVERSE(M1, M2)       DlInverse(&(M1), &(M2))

#define CALL_PRIM_DETERMINANT(F, M)     DlDeterminant(&(F), &(M))

#define CALL_PRIM_TRANSLATE(M1, M2, P)  DlTranslate(&(M1), &(M2), &(P))

#define CALL_PRIM_ROTATE(M1, M2, F, V)  DlRotate(&(M1), &(M2), &(F), &(V))

#define CALL_PRIM_SCALE(M1, M2, P)      DlScale(&(M1), &(M2), &(P))


////////////////////////////////////////////////////////////////////////////
// Functions for access functions
////////////////////////////////////////////////////////////////////////////

#define CALL_PRIM_TEXTUREINFO(RES, NAME, DATA, VAR, IS_VAR, TYPE, NB_ELEMS)  \
	DlTextureInfo(&(RES), (NAME), (DATA), &(VAR), int(IS_VAR), \
		DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
		0, 1 , shaderSupport)

#define CALL_PRIM_TEXTUREINFO_SIMD(RES, NAME, DATA, VAR, IS_VAR, TYPE, NB_ELEMS)  \
	DlTextureInfo(&(RES), (NAME), (DATA), (VAR), int(IS_VAR), \
		DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
		0, numMp, shaderSupport)

#define CALL_PRIM_TEXTUREINFO_SIMD_IF_TRUE(RES, B, NAME, DATA, VAR, IS_VAR, TYPE, NB_ELEMS)  \
	DlTextureInfo(&(RES), (NAME), (DATA), (VAR), int(IS_VAR), \
		DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
		(B),  numMp, shaderSupport)

#define CALL_PRIM_ATTRIBUTE(RES, NAME, VAR, IS_VAR, TYPE, NB_ELEMS)  \
	DlAttribute(&(RES), (NAME), &(VAR), int(IS_VAR), \
		DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
		0, 1, i_context)

#define CALL_PRIM_ATTRIBUTE_SIMD(RES, NAME, VAR, IS_VAR, TYPE, NB_ELEMS)  \
	DlAttribute(&(RES), (NAME), (VAR), int(IS_VAR), \
		DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
		0, numMp, i_context)

#define CALL_PRIM_ATTRIBUTE_SIMD_IF_TRUE(RES, B, NAME, VAR, IS_VAR, TYPE, NB_ELEMS)  \
	DlAttribute(&(RES), (NAME), (VAR), int(IS_VAR), \
		DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
		(B),  numMp, i_context)


#define CALL_PRIM_OPTION(RES, NAME, VAR, IS_VAR, TYPE, NB_ELEMS)  \
	DlOption(&(RES), (NAME), &(VAR), int(IS_VAR), \
		DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
		0, 1, i_context)

#define CALL_PRIM_OPTION_SIMD(RES, NAME, VAR, IS_VAR, TYPE, NB_ELEMS)  \
	DlOption(&(RES), (NAME), (VAR), int(IS_VAR), \
		DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
		0, numMp, i_context)

#define CALL_PRIM_OPTION_SIMD_IF_TRUE(RES, B, NAME, VAR, IS_VAR, TYPE, NB_ELEMS)  \
	DlOption(&(RES), (NAME), (VAR), int(IS_VAR), \
		DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
		(B), numMp, i_context)



#define CALL_PRIM_RENDERERINFO(RES, NAME, VAR, IS_VAR, TYPE, NB_ELEMS)  \
	DlRendererInfo(&(RES), (NAME), &(VAR), int(IS_VAR), \
		DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
		0, 1)

#define CALL_PRIM_RENDERERINFO_SIMD(RES, NAME, VAR, IS_VAR, TYPE, NB_ELEMS)  \
	DlRendererInfo(&(RES), (NAME), (VAR), int(IS_VAR), \
		DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
		0, numMp)

#define CALL_PRIM_RENDERERINFO_SIMD_IF_TRUE(RES, B, NAME, VAR, IS_VAR, TYPE, NB_ELEMS)  \
	DlRendererInfo(&(RES), (NAME), (VAR), int(IS_VAR), \
		DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
		(B),  numMp)

#define CALL_PRIM_SURFACE(RES, NAME, VAR, IS_VAR, TYPE, NB_ELEMS)  \
	DlMessageParam(&(RES), e_SurfaceParam, \
		(NAME), &(VAR), int(IS_VAR), \
		DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
		0, 1, i_context, 0)

#define CALL_PRIM_SURFACE_SIMD(RES, NAME, VAR, IS_VAR, TYPE, NB_ELEMS)  \
	DlMessageParam(&(RES),  e_SurfaceParam, \
		(NAME), (VAR), int(IS_VAR), \
		DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
		0, numMp, i_context, 0)

#define CALL_PRIM_SURFACE_SIMD_IF_TRUE(RES, B, NAME, VAR, IS_VAR, TYPE, NB_ELEMS)  \
	DlMessageParam(&(RES),  e_SurfaceParam, \
		(NAME), (VAR), int(IS_VAR), \
		DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
		(B),  numMp, i_context, 0)


#define CALL_PRIM_DISPLACEMENT(RES, NAME, VAR, IS_VAR, TYPE, NB_ELEMS)  \
	DlMessageParam(&(RES), e_DisplacementParam, \
		(NAME), &(VAR), int(IS_VAR), \
		DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
		0, 1, i_context, 0)

#define CALL_PRIM_DISPLACEMENT_SIMD(RES, NAME, VAR, IS_VAR, TYPE, NB_ELEMS)  \
    DlMessageParam(&(RES),  e_DisplacementParam, \
                                (NAME), (VAR), int(IS_VAR), \
                                DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
                                0, numMp, i_context, 0)

#define CALL_PRIM_DISPLACEMENT_SIMD_IF_TRUE(RES, B, NAME, VAR, IS_VAR, TYPE, NB_ELEMS)  \
    DlMessageParam(&(RES),  e_DisplacementParam, \
                                (NAME), (VAR), int(IS_VAR), \
                                DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
                                (B),  numMp, i_context, 0)


#define CALL_PRIM_LIGHTSOURCE(RES, NAME, VAR, IS_VAR, TYPE, NB_ELEMS)  \
    DlMessageParam(&(RES), e_LightsourceParam, \
                                (NAME), &(VAR), int(IS_VAR), \
                                DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
                                0, 1, i_context, lightCnt)

#define CALL_PRIM_LIGHTSOURCE_SIMD(RES, NAME, VAR, IS_VAR, TYPE, NB_ELEMS)  \
    DlMessageParam(&(RES),  e_LightsourceParam, \
                                (NAME), (VAR), int(IS_VAR), \
                                DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
                                0, numMp, i_context, lightCnt)

#define CALL_PRIM_LIGHTSOURCE_SIMD_IF_TRUE(RES, B, NAME, VAR, IS_VAR, TYPE, NB_ELEMS)  \
    DlMessageParam(&(RES),  e_LightsourceParam, \
                                (NAME), (VAR), int(IS_VAR), \
                                DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
                                (B),  numMp, i_context, lightCnt)


#define CALL_PRIM_ATMOSPHERE(RES, NAME, VAR, IS_VAR, TYPE, NB_ELEMS)  \
    DlMessageParam(&(RES), e_AtmosphereParam, \
                                (NAME), &(VAR), int(IS_VAR), \
                                DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
                                0, 1, i_context, 0)

#define CALL_PRIM_ATMOSPHERE_SIMD(RES, NAME, VAR, IS_VAR, TYPE, NB_ELEMS)  \
    DlMessageParam(&(RES),  e_AtmosphereParam, \
                                (NAME), (VAR), int(IS_VAR), \
                                DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
                                0, numMp, i_context, 0)

#define CALL_PRIM_ATMOSPHERE_SIMD_IF_TRUE(RES, B, NAME, VAR, IS_VAR, TYPE, NB_ELEMS)  \
    DlMessageParam(&(RES),  e_AtmosphereParam, \
                                (NAME), (VAR), int(IS_VAR), \
                                DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
                                (B),  numMp, i_context, 0)


#define CALL_PRIM_INCIDENT(RES, NAME, VAR, IS_VAR, TYPE, NB_ELEMS)  \
    DlMessageParam(&(RES), e_IncidentParam, \
                                (NAME), &(VAR), int(IS_VAR), \
                                DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
                                0, 1, i_context, 0)

#define CALL_PRIM_INCIDENT_SIMD(RES, NAME, VAR, IS_VAR, TYPE, NB_ELEMS)  \
    DlMessageParam(&(RES),  e_IncidentParam, \
                                (NAME), (VAR), int(IS_VAR), \
                                DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
                                0, numMp, i_context, 0)

#define CALL_PRIM_INCIDENT_SIMD_IF_TRUE(RES, B, NAME, VAR, IS_VAR, TYPE, NB_ELEMS)  \
    DlMessageParam(&(RES),  e_IncidentParam, \
                                (NAME), (VAR), int(IS_VAR), \
                                DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
                                (B),  numMp, i_context, 0)

#define CALL_PRIM_OPPOSITE(RES, NAME, VAR, IS_VAR, TYPE, NB_ELEMS)  \
    DlMessageParam(&(RES), e_OppositeParam, \
                                (NAME), &(VAR), int(IS_VAR), \
                                DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
                                0, 1, i_context, 0)

#define CALL_PRIM_OPPOSITE_SIMD(RES, NAME, VAR, IS_VAR, TYPE, NB_ELEMS)  \
    DlMessageParam(&(RES),  e_OppositeParam, \
                                (NAME), (VAR), int(IS_VAR), \
                                DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
                                0, numMp, i_context, 0)

#define CALL_PRIM_OPPOSITE_SIMD_IF_TRUE(RES, B, NAME, VAR, IS_VAR, TYPE, NB_ELEMS)  \
    DlMessageParam(&(RES),  e_OppositeParam, \
                                (NAME), (VAR), int(IS_VAR), \
                                DlTokenInfo::DlType(int(TYPE)), int(NB_ELEMS), \
                                (B),  numMp, i_context, 0)

////////////////////////////////////////////////////////////////////////////
// BOOLEAN FLAGS MACROS
////////////////////////////////////////////////////////////////////////////

#define SET_UNIFORM_BOOL_FLAG(BF, VF)  \
{  \
   (BF) = DL_SB_FALSE;  \
   for (polyCnt = 0; polyCnt < numMp; ++polyCnt)  \
   {  \
      if ((VF)[polyCnt]) { (BF) = DL_SB_TRUE; break; } \
   }  \
}

#define INVERT_BOOL_FLAG(RF, BF)  \
{  \
   for (polyCnt = 0; polyCnt < numMp; ++polyCnt)  \
   {  \
      (RF)[polyCnt] = !(BF)[polyCnt]; \
   }  \
}

#define INVERT_INNER_BOOL_FLAG(RF, BF, OF)  \
{  \
   for (polyCnt = 0; polyCnt < numMp; ++polyCnt)  \
   {  \
      (RF)[polyCnt] = ((OF)[polyCnt] && !((BF)[polyCnt])); \
   }  \
}


#define OR_BOOL_FLAG(BF, LF, RF)  \
{  \
   for (polyCnt = 0; polyCnt < numMp; ++polyCnt)  \
   {  \
      (BF)[polyCnt] = ((LF)[polyCnt] || (RF)[polyCnt]); \
   } \
}


////////////////////////////////////////////////////////////////////////////
// BMRT functions
////////////////////////////////////////////////////////////////////////////

#define CALL_PRIM_TRANSMISSION(C, P1, P2) \
	DlTransmission(&(C), &(P1), &(P2), 0, 1, 1, \
	i_context.RayInfos(), shaderSupport)

#define CALL_PRIM_TRANSMISSION_SIMD(C, P1, P2) \
	DlTransmission((C), (P1), (P2), 0, numMpU, numMpV, \
	i_context.RayInfos(), shaderSupport )

#define CALL_PRIM_TRANSMISSION_SIMD_IF_TRUE(C, B, P1, P2) \
	DlTransmission((C), (P1), (P2), (B), numMpU, numMpV, \
	i_context.RayInfos(), shaderSupport )

#define CALL_PRIM_ISSHADOWRAY(F) \
         DlIsShadowRay(&(F), i_context.RayInfos() )

#define CALL_PRIM_ISINDIRECTRAY( F ) \
         DlIsIndirectRay(&(F), i_context.RayInfos() )

#define CALL_PRIM_RAYLEVEL(F) \
         DlRayLevel(&(F), i_context.RayInfos() )

/* float trace(point, vector) shadeop */
#define CALL_PRIM_RAYHITTEST(RES, FROM, DIR) \
	DlRayHitTest(&(RES), &(FROM), &(DIR), 0x0, 1, 1, i_context.RayInfos(), shaderSupport)

#define CALL_PRIM_RAYHITTEST_SIMD(RES, FROM, DIR) \
	DlRayHitTest((RES), (FROM), (DIR), 0x0, numMpU, numMpV, i_context.RayInfos(), shaderSupport)

#define CALL_PRIM_RAYHITTEST_SIMD_IF_TRUE(RES, B, FROM, DIR) \
	DlRayHitTest((RES), (FROM), (DIR), (B), numMpU, numMpV, i_context.RayInfos(), shaderSupport)

/*
#define CALL_PRIM_FULLTRACE(RES, FROM, DIR, HITCOL, HITDIST, PHIT, NHIT, PMISS, HMISS) \
	DlFulltrace(&(RES), &(FROM), &(DIR), &(HITCOL), &(HITDIST), &(PHIT), &(NHIT), &(PMISS), &(HMISS), 0, 1, i_context.RayInfos(), shaderSupport)

#define CALL_PRIM_FULLTRACE_SIMD(RES, FROM, DIR, HITCOL, HITDIST, PHIT, NHIT, PMISS, HMISS) \
	DlFulltrace((RES), (FROM), (DIR), (HITCOL), (HITDIST), (PHIT), (NHIT), (PMISS), (HMISS), 0, numMp, i_context.RayInfos(), shaderSupport)

#define CALL_PRIM_FULLTRACE_SIMD_IF_TRUE(RES, B, FROM, DIR, HITCOL, HITDIST, PHIT, NHIT, PMISS, HMISS) \
	DlFulltrace((RES), (FROM), (DIR), (HITCOL), (HITDIST), (PHIT), (NHIT), (PMISS), (HMISS), (B), numMp, i_context.RayInfos(), shaderSupport)
*/

////////////////////////////////////////////////////////////////////////////
// FIXME: 'pack' and 'unpack' functions
////////////////////////////////////////////////////////////////////////////

#define CALL_PRIM_PACK(F1, F2, F3, F4, F5) \
     { \
          float temp1 = (F2); \
          float temp2 = (F3); \
          float temp3 = (F4); \
          float temp4 = (F5); \
          DlPack(&(F1), &(temp1), &(temp2), &(temp3), &(temp4), 0, 1); \
     }

#define CALL_PRIM_PACK_SIMD(F1, F2, F3, F4, F5) \
	DlPack((F1), (F2), (F3), (F4), (F5), 0, numMp)

#define CALL_PRIM_PACK_SIMD_IF_TRUE(F1, B6, F2, F3, F4, F5) \
	DlPack((F1), (F2), (F3), (F4), (F5), (B6), numMp)


#define CALL_PRIM_UNPACK(F1, F2, F3, F4, F5) \
{ \
     float temp1 = (F1); \
     DlUnpack(&temp, &(F2), &(F3), &(F4), &(F5), 0, 1); \
}

#define CALL_PRIM_UNPACK_SIMD(F1, F2, F3, F4, F5) \
     DlUnpack((F1), (F2), (F3), (F4), (F5), 0, numMp)

#define CALL_PRIM_UNPACK_SIMD_IF_TRUE(F1, B6, F2, F3, F4, F5) \
     DlUnpack((F1), (F2), (F3), (F4), (F5), (B6),  numMp)

#endif
